﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;© Big Eared Software 2017.
;==========================
;   Created with Purebasic 5.46 for Windows (tested on Win 7 and Ubuntu 16.04).
;
;   Platforms:  ALL.
;
;minScript compiler test program.
;qdMin - 'Quick and Dirty' editor. Used during our testing. Ideal to learn from!
;
;Use this little app to load a text file containing a script (including any of our demo ones), edit them, run them and save them etc.
;/////////////////////////////////////////////////////////////////////////////////

;Scripts run much faster with the debugger disabled.
DisableDebugger

;Add a reference to the Splitter library.
  XIncludeFile "splitter.pbi"

EnableExplicit

UsePNGImageDecoder()


;INIT minScript.
  ;Set the following constant to point to the main minScript source folder.
;Set the following constant to point to the main minScript source folder.
  CompilerSelect #PB_Compiler_OS
    CompilerCase #PB_OS_Windows
;      #minScript_HOME = "c:/PurebasicCode/minScript/Source/"  
    CompilerCase #PB_OS_Linux
;      #minScript_HOME = "/home/stephen/PurebasicCode/minScript/Source/"
  CompilerEndSelect
  CompilerIf Defined(minScript_HOME, #PB_Constant) = 0
    CompilerError "qdMin cannot run until you declare the constant : '#minScript_HOME' to point at the minScript source folder!"
  CompilerEndIf

  ;The following constant allows our debugger to retrieve disassembled object code during compilation.
    #minScript_OUTPUTASMTODEBUGGER = 1
  ;Include the minScript source and check that it initialized ok.
    XIncludeFile #minScript_HOME + "minScript.pbi"
    If minScriptINIT <> #minScript_OKAY
      MessageRequester("minScript - qdMin", "minScript failed to initialize correctly."+#CRLF$+"Closing qdMin", #PB_MessageRequester_Ok)
      End
    EndIf

;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES.
;==========

  Declare.i qdMin_OpenGUI()
  Declare qdMin_DisableGUIElementsForExecution(state)
  Declare qdMin_DisplayStatusMessage(text$)
  Declare.i qdMin_Init()
  Declare qdMin_ScintillaCallback(id, *scinotify.SCNotification)
  Declare.s qdMin_ScintillaGetText(id)
  Declare qdMin_ScintillaSetText(id, text$)
  Declare.i qdMin_minScriptDebugger(codeObject.minScriptCode, uMsg, param1, param2)
  Declare qdMin_OnButtonEvents()
  Declare qdMin_OnCloseMainWindow()
  Declare qdMin_OnCustomEvents()
  Declare qdMin_OnEditorEvents()
  Declare qdMin_OnMainMenuEvents()
  Declare qdMin_OnMainWindowResize()
  Declare.i qdMin_OnLeftSplitter(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)
  Declare.i qdMin_OnRightSplitter(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)
  Declare.i qdMin_OnVerticalSplitter(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS / STRUCTURES.
;========================

  ;Splitter metrics.
    #Splitter_Margin = 2  ;A buffer area we create around the slider.
    #Splitter_SliderWidth = 6   

  #qdMin_MarginAroundMainSplitter = 10 ;A gap we leave around the sides and bottom of the main splitter.

  #qdMin_GUI_MainWindow = 1

  ;Menus#'s.
    Enumeration
      #qdMin_GUI_MainMenu_CloseScript
      #qdMin_GUI_MainMenu_OpenScript
      #qdMin_GUI_MainMenu_SaveScript
    EndEnumeration
  
  ;Main GUI elements.
    Enumeration
      #qdMin_GUI_splVertical
        #qdMin_GUI_splLeft
          #qdMin_GUI_conScript
            #qdMin_GUI_imbCloseScript
            #qdMin_GUI_imbOpenScript
            #qdMin_GUI_imbSaveScript
            #qdMin_GUI_txtScript
            #qdMin_GUI_sciScript
            #qdMin_GUI_imbCompile
          #qdMin_GUI_conImmediates
            #qdMin_GUI_txtImmediates
            #qdMin_GUI_edtImmediate
            #qdMin_GUI_imbExecute
            #qdMin_GUI_imbKillExecution
        #qdMin_GUI_splRight
          #qdMin_GUI_ConOutput
            #qdMin_GUI_panOutput
              ;Panel 1.
                #qdMin_GUI_txtOutput
                #qdMin_GUI_lvOutput
                #qdMin_GUI_btnClearOutput
              ;Panel 2.
                #qdMin_GUI_lvASM
            #qdMin_GUI_conStatus
              #qdMin_GUI_txtStatus
              #qdMin_GUI_lvStatus
              #qdMin_GUI_btnClearStatus
    EndEnumeration

  ;Fonts.
    Enumeration
      #qdMin_FONT_status    
    EndEnumeration
  
  ;Custom events
    Enumeration #PB_Event_FirstCustomValue
      #qdMin_EVENT_STARTEXECUTION
      #qdMin_EVENT_ENDEXECUTION
      #qdMin_EVENT_OUTPUTTEXT
    EndEnumeration

  ;The following allows our worker thread to pass strings to the main process to update the text content of various list view gadgets.
    Structure qdMin_ThreadQueue
      gadgetID.i
      text$
    EndStructure

  ;The following structure wraps up all of our global variables.
    Structure qdMin_globals
      blnQuit.i
      toolbarHeight.i
      edtScriptHasHadFocus.i
      blnSaveEnabled.i
      fileName$
      ;Minscript fields.
        codeObject.minScriptCode
        *returnVar.minScriptVariant
      ;Execution fields - allows us to use a separate thread.
        execText$
        blnKillExecution.i
      ;Thread fields.
        threadMutex.i
        List threadQueue.qdMin_ThreadQueue()
        blnThreadRunning.i
      ;Splitter fields.
        verticalSplitter.Splitter
        leftSplitter.Splitter
        rightSplitter.Splitter
      ;Assist with GUI placement (issues on Linux GTK3)  
        btnClearOutputReqWidth.i
        btnClearStatusReqWidth.i
      ;Images.
        closeScriptPNG.i
        compileScriptPNG.i
        executePNG.i
        killExecutionPNG.i
        openScriptPNG.i
        saveScriptPNG.i
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS.
;========================

  Global gQDMIN.qdMin_globals
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-MAIN ROUTINE.
;==============

;Init qdMin.
  If qdMin_Init() = 0
    MessageRequester("minScript - qdMin", "qdMin failed to initialize correctly."+#CRLF$+"Closing qdMin", #PB_MessageRequester_Ok)
    End
  EndIf
;OPEN GUI.
  If qdMin_OpenGUI()
    Define eventID
    ;Main event loop. This covers any modal dialogs we may throw into the mix.
      Repeat
        eventID = WaitWindowEvent()
      Until gQDMIN\blnQuit
    gQDMIN\codeObject\Destroy()
  Else
    MessageRequester("minScript - qdMin", "Unable to open GUI."+#CRLF$+"Closing qdMin", #PB_MessageRequester_Ok)
  EndIf
End
;/////////////////////////////////////////////////////////////////////////////////


;-GUI FUNCTIONS.
;===============

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i qdMin_OpenGUI()
  Protected result = #True, text$, utf8Buffer, vertX, vertY
  result = OpenWindow(#qdMin_GUI_MainWindow, 300, 50, 800, 600, "minScript version " + #minScript_VERSION$ + " - qdMin", #PB_Window_MinimizeGadget|#PB_Window_MaximizeGadget|#PB_Window_SizeGadget|#PB_Window_TitleBar|#PB_Window_Maximize)
  If result
    CompilerIf #PB_Compiler_OS = #PB_OS_Windows 
      SetWindowLongPtr_(WindowID(#qdMin_GUI_MainWindow), #GWL_STYLE, GetWindowLongPtr_(WindowID(#qdMin_GUI_MainWindow), #GWL_STYLE) | #WS_CLIPCHILDREN)
    CompilerEndIf
    RemoveKeyboardShortcut(#qdMin_GUI_MainWindow, #PB_Shortcut_Tab) ;Required for the tab key to function correctly when the Scintilla control has the focus.
    AddKeyboardShortcut(#qdMin_GUI_MainWindow, #PB_Shortcut_Control|#PB_Shortcut_O, #qdMin_GUI_MainMenu_OpenScript)
    AddKeyboardShortcut(#qdMin_GUI_MainWindow, #PB_Shortcut_Control|#PB_Shortcut_S, #qdMin_GUI_MainMenu_SaveScript)
    AddKeyboardShortcut(#qdMin_GUI_MainWindow, #PB_Shortcut_Control|#PB_Shortcut_W, #qdMin_GUI_MainMenu_CloseScript)
    ;Main GUI elements.
      SetGadgetFont(#PB_Default, FontID(#qdMin_FONT_status))
    ;Toolbar replacement!
      ButtonImageGadget(#qdMin_GUI_imbOpenScript, 10, 0, 32, 32, ImageID(gQDMIN\openScriptPNG))
        GadgetToolTip(#qdMin_GUI_imbOpenScript, "Open script file (Ctrl+O)")
      ButtonImageGadget(#qdMin_GUI_imbSaveScript, 42, 0, 32, 32, ImageID(gQDMIN\saveScriptPNG))
        GadgetToolTip(#qdMin_GUI_imbSaveScript, "Save script file (Ctrl+S)")
      ButtonImageGadget(#qdMin_GUI_imbCloseScript, 74, 0, 32, 32, ImageID(gQDMIN\closeScriptPNG))
        GadgetToolTip(#qdMin_GUI_imbCloseScript, "Empty script editor (Ctrl+W)")
    gQDMIN\toolbarHeight = GadgetHeight(#qdMin_GUI_imbCloseScript) + 10
    ;Add gadgets.
    ;Left splitter and children.
      ContainerGadget(#qdMin_GUI_conScript, #qdMin_MarginAroundMainSplitter, gQDMIN\toolbarHeight, 0, 0, #PB_Container_Flat)
        TextGadget(#qdMin_GUI_txtScript, 10, 10, 120, 20, "Script panel")
        ScintillaGadget(#qdMin_GUI_sciScript, 10, 30, 0, 0, @qdMin_ScintillaCallback())
        ButtonImageGadget(#qdMin_GUI_imbCompile, 0, 0, 32, 32, ImageID(gQDMIN\compileScriptPNG))
        GadgetToolTip(#qdMin_GUI_imbCompile, "Compile current script ready for execution")
      CloseGadgetList()
      CompilerIf #PB_Compiler_OS = #PB_OS_Windows 
        SetWindowLongPtr_(GadgetID(#qdMin_GUI_conScript), #GWL_STYLE, GetWindowLongPtr_(GadgetID(#qdMin_GUI_conScript), #GWL_STYLE) | #WS_CLIPCHILDREN)
      CompilerEndIf
      ;We set a fixed line-number margin width for this unsophisticated demo program.
        ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_SETMARGINWIDTHN, 0, 30)
          ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_SETTABWIDTH, 2)
          ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_SETUSETABS, 0)
          ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLESETBACK, #STYLE_DEFAULT, $C3F8FF)
          ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLESETSIZE, #STYLE_DEFAULT, 9)
          ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLECLEARALL)
      ContainerGadget(#qdMin_GUI_conImmediates, #qdMin_MarginAroundMainSplitter, 0, 0, 0, #PB_Container_Flat)
        TextGadget(#qdMin_GUI_txtImmediates, 10, 10, 120, 20, "Execution panel")
        EditorGadget(#qdMin_GUI_edtImmediate, 10, 30, 0, 0)
        ButtonImageGadget(#qdMin_GUI_imbExecute, 0, 0, 32, 32, ImageID(gQDMIN\executePNG))
          GadgetToolTip(#qdMin_GUI_imbExecute, "Execute")
        ButtonImageGadget(#qdMin_GUI_imbKillExecution, 0, 0, 32, 32, ImageID(gQDMIN\killExecutionPNG))
          GadgetToolTip(#qdMin_GUI_imbKillExecution, "Kill execution")
      CloseGadgetList()
      CompilerIf #PB_Compiler_OS = #PB_OS_Windows 
        SetWindowLongPtr_(GadgetID(#qdMin_GUI_conImmediates), #GWL_STYLE, GetWindowLongPtr_(GadgetID(#qdMin_GUI_conImmediates), #GWL_STYLE) | #WS_CLIPCHILDREN)
      CompilerEndIf
    ;Right splitter and children.
      ContainerGadget(#qdMin_GUI_ConOutput, 0, gQDMIN\toolbarHeight, 0, 0, #PB_Container_Flat)
        PanelGadget(#qdMin_GUI_panOutput, 0, 0, 0, 0)
          AddGadgetItem(#qdMin_GUI_panOutput, -1, "Output")
            TextGadget(#qdMin_GUI_txtOutput, 10, 10, 10, 20, "Output/Debug Panel")
            ResizeGadget(#qdMin_GUI_txtOutput, #PB_Ignore, #PB_Ignore, GadgetWidth(#qdMin_GUI_txtOutput, #PB_Gadget_RequiredSize), GadgetHeight(#qdMin_GUI_txtOutput, #PB_Gadget_RequiredSize))
            ListViewGadget(#qdMin_GUI_lvOutput, 10, 30, 0, 0)
            ButtonGadget(#qdMin_GUI_btnClearOutput, 0, 0, 80, 20, "Clear Output") 
            AddGadgetItem(#qdMin_GUI_panOutput, -1, "Disassembled object code (ASM)")
            ListViewGadget(#qdMin_GUI_lvASM, 10, 10, 0, 0)
        CloseGadgetList()
      CloseGadgetList()
      gQDMIN\btnClearOutputReqWidth = GadgetWidth(#qdMin_GUI_btnClearOutput, #PB_Gadget_RequiredSize)
      CompilerIf #PB_Compiler_OS = #PB_OS_Windows 
        SetWindowLongPtr_(GadgetID(#qdMin_GUI_ConOutput), #GWL_STYLE, GetWindowLongPtr_(GadgetID(#qdMin_GUI_ConOutput), #GWL_STYLE) | #WS_CLIPCHILDREN)
      CompilerEndIf
      ContainerGadget(#qdMin_GUI_conStatus, 0, 0, 0, 0, #PB_Container_Flat)
        TextGadget(#qdMin_GUI_txtStatus, 10, 10, 120, 20, "Status")
        ListViewGadget(#qdMin_GUI_lvStatus, 10, 30, 0, 0)
;BUG IN PB/LINUX MEANS WE CANNOT SET COLORS.
;          SetGadgetColor(#qdMin_GUI_lvStatus, #PB_Gadget_BackColor, 0)
;          SetGadgetColor(#qdMin_GUI_lvStatus, #PB_Gadget_FrontColor, $FFFFFF)
        ButtonGadget(#qdMin_GUI_btnClearStatus, 0, 0, 80, 20, "Clear Status") 
      CloseGadgetList()
      gQDMIN\btnClearStatusReqWidth = GadgetWidth(#qdMin_GUI_btnClearStatus, #PB_Gadget_RequiredSize)
      CompilerIf #PB_Compiler_OS = #PB_OS_Windows 
        SetWindowLongPtr_(GadgetID(#qdMin_GUI_conStatus), #GWL_STYLE, GetWindowLongPtr_(GadgetID(#qdMin_GUI_conStatus), #GWL_STYLE) | #WS_CLIPCHILDREN)
      CompilerEndIf
    ;Position the main splitter.
      vertX = 2*WindowWidth(#qdMin_GUI_MainWindow)/3
      vertY = 2*WindowHeight(#qdMin_GUI_MainWindow)/3
      gQDMIN\verticalSplitter = NewSplitter(vertX, gQDMIN\toolbarHeight, #Splitter_SliderWidth, WindowHeight(#qdMin_GUI_MainWindow) - gQDMIN\toolbarHeight - #qdMin_MarginAroundMainSplitter, @qdMin_OnVerticalSplitter(), #Splitter_VERTICAL)
      ;Position the left splitter.  
      gQDMIN\leftSplitter = NewSplitter(#qdMin_MarginAroundMainSplitter, vertY, vertX - #Splitter_Margin - #qdMin_MarginAroundMainSplitter, #Splitter_SliderWidth, @qdMin_OnLeftSplitter())
    ;Position the right splitter.  
      gQDMIN\rightSplitter = NewSplitter(vertX + #Splitter_SliderWidth + #Splitter_Margin, vertY, WindowWidth(#qdMin_GUI_MainWindow) - vertX - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter, #Splitter_SliderWidth, @qdMin_OnRightSplitter())
    ;Create event handlers.
      ;Close window.
        BindEvent(#PB_Event_CloseWindow, @qdMin_OnCloseMainWindow(), #qdMin_GUI_MainWindow)
      ;Menus.
        BindEvent(#PB_Event_Menu, @qdMin_OnMainMenuEvents())
      ;Resizing.
        BindEvent(#PB_Event_SizeWindow, @qdMin_OnMainWindowResize(), #qdMin_GUI_MainWindow)
      ;Editor events.
        BindGadgetEvent(#qdMin_GUI_edtImmediate, @qdMin_OnEditorEvents(), #PB_EventType_Focus)
        BindGadgetEvent(#qdMin_GUI_edtImmediate, @qdMin_OnEditorEvents(), #PB_EventType_Change)
      ;Buttons.
        BindGadgetEvent(#qdMin_GUI_imbOpenScript, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_imbSaveScript, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_imbCloseScript, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_imbCompile, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_imbExecute, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_imbKillExecution, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_btnClearOutput, @qdMin_OnButtonEvents())
        BindGadgetEvent(#qdMin_GUI_btnClearStatus, @qdMin_OnButtonEvents())
      ;Custom events.
        BindEvent(#qdMin_EVENT_STARTEXECUTION, @qdMin_OnCustomEvents(), #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_STARTEXECUTION)
        BindEvent(#qdMin_EVENT_ENDEXECUTION, @qdMin_OnCustomEvents(), #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_ENDEXECUTION)
        BindEvent(#qdMin_EVENT_OUTPUTTEXT, @qdMin_OnCustomEvents(), #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_OUTPUTTEXT)
    ;Finalise.
      ;Set initial message in Scintilla control.
        text$ = ">>>USE THIS EDITOR TO ENTER/LOAD/COPY SCRIPT CODE (in the form of CLASS/ENDCLASS and/or FUNCTION/ENDFUNCTION blocks)<<<" + #LF$ + 
                #LF$ + "For example : " + #LF$ + 
                #LF$ + "      FUNCTION AddIntegers(a AS INTEGER, b AS INTEGER)" + #LF$ +
                "        RETURN a+b" + #LF$ +
                "      ENDFUNCTION" + #LF$ +
                #LF$ + "Hit the compile button (below) before attempting to test the code using the editor below." + #LF$ +
                "Any compilation errors can be seen in the panels on the right hand side of the main window." + #LF$ +
                #LF$ +  "Add as many functions/classes to this editor as you wish." + #LF$ +
                #LF$ +  "NOTE THAT THIS SIMPLE EDITOR ONLY ALLOWS A SINGLE SCRIPT MODULE."
        qdMin_ScintillaSetText(#qdMin_GUI_sciScript, text$)
      ;Set initial message in immediates editor.
        SetGadgetText(#qdMin_GUI_edtImmediate, ">>>USE THIS EDITOR TO TEST THE SCRIPTS ENTERED IN THE EDITOR ABOVE<<<")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "Type in some code directly to call script functions (or standard run-time functions). For example, enter the following command :")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "")
          AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "      DEBUG(AddIntegers(10, 20))")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "And hit the 'EXECUTE' button (below).")
          AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "Any errors or return values can be seen in the panels on the right hand side of the main window.")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "")
        AddGadgetItem(#qdMin_GUI_edtImmediate, -1, "IF EXECUTION FREEZES OR RUNS AN INFINITE LOOP THEN HIT THE 'KILL EXECUTION' BUTTON.")
      ;Set initial messages in status panel.
        qdMin_DisplayStatusMessage("Utilising minScript version " + #minScript_VERSION$)
        qdMin_DisplayStatusMessage("qdMin initialised OK")
        qdMin_DisplayStatusMessage("Ready")
      ;Disable / enable as appropriate.
        DisableGadget(#qdMin_GUI_imbSaveScript, #True)
        DisableGadget(#qdMin_GUI_imbKillExecution, #True)
        DisableGadget(#qdMin_GUI_imbExecute, #True)
    ;A workaround for a problem with retrieving gadget positions under PB 5.46 Ubuntu GTK3.
      Delay(200) : While WindowEvent() : Wend
  qdMin_OnMainWindowResize()  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-MISCELLANEOUS FUNCTIONS.
;=========================

;/////////////////////////////////////////////////////////////////////////////////
Procedure qdMin_DisableGUIElementsForExecution(state)
  DisableGadget(#qdMin_GUI_imbCloseScript, state)
  DisableGadget(#qdMin_GUI_imbOpenScript, state)
  DisableGadget(#qdMin_GUI_imbCompile, state)
  DisableGadget(#qdMin_GUI_edtImmediate, state)
  DisableGadget(#qdMin_GUI_imbExecute, state)
  DisableGadget(#qdMin_GUI_imbKillExecution, 1-state)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure qdMin_DisplayStatusMessage(text$)
  AddGadgetItem(#qdMin_GUI_lvStatus, -1, "["+FormatDate("%hh:%ii:%ss", Date()) +"] " + text$)
  SetGadgetState(#qdMin_GUI_lvStatus, CountGadgetItems(#qdMin_GUI_lvStatus)-1)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;INIT function.
Procedure.i qdMin_Init()
  Protected result
  CompilerIf #PB_Compiler_OS = #PB_OS_Windows
    CompilerIf #PB_Compiler_Processor = #PB_Processor_x86 
      If InitScintilla("Scintillax86.dll")
    CompilerElse
      If InitScintilla("Scintillax64.dll")
    CompilerEndIf
  CompilerElse
    If InitScintilla() ;Not really needed.
  CompilerEndIf
    ;Create a thread mutex.
      gQDMIN\threadMutex = CreateMutex()
    If gQDMIN\threadMutex
      ;Init the minScript code object.
        gQDMIN\codeObject = minScript_NewCodeObject()
      If gQDMIN\codeObject
        gQDMIN\returnVar = gQDMIN\codeObject\GetReturnVariantPtr()
        gQDMIN\codeObject\SetDebuggerCallback(@qdMin_minScriptDebugger())
        ;We assume there will be no problem loading the images or fonts.
          gQDMIN\closeScriptPNG = CatchImage(#PB_Any, ?CloseScriptPNG)
          gQDMIN\compileScriptPNG = CatchImage(#PB_Any, ?CompileScriptPNG)
          gQDMIN\executePNG = CatchImage(#PB_Any, ?ExecutePNG)
          gQDMIN\killExecutionPNG = CatchImage(#PB_Any, ?KillExecutionPNG)
          gQDMIN\openScriptPNG = CatchImage(#PB_Any, ?OpenScriptPNG)
          gQDMIN\saveScriptPNG = CatchImage(#PB_Any, ?SaveScriptPNG)
          LoadFont(#qdMin_FONT_status, "", 9)
        result = #True
      Else
        FreeMutex(gQDMIN\threadMutex)
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following converts any minScript error code to text (English).
;An empty return denotes #minScript_OKAY
Procedure.s qdMin_minScriptStatusCode2Text(errorCode)
  Protected result$
  Select errorCode
    Case #minScript_OKAY ;Do nothing.
    Case #minScript_KILLEXECUTION
      result$ = "Execution killed."    
    Case #minScript_UNEXPECTEDERROR
      result$ = "Unexpected error."
    Case #minScript_OUTOFMEMORY
      result$ = "Out of memory."
    Case #minScript_INVALIDARGUMENT
      result$ = "Invalid argument."
    Case #minScript_INCOMPATIBLEDATATYPES
      result$ = "Incompatible data types."
    Case #minScript_INVALIDDATATYPE
      result$ = "Invalid data type."
    Case #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
      result$ = "Symbol name matches compiler keyword."
    Case #minScript_FUNCTIONIDMISMATCH
      result$ = "Function ID mismatch."
    Case #minScript_ACCESSDENIED
      result$ = "Access denied."
    Case #minScript_FUNCTIONIDALREADYEXISTS
      result$ = "Function ID already exists."
    Case #minScript_MODULENOTFOUND
      result$ = "Module not found."
    Case #minScript_FUNCTIONNOTFOUND
      result$ = "Function not found."
    Case #minScript_FUNCTIONORMETHODALREADYDEFINED
      result$ = "Function or method already defined."
    Case #minScript_INVALIDTOKEN
      result$ = "Invalid token."  
    Case #minScript_CODEOUTSIDEFUNCTIONBLOCK
      result$ = "This code is not allowed outside of function/method block."
    Case #minScript_NESTEDFUNCTIONSORMETHODS
      result$ = "Cannot declare a function/method whilst inside another function/method."
    Case #minScript_INVALIDFUNCTIONORMETHODDECLARATION
      result$ = "Invalid function/method declaration."
    Case #minScript_INVALIDUSEOFPRIVATE
      result$ = "Invalid use of 'PRIVATE'."
    Case #minScript_INVALIDPARAMETER
      result$ = "Invalid parameter."
    Case #minScript_SYNTAXERROR
      result$ = "Syntax error."
    Case #minScript_INVALIDSYMBOLNAME
      result$ = "Invalid symbol name."
    Case #minScript_STATEMENTTOOCOMPLEX
      result$ = "Statement too complex."
    Case #minScript_MISSINGTERMINATINGSYMBOL
      result$ = "Missing terminating symbol."
    Case #minScript_CONSTANTNOTFOUND
      result$ = "Constant not found."
    Case #minScript_CONSTANTALREADYDECLARED
      result$ = "Constant already declared."
    Case #minScript_INVALIDCONSTANTEXPRESSION
      result$ = "Invalid constant expression."
    Case #minScript_VARIABLENOTFOUND
      result$ = "Variable not found."
    Case #minScript_MISSINGENDFUNCTIONORMETHOD
      result$ = "Missing ENDFUNCTION/ENDMETHOD."
    Case #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
      result$ = "Symbol name matches run-time function name."
    Case #minScript_UNRECOGNISEDVARIABLETYPE
      result$ = "Unrecognised variable type."
    Case #minScript_TOOMANYPARAMETERS
      result$ = "Too many parameters."
    Case #minScript_VARIABLEORPROPERTYALREADYDECLARED
      result$ = "Variable or property already declared."
    Case #minScript_OBJECTNOTINITIALIZED
      result$ = "Object not initialised."
    Case #minScript_MISSINGENDCONSTRUCT
      result$ = "Missing end construct."
    Case #minScript_UNEXPECTEDENDCONSTRUCT
      result$ = "Unexpected end construct."
    Case #minScript_CASEORDEFAULTEXPECTED
      result$ = "CASE or DEFAULT expected."
    Case #minScript_DEFAULTORELSEALREADYENCOUNTERED
      result$ = "DEFAULT or ELSE already encountered."
    Case #minScript_INVALIDSTEPVALUE
      result$ = "Invalid step value."
    Case #minScript_INVALIDBREAKLEVEL
      result$ = "Invalid break level."
    Case #minScript_INVALIDUSEOFBREAK
      result$ = "Invalid use of BREAK."
    Case #minScript_INVALIDORNULLOBJECTVARIABLE
      result$ = "Invalid or null object variable."
    Case #minScript_INVALIDUSEOFTYPEMODIFIER
      result$ = "Invalid use of a type modifier."
    Case #minScript_INVALIDTYPEMODIFIER
      result$ = "Invalid type modifier."
    Case #minScript_TYPEMISMATCH
      result$ = "Type mismatch."
    Case #minScript_NOTALLOWEDOUTSIDEOFMODULE
      result$ = "Not allowed outside of a module."
    Case #minScript_CLASSNOTFOUND
      result$ = "Class not found."
    Case #minScript_CLASSALREADYDECLARED
      result$ = "Class already declared."
    Case #minScript_MISSINGENDCLASS
      result$ = "Missing ENDCLASS."
    Case #minScript_INVALIDCLASSDECLARATION
      result$ = "Invalid class declaration."
    Case #minScript_NESTEDCLASSES
      result$ = "Cannot declare a class within another class declaration."
    Case #minScript_NOTALLOWEDOUTSIDEOFCLASS
      result$ = "Not allowed outside of a class declaration."
    Case #minScript_NOTALLOWEDINSIDECLASS
      result$ = "Not allowed inside a class declaration."
    Case #minScript_INTERNALSTACKOVERFLOW
      result$ = "Internal stack overflow."
    Case #minScript_VARIANTSTACKOVERFLOW
      result$ = "Variant stack overflow."
    Case #minScript_INVALIDUSEOFVARIANTSTACK
      result$ = "Invalid use of variant stack."
    Case #minScript_VARIANTSTACKCORRUPTED
      result$ = "Variant stack corrupted."
    Case #minScript_DIVISIONBYZERO
      result$ = "Attempted division by zero."
    Case #minScript_INCORRECTNUMBEROFPARAMETERS
      result$ = "Incorrect number of parameters."
    Case #minScript_METHODORPROPERTYNOTSUPPORTED
      result$ = "Method or property not supported."
    Case #minScript_PROPERTYISREADONLY
      result$ = "Property is read only."
    Case #minScript_NOCURRENTELEMENT
      result$ = "There is no current element."
    Case #minScript_INVALIDORMISSINGARRAYINDEX
      result$ = "Invalid or missing array index."
    Case #minScript_NULLPOINTER
      result$ = "Null pointer."
    Case #minScript_TOOMANYHASHMAPPUSHES
      result$ = "Too many hashmap pushes."
    Case #minScript_HASHMAPPOSITIONNOTPUSHED
      result$ = "Hashmap position not pushed."
    Default
      result$ = "Unknown error." ;Just in case we've missed one or a script THROWS an unknown error!
  EndSelect
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Our Scintilla callback.
Procedure qdMin_ScintillaCallback(id, *scinotify.SCNotification)
  Select *scinotify\nmhdr\code
    Case #SCN_FOCUSIN
      If gQDMIN\edtScriptHasHadFocus = #False
        ScintillaSendMessage(id, #SCI_CLEARALL)
        gQDMIN\edtScriptHasHadFocus = #True
        DisableGadget(#qdMin_GUI_imbSaveScript, #False)
        gQDMIN\blnSaveEnabled = #True
      EndIf
    Case #SCN_MODIFIED
      ;We arrange for the compile script toolbar button to be shown or otherwsise.
      If *scinotify\modificationType &(#SC_MOD_INSERTTEXT|#SC_MOD_DELETETEXT)
        DisableGadget(#qdMin_GUI_imbSaveScript, #False)
        If gQDMIN\edtScriptHasHadFocus And ScintillaSendMessage(id, #SCI_GETLENGTH)
          DisableGadget(#qdMin_GUI_imbCompile, #False)
          gQDMIN\blnSaveEnabled = #True
        Else          
          DisableGadget(#qdMin_GUI_imbCompile, #True)
          ClearGadgetItems(#qdMin_GUI_lvASM)
        EndIf
      EndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function retrieves the text for the entire control.
Procedure.s qdMin_ScintillaGetText(id)
  Protected text$, utf8Buffer, numBytes
  numBytes = ScintillaSendMessage(id, #SCI_GETLENGTH)
  If numBytes
    utf8Buffer = AllocateMemory(numBytes+1)
    If utf8Buffer 
      ScintillaSendMessage(id, #SCI_GETTEXT, numBytes + 1, utf8Buffer)
      text$ = PeekS(utf8Buffer, -1, #PB_UTF8)
      FreeMemory(utf8Buffer)
      EndIf
    EndIf
  ProcedureReturn text$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure qdMin_ScintillaSetText(id, text$)
  Protected utf8Buffer
  utf8Buffer = AllocateMemory(StringByteLength(text$, #PB_UTF8)+1)
  If utf8Buffer 
    PokeS(utf8Buffer, text$, -1, #PB_UTF8)
    ScintillaSendMessage(id, #SCI_SETTEXT, 0, utf8Buffer)
    FreeMemory(utf8Buffer)
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;-THREAD PROCEDURES.
;===================

;/////////////////////////////////////////////////////////////////////////////////
;The following runs in a separate thread and is used to execute code, leaving the main process free to monitor the kill execution button etc.
Procedure qdMin_ExecutionThread(param)
  gQDMIN\blnThreadRunning = #True
  ;Execute the code.
    gQDMIN\codeObject\Execute(gQDMIN\execText$, #True) ;#True to enable debugging.
  ;Finally, we post a #qdMin_EVENT_ENDEXECUTION message.
    PostEvent(#qdMin_EVENT_ENDEXECUTION, #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_ENDEXECUTION)
  gQDMIN\blnThreadRunning = #False
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-MISCELLANEOUS MINSCRIPT FUNCTIONS.
;===================================

;/////////////////////////////////////////////////////////////////////////////////
;The following is our minScript debugger which we use firstly to output disassembled object code during compilation and secondly to
;determine if the user kills the execution of object code etc.
;This can be invoked either from the main process (since we don't bother with a separate thread to compile the main scripts) or from the execution thread.
Procedure.i qdMin_minScriptDebugger(codeObject.minScriptCode, uMsg, param1, param2)
  Protected result = #minScript_OKAY, *ptrString.STRING, *ptrVariant.minScriptVariant
  Select uMsg
    Case #minScript_DEBUGGER_COMPILE_OUTPUTASM
      *ptrString = param2
      ;Is the thread running?
      If gQDMIN\blnThreadRunning
        ;We need to add the new ASM to the thread queue.
          LockMutex(gQDMIN\threadMutex)
            LastElement(gQDMIN\threadQueue())
            AddElement(gQDMIN\threadQueue())
            gQDMIN\threadQueue()\gadgetID = #qdMin_GUI_lvASM
            gQDMIN\threadQueue()\text$ = *ptrString\s
          UnlockMutex(gQDMIN\threadMutex)
        ;Post a suitable #qdMin_EVENT_OUTPUTTEXT message.
          PostEvent(#qdMin_EVENT_OUTPUTTEXT, #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_OUTPUTTEXT)
      Else ;We can output the ASM directly.
        AddGadgetItem(#qdMin_GUI_lvASM, -1, *ptrString\s)
      EndIf
    Case #minScript_DEBUGGER_DEBUG
      *ptrVariant = param1
      ;Convert to a string.
      If minScript_VarChangeType(*ptrVariant, #minScript_STRING) = #minScript_OKAY
        ;We need to add the new output to the thread queue.
          LockMutex(gQDMIN\threadMutex)
            LastElement(gQDMIN\threadQueue())
            AddElement(gQDMIN\threadQueue())
            gQDMIN\threadQueue()\gadgetID = #qdMin_GUI_lvOutput
            If *ptrVariant\sPtr
              gQDMIN\threadQueue()\text$ = PeekS(*ptrVariant\sPtr)
            Else
              gQDMIN\threadQueue()\text$ = ""
            EndIf
          UnlockMutex(gQDMIN\threadMutex)
        ;Post a suitable #qdMin_EVENT_OUTPUTTEXT message.
          PostEvent(#qdMin_EVENT_OUTPUTTEXT, #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_OUTPUTTEXT)
        ;A short delay allows the main process time to keep an eye on the kill execution button in the case that a massive loop full of DEBUG statements is
        ;holding things up.
          Delay(10)
      EndIf
    Default ;This will be invoked only whilst a script is being executed. We use this to check whether the kill execution button has been pushed.
      If gQDMIN\blnKillExecution
        result = #minScript_KILLEXECUTION
      EndIf
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-EVENT HANDLERS.
;================

;/////////////////////////////////////////////////////////////////////////////////
Procedure qdMin_OnButtonEvents()
  Protected result = #minScript_OKAY, text$, file$, fileID, stringFormat, i
  Select EventGadget()
    Case #qdMin_GUI_imbOpenScript
      file$ = OpenFileRequester("Please choose source file to open", "", "Text (*.txt)|*.txt|All files (*.*)|*.*", 0)
      If file$
        ;Open the file.
        fileID = ReadFile(#PB_Any, file$)
        If fileID
          ;Read BOM.
            stringFormat = ReadStringFormat(fileID)
          If stringFormat = #PB_Ascii Or stringFormat = #PB_Unicode Or StringFormat = #PB_UTF8
            ;Read file.
              While Eof(fileID) = #False
                If text$ = ""
                  text$ = ReadString(FileID, stringFormat)
                Else
                  text$ + #LF$ + ReadString(FileID, stringFormat)
                EndIf
              Wend
            gQDMIN\edtScriptHasHadFocus = #True
            gQDMIN\fileName$ = file$
            ;Set scintilla text.
              qdMin_ScintillaSetText(#qdMin_GUI_sciScript, text$)
            gQDMIN\codeObject\Reset(#minScript_RESET_REMOVEALL)
            qdMin_DisplayStatusMessage("'" + GetFilePart(file$) + "' successfully opened.")
          Else
            qdMin_DisplayStatusMessage("Unable to read file!")
          EndIf
          CloseFile(fileID)
        Else
          qdMin_DisplayStatusMessage("Unable to open file!")
        EndIf       
      Else
        qdMin_DisplayStatusMessage("Operation cancelled.")
      EndIf
    Case #qdMin_GUI_imbSaveScript
      ;Retrieve text.
        ;Convert all EOL characters to #CRLF.
          ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_CONVERTEOLS, #SC_EOL_CRLF)
        text$ = qdMin_ScintillaGetText(#qdMin_GUI_sciScript)
      file$ = SaveFileRequester("Please choose source file to save (text format)", gQDMIN\fileName$, "Text (*.txt)|*.txt|All files (*.*)|*.*", 0)
      If file$
        If GetExtensionPart(file$) = ""
          file$ + ".txt"
        EndIf          
        ;Does file already exist.
          If FileSize(file$) >= 0
            If MessageRequester("minScript - qdMin", "Delete existing file (" + file$ + ") (Y/N)?", #PB_MessageRequester_YesNoCancel) <> #PB_MessageRequester_Yes 
              qdMin_DisplayStatusMessage("Operation cancelled.")
              ProcedureReturn
            EndIf            
          EndIf
        FileID = CreateFile(#PB_Any, file$, #PB_File_NoBuffering)
        If fileID
          If WriteStringFormat(fileID, #PB_UTF8)
            If text$ = "" Or WriteString(fileID, text$)
              CloseFile(fileID)
              qdMin_DisplayStatusMessage("'" + GetFilePart(file$) + "' successfully saved.")
              gQDMIN\fileName$ = file$
            Else
              qdMin_DisplayStatusMessage("Unable to write file!")
              CloseFile(fileID)
              DeleteFile(file$)
            EndIf
          Else
            qdMin_DisplayStatusMessage("Unable to write file!")
            CloseFile(fileID)
            DeleteFile(file$)
          EndIf
        Else
          qdMin_DisplayStatusMessage("Unable to create file!")
          qdMin_DisplayStatusMessage("File not saved.")
        EndIf
      Else
        qdMin_DisplayStatusMessage("Operation cancelled.")
      EndIf
    Case #qdMin_GUI_imbCloseScript
      qdMin_ScintillaSetText(#qdMin_GUI_sciScript, "")
      gQDMIN\edtScriptHasHadFocus = #True
      gQDMIN\codeObject\Reset(#minScript_RESET_REMOVEALL)
      gQDMIN\fileName$ = ""
      qdMin_DisplayStatusMessage("Script closed.")
    Case #qdMin_GUI_imbCompile
      ;Clear the ASM output.
        ClearGadgetItems(#qdMin_GUI_lvASM)
      ;Reset the code object.
        result = gQDMIN\codeObject\Reset(#minScript_RESET_REMOVEALL)
      If result = #minScript_OKAY
        ;Retrieve text to compile. Guaranteed to be non-empty.
          text$ = qdMin_ScintillaGetText(#qdMin_GUI_sciScript)
        ;Compile the code in the form of a module.
          result = gQDMIN\codeObject\AddModule("Module1", @text$)
          SetGadgetState(#qdMin_GUI_lvASM, CountGadgetItems(#qdMin_GUI_lvASM)-1)
        If result = #minScript_OKAY
          qdMin_DisplayStatusMessage("Compilation successful.")
        Else
          qdMin_DisplayStatusMessage("Compilation error (code " + Str(result) + ") : " + qdMin_minScriptStatusCode2Text(result))
          qdMin_DisplayStatusMessage("    Error line number : " + Str(gQDMIN\codeObject\GetLastErrorLineNumber()))
          qdMin_DisplayStatusMessage("    Source line : " + gQDMIN\codeObject\GetLastCompilationErrorLine())
        EndIf
      Else
        text$ = "Error : " + qdMin_minScriptStatusCode2Text(result)
        qdMin_DisplayStatusMessage(text$)
      EndIf
    Case #qdMin_GUI_imbExecute
      ;Retrieve immediate text. Guaranteed not to be empty.
        gQDMIN\execText$ = GetGadgetText(#qdMin_GUI_edtImmediate)
      ;Post a #qdMin_EVENT_STARTEXECUTION message to the main window which will execute the code in a separate thread.
        PostEvent(#qdMin_EVENT_STARTEXECUTION, #qdMin_GUI_MainWindow, 0, #qdMin_EVENT_STARTEXECUTION)
    Case #qdMin_GUI_imbKillExecution
      gQDMIN\blnKillExecution = #True
    Case #qdMin_GUI_btnClearStatus
      ClearGadgetItems(#qdMin_GUI_lvStatus)
    Case #qdMin_GUI_btnClearOutput
      ClearGadgetItems(#qdMin_GUI_lvOutput)
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure qdMin_OnCloseMainWindow()
  gQDMIN\blnQuit = #True
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure qdMin_OnCustomEvents()
  Protected result, name$
  Select EventType()
    Case #qdMin_EVENT_STARTEXECUTION
      ;Disable all pertinant GUI elements but the killExecution button.
        qdMin_DisableGUIElementsForExecution(#True)
      ;Set back color of the scintilla control to white.
        ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLESETBACK, #STYLE_DEFAULT, $FFFFFF)
        ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLECLEARALL)
      ;Clear list of disassembled object code.
        ClearGadgetItems(#qdMin_GUI_lvASM)
      ;Reset the blnKillExecution global field.
        gQDMIN\blnKillExecution = #False
      ;Create thread to execute the code.
        CreateThread(@qdMin_ExecutionThread(), 0)
    Case #qdMin_EVENT_ENDEXECUTION
      ;Enable all pertinant GUI elements but the killExecution button.
        qdMin_DisableGUIElementsForExecution(#False)
      ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLESETBACK, #STYLE_DEFAULT, $C3F8FF)
      ScintillaSendMessage(#qdMin_GUI_sciScript, #SCI_STYLECLEARALL)
      ;Report errors/status.
        result = gQDMIN\codeObject\GetLastErrorCode()
        If result = #minScript_OKAY
          qdMin_DisplayStatusMessage("No execution errors.")
        Else
          qdMin_DisplayStatusMessage("Execution error (code " + Str(result) + ") : " + qdMin_minScriptStatusCode2Text(result))
          If result <> #minScript_KILLEXECUTION
            qdMin_DisplayStatusMessage("    Error Source line number : " + Str(gQDMIN\codeObject\GetLastErrorLineNumber()))
            name$ = gQDMIN\codeObject\GetLastExecutionErrorFunctionName()
            If name$
              qdMin_DisplayStatusMessage("    Error occurred within function : " + Chr(34) + name$ + Chr(34))
            Else
              qdMin_DisplayStatusMessage("    Error occurred in the statement submitted for execution.")
            EndIf
          EndIf
        EndIf
    Case #qdMin_EVENT_OUTPUTTEXT
      ;We just retrieve the first entry in the thread queue and display as appropriate.
      LockMutex(gQDMIN\threadMutex)
        FirstElement(gQDMIN\threadQueue())
        Select gQDMIN\threadQueue()\gadgetID
          Case #qdMin_GUI_lvASM
            AddGadgetItem(#qdMin_GUI_lvASM, -1, gQDMIN\threadQueue()\text$)
            SetGadgetState(#qdMin_GUI_lvASM, CountGadgetItems(#qdMin_GUI_lvASM)-1)
          Case #qdMin_GUI_lvOutput
            AddGadgetItem(#qdMin_GUI_lvOutput, -1, gQDMIN\threadQueue()\text$)
            SetGadgetState(#qdMin_GUI_lvOutput, CountGadgetItems(#qdMin_GUI_lvOutput)-1)
        EndSelect
        DeleteElement(gQDMIN\threadQueue())
      UnlockMutex(gQDMIN\threadMutex)
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;We use the following event handler to wipe the initial messages we post in the various editor gadgets when the said gadget first gets the focus.
Procedure qdMin_OnEditorEvents()
  Static st_edtImmediateHasHadFocus
  Select EventType()
    Case #PB_EventType_Focus 
      If st_edtImmediateHasHadFocus = #False
        SetGadgetText(#qdMin_GUI_edtImmediate, "")
        st_edtImmediateHasHadFocus = #True
      EndIf
    Case #PB_EventType_Change
      If GetGadgetText(#qdMin_GUI_edtImmediate)
        DisableGadget(#qdMin_GUI_imbExecute, #False)      
      Else
        DisableGadget(#qdMin_GUI_imbExecute, #True)      
      EndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The main menu handler.
Procedure qdMin_OnMainMenuEvents()
  Select EventMenu()
    Case #qdMin_GUI_MainMenu_CloseScript
      PostEvent(#PB_Event_Gadget, #qdMin_GUI_MainWindow, #qdMin_GUI_imbCloseScript)
    Case #qdMin_GUI_MainMenu_OpenScript
      PostEvent(#PB_Event_Gadget, #qdMin_GUI_MainWindow, #qdMin_GUI_imbOpenScript)
    Case  #qdMin_GUI_MainMenu_SaveScript
      If gQDMIN\blnSaveEnabled
        PostEvent(#PB_Event_Gadget, #qdMin_GUI_MainWindow, #qdMin_GUI_imbSaveScript)
      EndIf
  EndSelect 
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The main window resizing handler.
Procedure qdMin_OnMainWindowResize()
  If gQDMIN\verticalSplitter ;Just in case.
    gQDMIN\verticalSplitter\Resize(#PB_Ignore, #PB_Ignore, #PB_Ignore, WindowHeight(#qdMin_GUI_MainWindow) - gQDMIN\toolbarHeight - #qdMin_MarginAroundMainSplitter)
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-Custom splitter control event handlers.
;----------------------------------------

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i qdMin_OnLeftSplitter(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)
  Protected result, width, height
  Static sBlnCon1Hidden, sBlnCon2Hidden
  Select uMsg
    Case #Splitter_RequestSliderPosChange
      If *y\i <= gQDMIN\toolbarHeight + #Splitter_Margin
        *y\i = gQDMIN\toolbarHeight + #Splitter_Margin
        HideGadget(#qdMin_GUI_conScript, #True)  
        sBlnCon1Hidden = #True
      ElseIf sBlnCon1Hidden = #True
        HideGadget(#qdMin_GUI_conScript, #False)  
        sBlnCon1Hidden = #False
      EndIf
      If *y\i + #Splitter_SliderWidth + #Splitter_Margin > WindowHeight(#qdMin_GUI_MainWindow) - #qdMin_MarginAroundMainSplitter
        If sBlnCon1Hidden = #False
          *y\i = WindowHeight(#qdMin_GUI_MainWindow) - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter
        EndIf  
        HideGadget(#qdMin_GUI_conImmediates, #True)  
        sBlnCon2Hidden= #True
      ElseIf sBlnCon2Hidden = #True
        HideGadget(#qdMin_GUI_conImmediates, #False)  
        sBlnCon2Hidden = #False
      EndIf
      If *y\i <> this\GetY()
        result = #True
      EndIf  
    Case #Splitter_SliderPosChanged
      ResizeGadget(#qdMin_GUI_conScript, #PB_Ignore, #PB_Ignore, this\GetWidth(), *y\i - #Splitter_Margin - gQDMIN\toolbarHeight)
      width = GadgetWidth(#qdMin_GUI_conScript)
      height = GadgetHeight(#qdMin_GUI_conScript)
      ResizeGadget(#qdMin_GUI_sciScript, #PB_Ignore, #PB_Ignore, width-20, height-76)
      ResizeGadget(#qdMin_GUI_imbCompile, width-42, height-41,  #PB_Ignore, #PB_Ignore)
      If height > 76
        HideGadget(#qdMin_GUI_imbCompile, #False)  
      Else
        HideGadget(#qdMin_GUI_imbCompile, #True)  
      EndIf
      ResizeGadget(#qdMin_GUI_conImmediates, #PB_Ignore, *y\i + #Splitter_SliderWidth + #Splitter_Margin, width, WindowHeight(#qdMin_GUI_MainWindow) - *y\i - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter)
      height = GadgetHeight(#qdMin_GUI_conImmediates)
      ResizeGadget(#qdMin_GUI_edtImmediate, #PB_Ignore, #PB_Ignore, width-20, height-76)
      ResizeGadget(#qdMin_GUI_imbKillExecution, width-42, height-41,  #PB_Ignore, #PB_Ignore)
      ResizeGadget(#qdMin_GUI_imbExecute, width-90, height-41,  #PB_Ignore, #PB_Ignore)
      If height > 76
        HideGadget(#qdMin_GUI_imbKillExecution, #False)  
        HideGadget(#qdMin_GUI_imbExecute, #False)  
      Else
        HideGadget(#qdMin_GUI_imbKillExecution, #True)  
        HideGadget(#qdMin_GUI_imbExecute, #True)  
      EndIf
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i qdMin_OnRightSplitter(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)
  Protected result, width, height
  Static sBlnCon1Hidden, sBlnCon2Hidden
  Select uMsg
    Case #Splitter_RequestSliderPosChange
      If *y\i <= gQDMIN\toolbarHeight + #Splitter_Margin
        *y\i = gQDMIN\toolbarHeight + #Splitter_Margin
        HideGadget(#qdMin_GUI_ConOutput, #True)  
        sBlnCon1Hidden = #True
      ElseIf sBlnCon1Hidden = #True
        HideGadget(#qdMin_GUI_ConOutput, #False)  
        sBlnCon1Hidden = #False
      EndIf
      If *y\i + #Splitter_SliderWidth + #Splitter_Margin > WindowHeight(#qdMin_GUI_MainWindow) - #qdMin_MarginAroundMainSplitter
        If sBlnCon1Hidden = #False
          *y\i = WindowHeight(#qdMin_GUI_MainWindow) - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter
        EndIf  
        HideGadget(#qdMin_GUI_conStatus, #True)  
        sBlnCon2Hidden= #True
      ElseIf sBlnCon2Hidden = #True
        HideGadget(#qdMin_GUI_conStatus, #False)  
        sBlnCon2Hidden = #False
      EndIf
      If *y\i <> this\GetY()
        result = #True
      EndIf  
    Case #Splitter_SliderPosChanged
      ResizeGadget(#qdMin_GUI_ConOutput, *x\i, #PB_Ignore, this\GetWidth(), *y\i - #Splitter_Margin - gQDMIN\toolbarHeight)
      ResizeGadget(#qdMin_GUI_panOutput, #PB_Ignore, #PB_Ignore, GadgetWidth(#qdMin_GUI_ConOutput), GadgetHeight(#qdMin_GUI_ConOutput))
      width = GetGadgetAttribute(#qdMin_GUI_panOutput, #PB_Panel_ItemWidth)
      height = GetGadgetAttribute(#qdMin_GUI_panOutput, #PB_Panel_ItemHeight)
      ResizeGadget(#qdMin_GUI_lvOutput, #PB_Ignore, #PB_Ignore, width-20, height-70)
      ResizeGadget(#qdMin_GUI_btnClearOutput, width-gQDMIN\btnClearOutputReqWidth-10, height-35, #PB_Ignore, #PB_Ignore)
      ResizeGadget(#qdMin_GUI_lvASM, #PB_Ignore, #PB_Ignore, width-20, height-20)
      If height > 70
        HideGadget(#qdMin_GUI_btnClearOutput, #False)  
      Else
        HideGadget(#qdMin_GUI_btnClearOutput, #True)  
      EndIf
      ResizeGadget(#qdMin_GUI_conStatus, *x\i, *y\i + #Splitter_SliderWidth + #Splitter_Margin, this\GetWidth(), WindowHeight(#qdMin_GUI_MainWindow) - *y\i - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter)
      width = GadgetWidth(#qdMin_GUI_conStatus)
      height = GadgetHeight(#qdMin_GUI_conStatus)
      ResizeGadget(#qdMin_GUI_lvStatus, #PB_Ignore, #PB_Ignore, width-20, height-70)
      ResizeGadget(#qdMin_GUI_btnClearStatus, width-gQDMIN\btnClearStatusReqWidth-10, height-35, #PB_Ignore, #PB_Ignore)
      If height > 70
        HideGadget(#qdMin_GUI_btnClearStatus, #False)  
      Else
        HideGadget(#qdMin_GUI_btnClearStatus, #True)  
      EndIf
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i qdMin_OnVerticalSplitter(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)
  Protected result
  Static sBlnCon1Hidden, sBlnCon2Hidden
  Select uMsg
    Case #Splitter_RequestSliderPosChange
      If *x\i <= #qdMin_MarginAroundMainSplitter + #Splitter_Margin
        *x\i = #qdMin_MarginAroundMainSplitter + #Splitter_Margin
        HideGadget(#qdMin_GUI_conScript, #True)  
        HideGadget(#qdMin_GUI_conImmediates, #True)
        gQDMIN\leftSplitter\Hide(#True)
        sBlnCon1Hidden = #True
      ElseIf sBlnCon1Hidden = #True
        HideGadget(#qdMin_GUI_conScript, #False)  
        HideGadget(#qdMin_GUI_conImmediates, #False)  
        gQDMIN\leftSplitter\Hide(#False)
        sBlnCon1Hidden = #False
      EndIf
      If *x\i + #Splitter_SliderWidth + #Splitter_Margin > WindowWidth(#qdMin_GUI_MainWindow) - #qdMin_MarginAroundMainSplitter
        If sBlnCon1Hidden = #False
          *x\i = WindowWidth(#qdMin_GUI_MainWindow) - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter
        EndIf  
        HideGadget(#qdMin_GUI_ConOutput, #True)  
        HideGadget(#qdMin_GUI_conStatus, #True)  
        gQDMIN\rightSplitter\Hide(#True)
        sBlnCon2Hidden= #True
      ElseIf sBlnCon2Hidden = #True
        HideGadget(#qdMin_GUI_ConOutput, #False)  
        HideGadget(#qdMin_GUI_conStatus, #False)  
        gQDMIN\rightSplitter\Hide(#False)
        sBlnCon2Hidden = #False
      EndIf
      If *x\i <> this\GetX()
        result = #True
      EndIf  

    Case #Splitter_SliderPosChanged
      If gQDMIN\leftSplitter
        gQDMIN\leftSplitter\Resize(#PB_Ignore, #PB_Ignore, *x\i - #Splitter_Margin - #qdMin_MarginAroundMainSplitter, #Splitter_SliderWidth)
      EndIf
      If gQDMIN\rightSplitter
        gQDMIN\rightSplitter\Resize(*x\i + #Splitter_SliderWidth + #Splitter_Margin, #PB_Ignore, WindowWidth(#qdMin_GUI_MainWindow) - *x\i - #Splitter_SliderWidth - #Splitter_Margin - #qdMin_MarginAroundMainSplitter, #Splitter_SliderWidth)
      EndIf
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


DataSection
  CloseScriptPNG: 
    IncludeBinary "Images/close.png"
  CompileScriptPNG: 
    IncludeBinary "Images/compile.png"
  ExecutePNG:
    IncludeBinary "Images/execute.png"
  KillExecutionPNG:
    IncludeBinary "Images/kill.png"
  OpenScriptPNG:
    IncludeBinary "Images/open.png"
  SaveScriptPNG:
    IncludeBinary "Images/save.png"
EndDataSection
; IDE Options = PureBasic 5.62 (Windows - x86)
; CursorPosition = 32
; FirstLine = 3
; Folding = -----
; EnableThread
; EnableXP
; Executable = qdMin.exe
; EnableUnicode