﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Code Compiler class.
;
;© Big Eared Software 2016.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;/////////////////////////////////////////////////////////////////////////////////



;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScriptCodeCompilerClass_PARSER_ExtractNextSymbol(*this, *module.st_minScriptModule)
  Declare.i minScriptCodeCompilerClass_EXPRESSION_BitwiseShiftTerms(*this, expressionVT)
  Declare.i minScriptCodeCompilerClass_EXPRESSION_ComparisonTerms(*this, expressionVT)
  Declare.i minScriptCodeCompilerClass_EXPRESSION_MulDivTerms(*this, expressionVT, *ptrIsNegative.INTEGER)
  Declare.i minScriptCodeCompilerClass_EXPRESSION_PushNextOperand(*this, expressionVT, *ptrIsNegative.INTEGER)
  Declare.i minScriptCodeCompilerClass_EXPRESSION_Recurse(*this, expressionVT)

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

  #minScript_COMPILER_SIZEOFPARSETREE = 1024

  ;Parse tree symbol types.
    Enumeration 1
      #minScript_COMPILER_SYMBOL_KEYWORD
      #minScript_COMPILER_SYMBOL_SEPARATOR
      #minScript_COMPILER_SYMBOL_LOCALVARIABLE                    ;The variable may not have been declared at this point.
      #minScript_COMPILER_SYMBOL_MODULEVARIABLE                   ;The variable WILL have been declared at this point.
      #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
      #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
      #minScript_COMPILER_SYMBOL_LITERALNUMBER
      #minScript_COMPILER_SYMBOL_LITERALSTRING
      #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION
      #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION                 ;The function may not have been declared at this point.
      #minScript_COMPILER_SYMBOL_OBJECTMETHOD
    EndEnumeration

  ;Construct types. Used to check for the correct nesting of such constructs etc.
    Enumeration
      #minScript_COMPILER_CONSTRUCT_FOR
      #minScript_COMPILER_CONSTRUCT_IF
      #minScript_COMPILER_CONSTRUCT_REPEAT
      #minScript_COMPILER_CONSTRUCT_SELECT
      #minScript_COMPILER_CONSTRUCT_WHILE
    EndEnumeration
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-INTERFACES

  ;All methods return a minScript error code unless otherwise stated.

  ;The following interface is the 'public' face of this class.
    Interface minScriptCodeCompiler
      Destroy()
      Reset(flags=0)
      ;The following method is responsible for compiling code.
        CompileCode.i(*ptrCode.CHARACTER, *module.st_minScriptModule=0)
    EndInterface

  ;The following interface extends the above one and is used internally, mostly for compiling individual statements as listed.
  ;All methods return a minScript status code.
    Interface minScriptCodeCompilerEX Extends minScriptCodeCompiler
      EmitObjectCodeInstruction.i(opCode, operand.q=0)
      ;minScript keywords.
        Break.i(*module.st_minScriptModule)
        Case.i()
        CLASS.i(*module.st_minScriptModule)
        Debug.i()
        Default.i()
        Define.i(*module.st_minScriptModule, blnShortCutAssignment=0)
        Dim.i(*module.st_minScriptModule, blnInsideFunction)
        DIRECTIVE.i()
        Else.i()
        ElseIf.i()
        ENDCLASS.i()
        ENDFUNCTION.i()
        EndIf.i()
        EndSelect.i()
        For.i(*module.st_minScriptModule)
        FUNCTION.i(*module.st_minScriptModule)
        If.i()
        LET.i(blnShortCutAssignment=0)
        Next.i()
        POKEYWORD.i(address)
        PROPERTY.i()
        Repeat.i()
        Return.i()
        Select.I()
        Static.i(*module.st_minScriptModule)
        Until.i()
        Wend.i()
        While.i()
      ;Methods for assisting with code labels
        AddVirtualCodeLabel.i(label$, objectCodeOffset=-1)    ;Returns the address of the nominal entry in the code labels map.
                                                              ;If zero is returned then there is a memory allocation issue.
        FixUpCodeReferences(*objectCodeBufferPtr.st_minScript_memoryPointer)
      ;Methods for assisting with nested constructs.
        PopConstruct.i()                  ;Returns 0 if there are no constructs pushed; else returns 1.
        PushConstruct.i(constructType)    ;Returns an eScript error code (only possible error is #eScript_OUTOFMEMORY).
        
    EndInterface
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following structure forms an array of parse symbols.
    Structure st_minScript_ParseTree
      symbol$                         ;Lowercase (except for literal strings)
      ;The following union allows us to quickly pull out details of the underlying symbol when the symbol represents a number,
      ;a variable or a function call etc.
      StructureUnion
        *localVariable.st_minScript_LocalVariablesSymbol
        *moduleVariable.minScriptVariant
        *scriptVariable.minScriptVariant
        *runtimeFunction.st_minScriptRuntimeFunction
        *scriptFunction.st_minScriptFunction
        *constant.st_minScriptGlobalConstant
      EndStructureUnion
      typeModifierVT.u
      invokeType.u
      symbolType.b                    ;E.g. #minScript_COMPILER_SYMBOL_KEYWORD
    EndStructure

  ;The following is used to hold information on nested constructs; IF/ELSE/ENDIF etc. and are held in a stack.
    Structure st_minScript_Construct
      type.i              ;E.g. #minScript_COMPILER_CONSTRUCT_FOR
      constructIndex1.i
      ;The following fields are dependent on the underlying construct.
        StructureUnion
          blnELSEEncountered.i
          blnDEFAULTEncountered.i
          stepValue.i             ;FOR/NEXT
        EndStructureUnion      
        StructureUnion
          opPushVar.i             ;FOR/NEXT
          numCASEs.i
          numELSEIFs.i
        EndStructureUnion      
        StructureUnion
          varOperand.i            ;FOR/NEXT
        EndStructureUnion
        StructureUnion
          forVar$
        EndStructureUnion
    EndStructure

  ;The following is used for a local variable when compiling FUNCTION statements.
    Structure st_minScript_ParseParameters
      varName$[#minScript_MAXNUMFUNCTIONPARAMETERS]
      vt.u[#minScript_MAXNUMFUNCTIONPARAMETERS]
    EndStructure


  ;The following is the class template for the minScriptCodeCompiler class.
    Structure st_minScriptCodeCompilerClass
      *vTable
      thisEX.minScriptCodeCompilerEX  ;For convenience.
      *code.st_minScriptCodeClass     ;Parent object.
      ;Source code fields.
        *ptrSource.CHARACTER
        ;Current line being compiled.
          currentLineNumber.i
          currentLine$
      ;Fields for 'local code' labels.
      ;The map uses the label name (lowercase) as the key.
      ;Each entry is an offset from the beginning of the object code file. A value of -1 means that the label has not been initialised etc.
      ;Note that such labels are only used in an implicit fashion, not actually forming any part of the source code and as such no error
      ;checking is required.
        Map codeLabels.i()
      ;Parse tree.
        numParseSymbols.i
        parseTree.st_minScript_ParseTree[#minScript_COMPILER_SIZEOFPARSETREE]
        *parseCharPtr.CHARACTER             ;Used to index the next character in the source stream.
        *parseSymbol.CHARACTER              ;Used to point to the beginning of the next symbol.
      ;Expression fields.
        expressionSymbolIndex.i             ;Zero based index into the parseTree.
      ;Constructs.
        List constructs.st_minScript_Construct()
        numTotalConstructs.i                ;Used to generate unique internal labels.
        blnLastStatementASelect.i           ;To ensure only CASE/DEFAULT/ENDSELECT statements follow a SELECT.
      ;Class creation.
        *currentClass.st_minScriptClass     ;Points at the relevant code map entry to assist with class creation.
      ;Object code creation.
        memoryFile.MemoryFile
        blnObjectCodeBeingBuffered.i        ;When set, any object code subsequently created will not necessarily remain in the final build.
                                            ;Setting this will suspend any debug output.
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////



;-INTERNAL FUNCTIONS (those not intended for public consumption).

;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of the minScriptCodeCompiler class.
;Returns a newly instantiated object or zero if an error (out of memory).
Procedure.i minScript_NewCodeCompilerObject(code.minScriptCode)
  Protected *this.st_minScriptCodeCompilerClass
  ;Attempt to allocate memory for a st_minScriptCodeCompilerClass class template.
    *this = AllocateStructure(st_minScriptCodeCompilerClass)
  If *this
    *this\vTable = ?VTable_minScriptCodeCompilerClass
    *this\thisEX = *this
    *this\code = code
    *this\memoryFile = NewMemoryFile(#minScript_INITIALMEMORYFILESIZE, #minScript_MEMORYFILEPAGESIZE, #MEMORYFILE_VARIABLEBUFFER)
    If *this\memoryFile
  
    Else
      FreeStructure(*this)
      *this = 0    
    EndIf
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-COMPILER DEBUG FUNCTION.

;/////////////////////////////////////////////////////////////////////////////////
;The following is the 'central hub' through which all compiler debug output is channelled.
;Centralising this facility allows us to switch output mediums if required.
CompilerIf Defined(minScript_OUTPUTASMTODEBUGGER, #PB_Constant)
  Procedure.i minScriptCodeCompilerClass_FN_OutputDebugASM(*this.st_minScriptCodeCompilerClass, text$)
    Protected address
    ;If a debugger is installed, send a #minScript_DEBUGGER_COMPILE_OUTPUTASM debug message.
      If *this\code\debugger And minScript_Globals\blnNoDebugASM = #False
        address = @text$
        *this\code\debugger(*this\code, #minScript_DEBUGGER_COMPILE_OUTPUTASM, *this\currentLineNumber, @address)
      EndIf
  EndProcedure
CompilerEndIf
;/////////////////////////////////////////////////////////////////////////////////


;-PARSING FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following function begins the process of creating a parse tree (more of an array really) from the current line of source.
;If there are no errors then on top of initialising the process, it only extracts the first symbol.
Procedure.i minScriptCodeCompilerClass_PARSER_BeginParseTree(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule)
  Protected result = #minScript_OKAY
  ;Reset parse tree.
    *this\numParseSymbols = 0
    *this\parseCharPtr = @*this\currentLine$
    *this\parseSymbol = *this\parseCharPtr
  ;Extract first symbol.
    result = minScriptCodeCompilerClass_PARSER_ExtractNextSymbol(*this, *module)
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function continues the process of creating a parse tree from the current line of source up to the given delimiter (optional).
;Used for those instances where processing may need to be completed in phases.
;It does not examine the individual symbols in order to categorise those which are not already categorised. This is instead left to 
;minScriptCodeCompilerClass_PARSER_FinishParseTree() which must be called to complete the processing.
Procedure.i minScriptCodeCompilerClass_PARSER_ContinueParseTree(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule, delimiter$="")
  Protected result = #minScript_OKAY
  ;Extract all remaining symbols (at least until an error is detected!)
    While *this\parseCharPtr\c And *this\parseCharPtr\c <> ';' And result = #minScript_OKAY 
      result = minScriptCodeCompilerClass_PARSER_ExtractNextSymbol(*this, *module)
      If *this\numParseSymbols And *this\parseTree[*this\numParseSymbols-1]\symbol$ = delimiter$
        Break
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function finishes (or continues until a given delimiter is encountered) the process of creating a parse tree from the current line of source.
;We can only call this once for a given source string.
Procedure.i minScriptCodeCompilerClass_PARSER_FinishParseTree(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule, delimiter$="")
  Protected result = #minScript_OKAY, i, j
  ;Extract all remaining symbols (at least until an error is detected!)
    While *this\parseCharPtr\c And *this\parseCharPtr\c <> ';' And result = #minScript_OKAY 
      result = minScriptCodeCompilerClass_PARSER_ExtractNextSymbol(*this, *module)
      If *this\numParseSymbols And (delimiter$ And *this\parseTree[*this\numParseSymbols-1]\symbol$ = delimiter$)
        Break
      EndIf
    Wend
  ;If there were no errors, we scan all symbols which have not been categorised as yet in an attempt to categorise them.
    If result = #minScript_OKAY
      For i = 0 To *this\numParseSymbols - 1
        If *this\parseTree[i]\symbolType = 0
          ;Check first for a object method. This depends on the previous symbol being a '.'.
          If i And *this\parseTree[i-1]\symbol$ = "."
            result = minScript_SymbolsIsSymbolNameValid(@*this\parseTree[i]\symbol$)
            If result <> #minScript_OKAY
              result = #minScript_SYNTAXERROR
              Break
            EndIf
            ;If the following symbol is not an open bracket then we change the invokation to a property get (may be further changed to an object property get or a property set later).
            If i = *this\numParseSymbols - 1 Or *this\parseTree[i+1]\symbol$ <> "("
              *this\parseTree[i]\invokeType = #minScript_METHOD_GET
            Else
              *this\parseTree[i]\invokeType = #minScript_METHOD_INVOKE
            EndIf
            *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_OBJECTMETHOD
          ;Check next for a keyword.
          ElseIf minScript_IsCompilerKeyword(*this\parseTree[i]\symbol$) ;Keyword.
            *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_KEYWORD
          Else  ;We now either have a run-time function or a scripted function or a local variable.
                ;We can check for a function because the proceeding symbol should be an open parantheses.
                ;Let us check that the symbol name is valid. This will allow for a more appropriate error.
            result = minScript_SymbolsIsSymbolNameValid(@*this\parseTree[i]\symbol$)
            If result <> #minScript_OKAY
              result = #minScript_SYNTAXERROR
              Break
            EndIf
            If i < *this\numParseSymbols - 1 And *this\parseTree[i+1]\symbol$ = "("
              ;First check for a module:scriptFunction() call.
              ;If the previous symbol is a ':' then we are looking at a combination of 3 symbols of the form 'moduleName:functionName'.
                If i And *this\parseTree[i-1]\symbol$ = ":"
                  If i > 1 And *this\parseTree[i-2]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE
                    *this\parseTree[i-2]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
                    *this\parseTree[i-2]\symbol$ + ":" + *this\parseTree[i]\symbol$
                    *this\parseTree[i-2]\typeModifierVT = *this\parseTree[i]\typeModifierVT
                    *this\parseTree[i-2]\invokeType = *this\parseTree[i]\invokeType
                    ;Shunt everything down 2 elements.
                      For j = i - 1 To *this\numParseSymbols - 3
                        *this\parseTree[j]\symbolType = *this\parseTree[j+2]\symbolType
                        *this\parseTree[j]\localVariable = *this\parseTree[j+2]\localVariable
                        *this\parseTree[j]\symbol$ = *this\parseTree[j+2]\symbol$
                        *this\parseTree[j]\typeModifierVT = *this\parseTree[j+2]\typeModifierVT
                        *this\parseTree[j]\invokeType = *this\parseTree[j+2]\invokeType
                      Next
                      *this\numParseSymbols - 2
                      i - 2
                  Else
                    result = #minScript_SYNTAXERROR
                    Break
                  EndIf
                Else
                  ;Try a runtime function first.
                    CompilerIf #PB_Compiler_Thread
                      LockMutex(minScript_Globals\threadMutex)
                    CompilerEndIf
                    *this\parseTree[i]\runtimeFunction = FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), *this\parseTree[i]\symbol$)
                    CompilerIf #PB_Compiler_Thread
                      UnlockMutex(minScript_Globals\threadMutex)
                    CompilerEndIf
                    If *this\parseTree[i]\runtimeFunction 
                      *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION
                    Else ;Can only be a (possibly undeclared) scripted function.
                      *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
                      ;We can attempt to locate the script function in the given module. If unusuccessful, we simply assume that the function has not been declared yet.
                      *this\parseTree[i]\scriptFunction = FindMapElement(*module\functions(), *this\parseTree[i]\symbol$)
                    EndIf
                EndIf
            Else  ;Variable (possibly undeclared at this point).
                  ;The parser cannot throw an error at this point.
              *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE
              ;We can attempt to locate the variable. If unusuccessful it simply means that the variable has not been declared yet.
                If *this\code\currentScriptFunction
                  *this\parseTree[i]\localVariable = FindMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *this\parseTree[i]\symbol$)
                EndIf
                If *this\parseTree[i]\localVariable
                  ;Do we in fact have a static variable?
                    If *this\parseTree[i]\localVariable\staticModuleVar
                      *this\parseTree[i]\moduleVariable = *this\parseTree[i]\localVariable\staticModuleVar
                      *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
;                      *this\parseTree[i]\symbol$ = "s_" + *this\code\currentScriptFunction\name$ + ":" + *this\parseTree[i]\symbol$
                    EndIf
                Else
                  ;Try a module variable.
                    *this\parseTree[i]\moduleVariable = FindMapElement(*module\moduleVariables(), *this\parseTree[i]\symbol$)
                    If *this\parseTree[i]\moduleVariable
                      *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
                    EndIf
                EndIf
            EndIf
          EndIf
        EndIf
        ;Do we have a type modifier. If so we can perform some checks on the operand itself.
        If *this\parseTree[i]\typeModifierVT
          Select *this\parseTree[i]\symbolType
            Case #minScript_COMPILER_SYMBOL_KEYWORD, #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION, #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
              result = #minScript_INVALIDUSEOFTYPEMODIFIER
              Break             
            Case #minScript_COMPILER_SYMBOL_SEPARATOR
              ;The only separator we can allow is a close bracket, providing it is not followed by a '.' method invokation.
              If *this\parseTree[i]\symbol$ = ")"
                If i < *this\numParseSymbols - 1 And *this\parseTree[i+1]\symbol$ = "."
                  result = #minScript_INVALIDUSEOFTYPEMODIFIER
                  Break
                EndIf
              Else
                result = #minScript_INVALIDUSEOFTYPEMODIFIER
                Break
              EndIf
            Case #minScript_COMPILER_SYMBOL_OBJECTMETHOD
              ;The only permitted option here is for a modifier to apply to the final invokation in a series, part of which
              ;is covered by the above case.
                If i < *this\numParseSymbols - 1 And (*this\parseTree[i+1]\symbol$ = "." Or *this\parseTree[i+1]\symbol$ = "(")
                  result = #minScript_INVALIDUSEOFTYPEMODIFIER
                  Break
                EndIf
            Case #minScript_COMPILER_SYMBOL_LOCALVARIABLE
              ;Only predeclared non-object variables can take a modifier.
              If *this\parseTree[i]\localVariable = 0
                result = #minScript_VARIABLENOTFOUND
                Break
              EndIf
            Case #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE, #minScript_COMPILER_SYMBOL_MODULEVARIABLE
              ;Only predeclared non-object variables can take a modifier.
              If *this\parseTree[i]\scriptVariable\vt & #minScript_REFERENCE
;                result = #minScript_INVALIDDATATYPE
                Break
              EndIf
          EndSelect
        EndIf
      Next
    EndIf
;Debug *this\currentLine$
;For i = 0 To *this\numParseSymbols - 1
;  Debug "  " + Str(*this\parseTree[i]\symbolType) + " : " + *this\parseTree[i]\symbol$ ;+ " : " + Str(*this\parseTree[i]\typeModifierVT)
;Next
;Debug "========"
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function extracts the next symbol from the source stream.
Procedure.i minScriptCodeCompilerClass_PARSER_ExtractNextSymbol(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule)
  Protected result = #minScript_OKAY, blnCertainSymbolTypeExpected, blnDecimalPointFound, *ptr1.CHARACTER, blnProceedWithObjectDefaultInvoke
  Protected *ptrConstant.st_minScriptGlobalConstant, t1
  While *this\parseCharPtr\c And result = #minScript_OKAY
    If *this\numParseSymbols = #minScript_COMPILER_SIZEOFPARSETREE
      ;We cannot allow any more symbols. Since the source lines would have been trimmed, any remaining characters (whitespace or otherwise)
      ;means there are too many symbols.
      result = #minScript_STATEMENTTOOCOMPLEX
      Break
    EndIf
    *this\parseTree[*this\numParseSymbols]\symbolType = 0
    *this\parseTree[*this\numParseSymbols]\typeModifierVT = 0
    *this\parseTree[*this\numParseSymbols]\localVariable = 0
    *this\parseTree[*this\numParseSymbols]\invokeType = 0
    Select *this\parseCharPtr\c
      Case #TAB, ' ' ;Whitespace.
        *this\parseSymbol + SizeOf(CHARACTER)
        *this\parseCharPtr + SizeOf(CHARACTER)
      Case ';' ;Comment.
        Break
      Case '"' ;Literal string.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
          ;Let us retrieve the entire quoted string.
            *this\parseCharPtr + SizeOf(CHARACTER)
            While *this\parseCharPtr\c
              If *this\parseCharPtr\c = '"'
                With *this\parseTree[*this\numParseSymbols]
                  \symbol$ = PeekS(*this\parseSymbol, (*this\parseCharPtr - *this\parseSymbol) >> (SizeOf(CHARACTER)-1)+1)
                  \symbolType = #minScript_COMPILER_SYMBOL_LITERALSTRING
                EndWith
                *this\numParseSymbols + 1
                *this\parseSymbol = *this\parseCharPtr + SizeOf(CHARACTER)
                Break
              EndIf          
              *this\parseCharPtr + SizeOf(CHARACTER)
            Wend
          If *this\parseCharPtr\c <> '"'
            result = #minScript_MISSINGTERMINATINGSYMBOL
          Else
            *this\parseCharPtr = *this\parseSymbol ;Ready for the next symbol.
          EndIf
        EndIf
        Break
      Case '0' ;We check for a hexadecimal number of the form 0x... etc.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
          *ptr1 = *this\parseCharPtr + SizeOf(CHARACTER)
          If *ptr1\c = 0 Or (*ptr1\c <> 'x' And *ptr1\c <> 'X')
            Goto minScriptCodeCompilerClass_PARSER_ExtractNextSymbol_L2
          EndIf
        EndIf        
        ;We have what should be a hexadecimal number.
        *this\parseCharPtr = *ptr1 + SizeOf(CHARACTER)
        If *this\parseCharPtr\c = 0 ;
          result = #minScript_SYNTAXERROR
          Break
        EndIf
        *this\parseSymbol = *this\parseCharPtr
        While *this\parseCharPtr\c
          If *this\parseCharPtr\c = '.' Or *this\parseCharPtr\c = ';' Or *this\parseCharPtr\c = ' ' Or *this\parseCharPtr\c = 9 Or *this\parseCharPtr\c = '+' Or *this\parseCharPtr\c = '-' Or *this\parseCharPtr\c = '*' Or *this\parseCharPtr\c = '/' Or *this\parseCharPtr\c = '(' Or *this\parseCharPtr\c = ')' Or *this\parseCharPtr\c = ',' Or *this\parseCharPtr\c = '=' Or *this\parseCharPtr\c = '#' Or *this\parseCharPtr\c = '$' Or *this\parseCharPtr\c = '<' Or *this\parseCharPtr\c = '>' Or *this\parseCharPtr\c = ':' Or *this\parseCharPtr\c = '%' Or *this\parseCharPtr\c = '@' Or *this\parseCharPtr\c ='~' Or *this\parseCharPtr\c ='|' Or *this\parseCharPtr\c ='!' Or *this\parseCharPtr\c ='&' Or *this\parseCharPtr\c ='[' Or *this\parseCharPtr\c =']'
            Break
          EndIf          
          *this\parseCharPtr + SizeOf(CHARACTER)
        Wend
        *this\parseTree[*this\numParseSymbols]\symbol$ = PeekS(*this\parseSymbol, (*this\parseCharPtr - *this\parseSymbol) >> (SizeOf(CHARACTER)-1))
        ;Check this is indeed a valid hex number and convert to an integer.
          If minScript_SymbolsIsHexadecimal(@*this\parseTree[*this\numParseSymbols]\symbol$, @t1) = #minScript_OKAY
            *this\parseTree[*this\numParseSymbols]\symbol$ = Str(t1)
            *this\parseTree[*this\numParseSymbols]\symbolType = #minScript_COMPILER_SYMBOL_LITERALNUMBER
            *this\numParseSymbols + 1
            *this\parseSymbol = *this\parseCharPtr
          Else
            result = #minScript_SYNTAXERROR
          EndIf
        Break
      Case '0' To '9', '.' ;Literal number. We retrieve this in a separate step to avoid problems with the '.' object invokation.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
minScriptCodeCompilerClass_PARSER_ExtractNextSymbol_L2:
          blnDecimalPointFound = #False
          ;In the case that the first character is a '.', we need to determine whether we have the start of a decimal or a method invokation.
            If *this\parseCharPtr\c = '.'
              *ptr1 = *this\parseCharPtr + SizeOf(CHARACTER)
              If *ptr1\c = 0 ;
                result = #minScript_SYNTAXERROR
                Break
              ElseIf *ptr1\c < '0' Or *ptr1\c > '9'              
                Goto minScriptCodeCompilerClass_PARSER_ExtractNextSymbol_L1
              EndIf
              blnDecimalPointFound = #True
            EndIf
          ;Let us retrieve the entire 'number'.
            *this\parseCharPtr + SizeOf(CHARACTER)
            While *this\parseCharPtr\c
              If *this\parseCharPtr\c = '.'
                If blnDecimalPointFound = #True
                  result = #minScript_SYNTAXERROR
                  Break 2
                EndIf
                blnDecimalPointFound = #True
              ElseIf *this\parseCharPtr\c = ';' Or *this\parseCharPtr\c = ' ' Or *this\parseCharPtr\c = 9 Or *this\parseCharPtr\c = '+' Or *this\parseCharPtr\c = '-' Or *this\parseCharPtr\c = '*' Or *this\parseCharPtr\c = '/' Or *this\parseCharPtr\c = '(' Or *this\parseCharPtr\c = ')' Or *this\parseCharPtr\c = ',' Or *this\parseCharPtr\c = '=' Or *this\parseCharPtr\c = '#' Or *this\parseCharPtr\c = '$' Or *this\parseCharPtr\c = '<' Or *this\parseCharPtr\c = '>' Or *this\parseCharPtr\c = ':' Or *this\parseCharPtr\c = '%' Or *this\parseCharPtr\c = '@' Or *this\parseCharPtr\c ='~' Or *this\parseCharPtr\c ='|' Or *this\parseCharPtr\c ='!' Or *this\parseCharPtr\c ='&' Or *this\parseCharPtr\c ='[' Or *this\parseCharPtr\c =']'
                Break
              EndIf          
              *this\parseCharPtr + SizeOf(CHARACTER)
            Wend
            *this\parseTree[*this\numParseSymbols]\symbol$ = PeekS(*this\parseSymbol, (*this\parseCharPtr - *this\parseSymbol) >> (SizeOf(CHARACTER)-1))
          ;Check this is indeed a valid number.
            If minScript_SymbolsIsReal(@*this\parseTree[*this\numParseSymbols]\symbol$) = #minScript_OKAY
              *this\parseTree[*this\numParseSymbols]\symbolType = #minScript_COMPILER_SYMBOL_LITERALNUMBER
              *this\numParseSymbols + 1
              *this\parseSymbol = *this\parseCharPtr
            Else
              result = #minScript_SYNTAXERROR
            EndIf
        EndIf
        Break
      Case '#' ;(Global) constant.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
          blnCertainSymbolTypeExpected = #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
          *this\parseSymbol = *this\parseCharPtr + SizeOf(CHARACTER)
          *this\parseCharPtr = *this\parseSymbol ;Ready for the next symbol.
        EndIf
      Case '$' ;Script variable.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
          blnCertainSymbolTypeExpected = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
          *this\parseSymbol = *this\parseCharPtr + SizeOf(CHARACTER)
          *this\parseCharPtr = *this\parseSymbol ;Ready for the next symbol.
        EndIf
      Case '%' ;Type modifier.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
          If *this\numParseSymbols = 0
            result = #minScript_SYNTAXERROR
          Else
            ;Identify the modifier.
              *this\parseCharPtr + SizeOf(CHARACTER)
              Select *this\parseCharPtr\c
                Case 'r', 'R'
                  *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = #minScript_REAL
                Case 'c', 'C'
                  *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = #minScript_CURRENCY
                Case 'i', 'I'
                  *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = #minScript_INTEGER
                Case 's', 'S'
                  *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = #minScript_STRING
                Default          
                  result = #minScript_INVALIDTYPEMODIFIER
                  Break
              EndSelect
            *this\parseSymbol = *this\parseCharPtr + SizeOf(CHARACTER)
            *this\parseCharPtr = *this\parseSymbol ;Ready for the next symbol.
          EndIf
        EndIf
        Break
      Case '(' ;We separate this case in order to deal with object variable default method invokation.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
          ;We check the previous symbol, if any, to determine if it is an object variable.
          ;We also check the symbol previous to any object variable is not a '.'
            If *this\numParseSymbols 
              blnProceedWithObjectDefaultInvoke = #False
              ;First check for a local variable.
                If *this\code\currentScriptFunction And FindMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *this\parseTree[*this\numParseSymbols-1]\symbol$)
                  If *this\code\currentScriptFunction\localVariablesSymbols()\vt & #minScript_REFERENCE
                    blnProceedWithObjectDefaultInvoke = #True
                  EndIf
                ;Check for a script or module object variable.
                ElseIf *this\parseTree[*this\numParseSymbols-1]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE And *this\parseTree[*this\numParseSymbols-1]\scriptVariable\vt & #minScript_REFERENCE
                  blnProceedWithObjectDefaultInvoke = #True         
                ElseIf *module And FindMapElement(*module\moduleVariables(), *this\parseTree[*this\numParseSymbols-1]\symbol$)
                  If *module\moduleVariables()\vt & #minScript_REFERENCE
                    blnProceedWithObjectDefaultInvoke = #True
                  EndIf
                ElseIf *this\parseTree[*this\numParseSymbols-1]\symbol$ = ")"
                  blnProceedWithObjectDefaultInvoke = #True         
                EndIf
              If blnProceedWithObjectDefaultInvoke
                If *this\numParseSymbols < 2 Or *this\parseTree[*this\numParseSymbols-2]\symbol$ <> "."
                  If *this\numParseSymbols < #minScript_COMPILER_SIZEOFPARSETREE - 2
                    ;We add entries directly to the tree to invoke a default method.
                    With *this\parseTree[*this\numParseSymbols]
                      \symbol$ = "."
                      \symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                      \typeModifierVT = 0
                      \invokeType = 0
                    EndWith
                    *this\numParseSymbols + 1
                    With *this\parseTree[*this\numParseSymbols]
                      \symbol$ = #minScript_DEFAULTPROPERTY$
                      \symbolType = #minScript_COMPILER_SYMBOL_OBJECTMETHOD
                      \typeModifierVT = 0
                      \invokeType = #minScript_METHOD_GET
                    EndWith
                    *this\numParseSymbols + 1
                  Else
                    result = #minScript_STATEMENTTOOCOMPLEX
                    Break
                  EndIf
                EndIf
              EndIf
            EndIf
          Goto minScriptCodeCompilerClass_PARSER_ExtractNextSymbol_L1
        EndIf
      Case '+', '-', '*', '/', '(', ')', ',', '=', '<', '>', ':', '.', '%', '@', '~', '|', '!', '&', '[', ']' ;Separator.
        If blnCertainSymbolTypeExpected
          result = #minScript_SYNTAXERROR
        Else
minScriptCodeCompilerClass_PARSER_ExtractNextSymbol_L1:
          With *this\parseTree[*this\numParseSymbols]
            \symbol$ = Chr(*this\parseCharPtr\c)
            \symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
            \typeModifierVT = 0
            \invokeType = 0
          EndWith
          ;We may have one of the << or >> bitshift operators?
          If *this\parseCharPtr\c = '<'
            *ptr1 = *this\parseCharPtr + SizeOf(CHARACTER)
            If *ptr1\c = '<'
              *this\parseTree[*this\numParseSymbols]\symbol$ = "<<"
              *this\parseCharPtr = *ptr1
            EndIf          
          ElseIf *this\parseCharPtr\c = '>'
            *ptr1 = *this\parseCharPtr + SizeOf(CHARACTER)
            If *ptr1\c = '>'
              *this\parseTree[*this\numParseSymbols]\symbol$ = ">>"
              *this\parseCharPtr = *ptr1
            EndIf          
          EndIf
          *this\numParseSymbols + 1
          *this\parseSymbol = *this\parseCharPtr + SizeOf(CHARACTER)
          *this\parseCharPtr = *this\parseSymbol ;Ready for the next symbol.
        EndIf
        Break
      Default ;A symbol which needs checking for keywords etc.
        ;First we retrieve the symbol.        
          While *this\parseCharPtr\c
            If *this\parseCharPtr\c = ';' Or *this\parseCharPtr\c = ' ' Or *this\parseCharPtr\c = 9 Or *this\parseCharPtr\c = '+' Or *this\parseCharPtr\c = '-' Or *this\parseCharPtr\c = '*' Or *this\parseCharPtr\c = '/' Or *this\parseCharPtr\c = '(' Or *this\parseCharPtr\c = ')' Or *this\parseCharPtr\c = ',' Or *this\parseCharPtr\c = '=' Or *this\parseCharPtr\c = '#' Or *this\parseCharPtr\c = '$' Or *this\parseCharPtr\c = '<' Or *this\parseCharPtr\c = '>' Or *this\parseCharPtr\c = ':' Or *this\parseCharPtr\c = '.' Or *this\parseCharPtr\c = '%' Or *this\parseCharPtr\c = '@' Or *this\parseCharPtr\c = '~' Or *this\parseCharPtr\c ='|' Or *this\parseCharPtr\c ='!' Or *this\parseCharPtr\c ='&' Or *this\parseCharPtr\c ='[' Or *this\parseCharPtr\c =']'
              Break
            EndIf        
            *this\parseCharPtr + SizeOf(CHARACTER)
          Wend
        ;Is a global constant expected?
          If blnCertainSymbolTypeExpected = #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
            *this\parseTree[*this\numParseSymbols]\symbol$ = "#" + LCase(PeekS(*this\parseSymbol, (*this\parseCharPtr - *this\parseSymbol) >> (SizeOf(CHARACTER)-1)))
            *this\parseTree[*this\numParseSymbols]\symbolType = #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
            ;Locate the constant.
              ;First try a module level constant as these take precedence over global constants.
              *this\parseTree[*this\numParseSymbols]\constant = FindMapElement(*module\moduleConstants(), *this\parseTree[*this\numParseSymbols]\symbol$)
              If *this\parseTree[*this\numParseSymbols]\constant = 0
                ;Try a global constant.
                CompilerIf #PB_Compiler_Thread
                  LockMutex(minScript_Globals\threadMutex)
                CompilerEndIf
                *this\parseTree[*this\numParseSymbols]\constant = FindMapElement(minScript_Globals\minScriptGlobalConstants(), *this\parseTree[*this\numParseSymbols]\symbol$)
                CompilerIf #PB_Compiler_Thread
                  UnlockMutex(minScript_Globals\threadMutex)
                CompilerEndIf
              EndIf
            blnCertainSymbolTypeExpected = #False
          ;How about a script variable instead?
          ElseIf blnCertainSymbolTypeExpected = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
            *this\parseTree[*this\numParseSymbols]\symbol$ = "$" + LCase(PeekS(*this\parseSymbol, (*this\parseCharPtr - *this\parseSymbol) >> (SizeOf(CHARACTER)-1)))
            *this\parseTree[*this\numParseSymbols]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
            ;Locate the script variable. It may not be defined at this point.
              *this\parseTree[*this\numParseSymbols]\scriptVariable = FindMapElement(*this\code\scriptVariables(), *this\parseTree[*this\numParseSymbols]\symbol$)
            blnCertainSymbolTypeExpected = #False
          Else
            *this\parseTree[*this\numParseSymbols]\symbol$ = LCase(PeekS(*this\parseSymbol, (*this\parseCharPtr - *this\parseSymbol) >> (SizeOf(CHARACTER)-1)))
          EndIf
        *this\numParseSymbols+1
        *this\parseSymbol = *this\parseCharPtr
        Break
    EndSelect
  Wend
  If blnCertainSymbolTypeExpected
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following retrieves the next line of source code, ignoring any empty lines and any preceeding white-space, including tabs.
;Returns #True if a non-empty, non-commented line is found.
Procedure.i minScriptCodeCompilerClass_PARSER_GetNextSourceLine(*this.st_minScriptCodeCompilerClass)
  Protected result, basePos
  *this\currentLine$ = ""
  basePos = *this\ptrSource
  While *this\ptrSource\c And result = #False
    Select *this\ptrSource\c
      Case #TAB, 32
        ;Ignore preceeding white-space.
        If *this\ptrSource = basePos  ;A possible non-empty, non-commented source line.
          *this\ptrSource + SizeOf(CHARACTER)
          basePos = *this\ptrSource
        Else
          *this\ptrSource + SizeOf(CHARACTER)
        EndIf
      Case #LF
        *this\currentLineNumber + 1
        If *this\ptrSource <> basePos  ;A possible non-empty, non-commented source line.
          *this\currentLine$ = Trim(PeekS(basePos, (*this\ptrSource - basePos) >> (SizeOf(CHARACTER) - 1)))
          *this\ptrSource + SizeOf(CHARACTER)
          If *this\ptrSource\c = #CR
            *this\ptrSource + SizeOf(CHARACTER)
          EndIf
          If *this\currentLine$ And Left(*this\currentLine$, 1) <> ";"
            result = #True
          EndIf
        Else
          *this\ptrSource + SizeOf(CHARACTER)
          If *this\ptrSource\c = #CR
            *this\ptrSource + SizeOf(CHARACTER)
          EndIf
        EndIf
        basePos = *this\ptrSource
      Case #CR
        *this\currentLineNumber + 1
        If *this\ptrSource <> basePos  ;A possible non-empty, non-commented source line.
          *this\currentLine$ = Trim(PeekS(basePos, (*this\ptrSource - basePos) >> (SizeOf(CHARACTER) - 1)))
          *this\ptrSource + SizeOf(CHARACTER)
          If *this\ptrSource\c = #LF
            *this\ptrSource + SizeOf(CHARACTER)
          EndIf
          If *this\currentLine$ And Left(*this\currentLine$, 1) <> ";"
            result = #True
          EndIf
        Else
          *this\ptrSource + SizeOf(CHARACTER)
          If *this\ptrSource\c = #LF
            *this\ptrSource + SizeOf(CHARACTER)
          EndIf
        EndIf
        basePos = *this\ptrSource
      Default
        *this\ptrSource + SizeOf(CHARACTER)
        If *this\ptrSource\c = 0 ;End of stream.
          *this\currentLineNumber + 1
          *this\currentLine$ = Trim(PeekS(basePos, (*this\ptrSource - basePos) >> (SizeOf(CHARACTER) - 1)))
          If *this\currentLine$ And Left(*this\currentLine$, 1) <> ";"
            result = #True
          EndIf
        EndIf        
    EndSelect
  Wend
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-HELPER FUNCTIONS.

;////////////////////////////////////////////////////////////////////////
;The following is called, only in the case that DISABLEEXPLICIT is in force, to add a new local variable at the given point in the parse tree and vType
;*this\code\currentScriptFunction is guaranteed not to be null.
Procedure.i minScriptCodeCompilerClass_HELPER_AddLocalVariable(*this.st_minScriptCodeCompilerClass, *ptrParseTreeEntry.st_minScript_ParseTree, varType)
  Protected result = #minScript_OKAY
  ;In the case of a user-defined (or variant) type, we cannot allow the variable to have the same name as run-time function because such variables
  ;can be used to invoke certain 'methods'.
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
    If (varType & #minScript_REFERENCE = 0 And varType <> #minScript_VARIANT) Or FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), *ptrParseTreeEntry\symbol$) = 0
      ;Create a new local variable symbol.
      *ptrParseTreeEntry\localVariable = AddMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *ptrParseTreeEntry\symbol$)
      If *ptrParseTreeEntry\localVariable 
        *ptrParseTreeEntry\localVariable\EBPoffset = *this\code\currentScriptFunction\numLocalVariables * SizeOf(minScriptVariant)
        *ptrParseTreeEntry\localVariable\vt = varType
        *this\code\currentScriptFunction\numLocalVariables + 1
        ;Emit the appropriate object code.
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_LOCALVAR, varType)
      Else
        result = #minScript_OUTOFMEMORY                  
      EndIf
    Else
      result = #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
    EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following is called, only in the case that DISABLEEXPLICIT is in force and we are not in the GCL, to add a new script variable at the given point in the parse tree and vType
Procedure.i minScriptCodeCompilerClass_HELPER_AddScriptVariable(*this.st_minScriptCodeCompilerClass, *ptrParseTreeEntry.st_minScript_ParseTree, varType)
  Protected result = #minScript_OKAY
  *ptrParseTreeEntry\scriptVariable = AddMapElement(*this\code\scriptVariables(), *ptrParseTreeEntry\symbol$)
  If *ptrParseTreeEntry\scriptVariable 
    *ptrParseTreeEntry\scriptVariable\vt = varType
  Else
    result = #minScript_OUTOFMEMORY                  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



;-EXPRESSION FUNCTIONS.

;////////////////////////////////////////////////////////////////////////
;The following adds the given symbol to the underlying function's string table.
;Used to add literal strings and scripted function names when being called in script.
;If successful, the index of the table entry is placed into the given buffer.
Procedure.i minScriptCodeCompilerClass_EXPRESSION_AddSymbolToStringTable(*this.st_minScriptCodeCompilerClass, symbol$, *ptrIndex.INTEGER)
  Protected result = #minScript_OKAY, i, blnExistingStringFound
  ;We first search the string table to see if there is already a copy of the given string present.
    For i = 1 To *this\code\currentScriptFunction\numStrings
      If *this\code\currentScriptFunction\stringTable[i] = symbol$
        *ptrIndex\i = i
        blnExistingStringFound = #True
        Break
      EndIf
    Next
  ;If unsuccessful, we attempt to add a new entry.
    If blnExistingStringFound = #False
      If *this\code\currentScriptFunction\numStrings < #minScript_STRINGTABLESIZE
        *this\code\currentScriptFunction\numStrings + 1
        *this\code\currentScriptFunction\stringTable[*this\code\currentScriptFunction\numStrings] = symbol$
        *ptrIndex\i = *this\code\currentScriptFunction\numStrings      
      Else
        result = #minScript_OUTOFMEMORY
      EndIf
    EndIf
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following is the main expression compiler.
;expressionVT is used when we know the variant type of the expression result. This means we can promote certain operands depending on
;the individual operators (e.g. division) etc. If expressionVT = 0 then we do not need to worry about such promotions.
Procedure.i minScriptCodeCompilerClass_EXPRESSION_Compile(*this.st_minScriptCodeCompilerClass, expressionVT=0, delimiter$="")
  Protected result = #minScript_OKAY, i
  ;In the case that disableexplicit is in force, we run through the parse tree and attempt to add any local/script variables as appropriate.
  ;We have to do this up front (at least in the case of local variables) so as to emit the appropriate object code before the expression is itself compiled to object code.
  If *this\code\blnEnableExplicit = #False
    For i = *this\expressionSymbolIndex To *this\numParseSymbols - 1
      If delimiter$ And *this\parseTree[*this\expressionSymbolIndex]\symbol$ = delimiter$
        Break
      ElseIf *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE And *this\parseTree[i]\localVariable = 0 And *this\code\currentScriptFunction
        ;Attempt to locate what may have been a newly added local variable.
          *this\parseTree[i]\localVariable = FindMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *this\parseTree[i]\symbol$)
        If *this\parseTree[i]\localVariable = 0
          result = minScriptCodeCompilerClass_HELPER_AddLocalVariable(*this, *this\parseTree[i], #minScript_VARIANT)
        EndIf
      ElseIf *this\parseTree[i]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE And *this\parseTree[i]\scriptVariable = 0
        If *this\code <> minScriptGlobalClassLib
          result = minScriptCodeCompilerClass_HELPER_AddScriptVariable(*this, @*this\parseTree[i], #minScript_VARIANT)
        Else
          result = #minScript_ACCESSDENIED              
        EndIf
      EndIf
    Next
    If result = #minScript_OKAY
      ;Hand over to the recursive element of the expression engine.
        result = minScriptCodeCompilerClass_EXPRESSION_Recurse(*this, expressionVT)
    EndIf
  Else
  ;Hand over to the recursive element of the expression engine.
    result = minScriptCodeCompilerClass_EXPRESSION_Recurse(*this, expressionVT)
  EndIf
  If result = #minScript_OKAY 
    ;Perform additional error checking.
    ;We either exhausted the input stream or reached the delimiting keyword (if one is defined).
      If delimiter$ = ""
        ;Should have reached the end of the input stream.
        If *this\expressionSymbolIndex < *this\numParseSymbols
          result = #minScript_SYNTAXERROR
        EndIf
      ElseIf *this\expressionSymbolIndex >= *this\numParseSymbols Or *this\parseTree[*this\expressionSymbolIndex]\symbol$ <> delimiter$
        result = #minScript_SYNTAXERROR
      EndIf      
  EndIf
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_AddSubTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded, blnIsNegative
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      blnIsNegative = #False
      result = minScriptCodeCompilerClass_EXPRESSION_MulDivTerms(*this, expressionVT, @blnIsNegative)
      If result = #minScript_OKAY
        If blnOperatorRecorded = 1 ;Add.
          If blnIsNegative
            *this\thisEX\EmitObjectCodeInstruction(#MS_OP_SUB)
          Else
            *this\thisEX\EmitObjectCodeInstruction(#MS_OP_ADD)
          EndIf
        ElseIf blnOperatorRecorded = 2 ;Subtract.
          If blnIsNegative
            *this\thisEX\EmitObjectCodeInstruction(#MS_OP_ADD)
          Else          
            *this\thisEX\EmitObjectCodeInstruction(#MS_OP_SUB)
          EndIf
        ElseIf blnIsNegative
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_NEG)
        EndIf        
        blnOperatorRecorded = 0
        ;Examine the next symbol to see if it is a + or -.
        If *this\expressionSymbolIndex < *this\numParseSymbols
          If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "+"
            blnOperatorRecorded = 1 ;For add.
            *this\expressionSymbolIndex + 1
          ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "-"
            blnOperatorRecorded = 2 ;For subtract.
            *this\expressionSymbolIndex + 1
          Else
            Break
          EndIf
        EndIf
        If *this\expressionSymbolIndex >= *this\numParseSymbols
          If blnOperatorRecorded
            result = #minScript_SYNTAXERROR
          EndIf
          Break
        EndIf
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_AndTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      result = minScriptCodeCompilerClass_EXPRESSION_ComparisonTerms(*this, expressionVT)
      If result = #minScript_OKAY
        If blnOperatorRecorded
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_AND)
        EndIf
        ;Now look out for the next appropriate symbol.
        If *this\expressionSymbolIndex >= *this\numParseSymbols Or *this\parseTree[*this\expressionSymbolIndex]\symbol$ <> "and"
          Break
        EndIf                    
        blnOperatorRecorded = #True
        *this\expressionSymbolIndex + 1
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_BitwiseAndTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      result = minScriptCodeCompilerClass_EXPRESSION_BitwiseShiftTerms(*this, expressionVT)
      If result = #minScript_OKAY
        If blnOperatorRecorded
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_BITWISEAND)
        EndIf
        ;Now look out for the next appropriate symbol.
        If *this\expressionSymbolIndex >= *this\numParseSymbols Or *this\parseTree[*this\expressionSymbolIndex]\symbol$ <> "&"
          Break
        EndIf                    
        blnOperatorRecorded = #True
        *this\expressionSymbolIndex + 1
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_BitwiseORXORTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      result = minScriptCodeCompilerClass_EXPRESSION_BitwiseAndTerms(*this, expressionVT)
      If result = #minScript_OKAY
        If blnOperatorRecorded = 1 ;| (bitwise OR)
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_BITWISEOR)
        ElseIf blnOperatorRecorded = 2 ;! (bitwise XOR)
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_BITWISEXOR)
        EndIf        
        blnOperatorRecorded = 0
        ;Examine the next symbol to see if it is a OR or XOR.
        If *this\expressionSymbolIndex < *this\numParseSymbols
          If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "|"
            blnOperatorRecorded = 1 ;For bitwise OR.
            *this\expressionSymbolIndex + 1
          ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "!"
            blnOperatorRecorded = 2 ;For bitwise XOR.
            *this\expressionSymbolIndex + 1
          Else
            Break
          EndIf
        EndIf
        If *this\expressionSymbolIndex >= *this\numParseSymbols
          If blnOperatorRecorded
            result = #minScript_SYNTAXERROR
          EndIf
          Break
        EndIf
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_BitwiseShiftTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      result = minScriptCodeCompilerClass_EXPRESSION_AddSubTerms(*this, expressionVT)
      If result = #minScript_OKAY
        If blnOperatorRecorded = 1 ;| (bitwise <<)
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_BITWISESHIFTLEFT)
        ElseIf blnOperatorRecorded = 2 ;! (bitwise >>)
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_BITWISESHIFTRIGHT)
        EndIf        
        blnOperatorRecorded = 0
        ;Examine the next symbol to see if it is a << or >>.
        If *this\expressionSymbolIndex < *this\numParseSymbols
          If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "<<"
            blnOperatorRecorded = 1 ;For bitwise <<.
            *this\expressionSymbolIndex + 1
          ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ">>"
            blnOperatorRecorded = 2 ;For bitwise >>.
            *this\expressionSymbolIndex + 1
          Else
            Break
          EndIf
        EndIf
        If *this\expressionSymbolIndex >= *this\numParseSymbols
          If blnOperatorRecorded
            result = #minScript_SYNTAXERROR
          EndIf
          Break
        EndIf
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_ComparisonTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      result = minScriptCodeCompilerClass_EXPRESSION_BitwiseORXORTerms(*this, expressionVT)
      If result = #minScript_OKAY
        If blnOperatorRecorded = #MS_OP_EQU
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_EQU)
        ElseIf blnOperatorRecorded = #MS_OP_LT
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_LT)
        ElseIf blnOperatorRecorded = #MS_OP_LTE
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_LTE)
        ElseIf blnOperatorRecorded = #MS_OP_GT
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_GT)
        ElseIf blnOperatorRecorded = #MS_OP_GTE
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_GTE)
        ElseIf blnOperatorRecorded = #MS_OP_NEQ
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_NEQ)
        EndIf
      EndIf
      blnOperatorRecorded = 0
      ;Examine the next symbol to see if it is a < or <= etc.
      If *this\expressionSymbolIndex < *this\numParseSymbols
        Select *this\parseTree[*this\expressionSymbolIndex]\symbol$
          Case "="
            blnOperatorRecorded = #MS_OP_EQU
            *this\expressionSymbolIndex + 1
          Case "<"
            *this\expressionSymbolIndex + 1
            If *this\expressionSymbolIndex < *this\numParseSymbols
              If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ">"
                blnOperatorRecorded = #MS_OP_NEQ
                *this\expressionSymbolIndex + 1
              ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "="
                blnOperatorRecorded = #MS_OP_LTE
                *this\expressionSymbolIndex + 1
              Else
                blnOperatorRecorded = #MS_OP_LT
              EndIf
            Else
              result = #minScript_SYNTAXERROR
            EndIf
          Case ">"
            *this\expressionSymbolIndex + 1
            If *this\expressionSymbolIndex < *this\numParseSymbols
              If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "="
                blnOperatorRecorded = #MS_OP_GTE
                *this\expressionSymbolIndex + 1
              Else
                blnOperatorRecorded = #MS_OP_GT
              EndIf
            Else
              result = #minScript_SYNTAXERROR
            EndIf

          Default
            Break
        EndSelect
      EndIf
      If result = #minScript_OKAY
        If *this\expressionSymbolIndex >= *this\numParseSymbols
          If blnOperatorRecorded
            result = #minScript_SYNTAXERROR
          EndIf
          Break
        EndIf
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following covers string concatenation as well.
Procedure.i minScriptCodeCompilerClass_EXPRESSION_MulDivTerms(*this.st_minScriptCodeCompilerClass, expressionVT, *ptrIsNegative.INTEGER)
  Protected result = #minScript_OKAY, blnOperatorRecorded, blnIsNegative
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      blnIsNegative = #False
      result = minScriptCodeCompilerClass_EXPRESSION_PushNextOperand(*this, expressionVT, @blnIsNegative)
      If result = #minScript_OKAY
        If blnIsNegative
          *ptrIsNegative\i = #True - *ptrIsNegative\i
        EndIf
        If blnOperatorRecorded = 1 ;Multiply.
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_MUL)
        ElseIf blnOperatorRecorded = 2 ;Divide.
          expressionVT&#minScript_VARIANTMASK
          ;We need to examine the expressionVT to see if one of the operands needs promoting. This avoids things like integer divisions
          ;being utilised when the result of the underlying expression will be cast to a real value.
          If expressionVT = #minScript_REAL Or expressionVT = #minScript_FLOAT Or expressionVT = #minScript_CURRENCY
            *this\thisEX\EmitObjectCodeInstruction(#MS_OP_CAST, #minScript_REAL)
          EndIf
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_DIV)
        EndIf        
        blnOperatorRecorded = 0
        ;Examine the next symbol to see if it is a * or /.
        If *this\expressionSymbolIndex < *this\numParseSymbols
          If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "*"
            blnOperatorRecorded = 1 ;For multiply.
            *this\expressionSymbolIndex + 1
          ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "/"
            blnOperatorRecorded = 2 ;For divide.
            *this\expressionSymbolIndex + 1
          Else
            Break
          EndIf
        EndIf
        If *this\expressionSymbolIndex >= *this\numParseSymbols
          If blnOperatorRecorded
            result = #minScript_SYNTAXERROR
          EndIf
          Break
        EndIf
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerClass_EXPRESSION_ORXORTerms(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY, blnOperatorRecorded
  ;Pull out consecutive operands.
    While result = #minScript_OKAY
      result = minScriptCodeCompilerClass_EXPRESSION_AndTerms(*this, expressionVT)
      If result = #minScript_OKAY
        If blnOperatorRecorded = 1 ;OR.
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_OR)
        ElseIf blnOperatorRecorded = 2 ;XOR.
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_XOR)
        EndIf        
        blnOperatorRecorded = 0
        ;Examine the next symbol to see if it is a OR or XOR.
        If *this\expressionSymbolIndex < *this\numParseSymbols
          If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "or"
            blnOperatorRecorded = 1 ;For OR.
            *this\expressionSymbolIndex + 1
          ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "xor"
            blnOperatorRecorded = 2 ;For XOR.
            *this\expressionSymbolIndex + 1
          Else
            Break
          EndIf
        EndIf
        If *this\expressionSymbolIndex >= *this\numParseSymbols
          If blnOperatorRecorded
            result = #minScript_SYNTAXERROR
          EndIf
          Break
        EndIf
      EndIf
    Wend
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following function is the lowest level of the recursive descent expression parser and pushes individual operands.
;We do not perform expression folding with this engine, though we do look out for consecutive +/-'s unitary operators.
Procedure.i minScriptCodeCompilerClass_EXPRESSION_PushNextOperand(*this.st_minScriptCodeCompilerClass, expressionVT, *ptrIsNegative.INTEGER)
  Protected result = #minScript_OKAY, var.minScriptVariant, index, fnName$, *runtimeFN.st_minScriptRuntimeFunction, numParamsPassed, i
  Protected symbolType, methodName$, blnIsVarReference, blnIsAddressOf, blnBracketsEncountered.q, invokeType, blnIsNot
  ;We may have a combination of + or -'s or '~'
    While *this\expressionSymbolIndex < *this\numParseSymbols
      If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "-" And *this\parseTree[*this\expressionSymbolIndex]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
        *ptrIsNegative\i = #True - *ptrIsNegative\i
        *this\expressionSymbolIndex + 1
      ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "+" And *this\parseTree[*this\expressionSymbolIndex]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
        *this\expressionSymbolIndex + 1
      ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "~" And *this\parseTree[*this\expressionSymbolIndex]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
        blnIsNot = #True - blnIsNot
        *this\expressionSymbolIndex + 1
      Else
        Break
      EndIf      
    Wend
  ;Check for an @ operator.  
    If (*this\expressionSymbolIndex < *this\numParseSymbols) And (*this\parseTree[*this\expressionSymbolIndex]\symbol$ = "@")
      blnIsAddressOf = #True
      *this\expressionSymbolIndex + 1
    EndIf
  If *this\expressionSymbolIndex < *this\numParseSymbols
    ;Examine the operand.
    symbolType = *this\parseTree[*this\expressionSymbolIndex]\symbolType
    Select symbolType
      Case #minScript_COMPILER_SYMBOL_LOCALVARIABLE
        If *this\parseTree[*this\expressionSymbolIndex]\localVariable
          ;Are we dealing with the address of the variable?
          If blnIsAddressOf
            ;We can only emit instructions to push the variable address if the variable is not a part of a string of method calls etc.
              If (*this\expressionSymbolIndex = *this\numParseSymbols - 1) Or *this\parseTree[*this\expressionSymbolIndex+1]\symbol$ <> "."
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHADDRLOCALVAR, *this\parseTree[*this\expressionSymbolIndex]\localVariable\EBPoffset)
                blnIsAddressOf = #False   
              Else
                ;Emit code to push a copy of the relevant variable.            
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHLOCALVAR, *this\parseTree[*this\expressionSymbolIndex]\localVariable\EBPoffset)
              EndIf
          Else
            ;Emit code to push a copy of the relevant variable.            
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHLOCALVAR, *this\parseTree[*this\expressionSymbolIndex]\localVariable\EBPoffset)
          EndIf
          ;Record whether we have an object variable (we have to include variants).
            blnIsVarReference = *this\parseTree[*this\expressionSymbolIndex]\localVariable\vt & (#minScript_REFERENCE|#minScript_VARIANT)
          *this\expressionSymbolIndex + 1
        Else
          result = #minScript_VARIABLENOTFOUND
        EndIf
      Case #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE, #minScript_COMPILER_SYMBOL_MODULEVARIABLE
        If *this\parseTree[*this\expressionSymbolIndex]\scriptVariable
          ;Are we dealing with the address of the variable?
          If blnIsAddressOf
            ;We can only emit instructions to push the variable address if the variable is not a part of a string of method calls etc.
              If (*this\expressionSymbolIndex = *this\numParseSymbols - 1) Or *this\parseTree[*this\expressionSymbolIndex+1]\symbol$ <> "."
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHADDRSCRIPTVAR, *this\parseTree[*this\expressionSymbolIndex]\scriptVariable)
                blnIsAddressOf = #False   
              Else
                ;Emit code to push a copy of the relevant variable.            
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHSCRIPTVAR, *this\parseTree[*this\expressionSymbolIndex]\scriptVariable)
              EndIf
          Else
            ;Emit code to push a copy of the relevant variable.            
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHSCRIPTVAR, *this\parseTree[*this\expressionSymbolIndex]\scriptVariable)
          EndIf
          ;Record whether we have an object variable (we have to include variants).
            blnIsVarReference = *this\parseTree[*this\expressionSymbolIndex]\scriptVariable\vt & (#minScript_REFERENCE|#minScript_VARIANT)
          *this\expressionSymbolIndex + 1
        Else
          result = #minScript_VARIABLENOTFOUND
        EndIf
      Case #minScript_COMPILER_SYMBOL_LITERALNUMBER
        ;We need to determine the type of literal number we have here.
        ;For this we can use the minScript_VarUtils_InitNumericVariant() function.
        result = minScript_VarUtils_InitNumericVariant(var, @*this\parseTree[*this\expressionSymbolIndex]\symbol$)
        If result = #minScript_OKAY
          ;Emit the appropriate assembly instruction.
          Select var\vt
            Case #minScript_INTEGER
              If *ptrIsNegative\i
                var\i = -var\i
                *ptrIsNegative\i = #False
              EndIf
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHINT, var\i)
            Case #minScript_CURRENCY
              If *ptrIsNegative\i
                var\c = -var\c
                *ptrIsNegative\i = #False
              EndIf
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHCUR, var\c)
            Case #minScript_REAL
              If *ptrIsNegative\i
                var\r = -var\r
                *ptrIsNegative\i = #False
              EndIf
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHREAL, var\q)
          EndSelect
          *this\expressionSymbolIndex + 1
        EndIf
      Case #minScript_COMPILER_SYMBOL_LITERALSTRING
        ;First add the string to the string table. 
          result = minScriptCodeCompilerClass_EXPRESSION_AddSymbolToStringTable(*this, Mid(*this\parseTree[*this\expressionSymbolIndex]\symbol$, 2, Len(*this\parseTree[*this\expressionSymbolIndex]\symbol$)-2) , @index)
          If result = #minScript_OKAY
            If blnIsAddressOf
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHADDRSTR, index)
              blnIsAddressOf = #False
            Else
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHSTR, index)
            EndIf
            *this\expressionSymbolIndex + 1
          EndIf
      Case #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
        ;The constant will already have been located.
          If *this\parseTree[*this\expressionSymbolIndex]\constant
            Select *this\parseTree[*this\expressionSymbolIndex]\constant\value\vt
              Case #minScript_INTEGER
                If *ptrIsNegative\i
                  *this\parseTree[*this\expressionSymbolIndex]\constant\value\i = -*this\parseTree[*this\expressionSymbolIndex]\constant\value\i
                  *ptrIsNegative\i = #False
                EndIf
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHINT, *this\parseTree[*this\expressionSymbolIndex]\constant\value\i)
              Case #minScript_CURRENCY
                If *ptrIsNegative\i
                  *this\parseTree[*this\expressionSymbolIndex]\constant\value\c = -*this\parseTree[*this\expressionSymbolIndex]\constant\value\c
                  *ptrIsNegative\i = #False
                EndIf
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHCUR, *this\parseTree[*this\expressionSymbolIndex]\constant\value\c)
              Case #minScript_REAL
                If *ptrIsNegative\i
                  *this\parseTree[*this\expressionSymbolIndex]\constant\value\r = -*this\parseTree[*this\expressionSymbolIndex]\constant\value\r
                  *ptrIsNegative\i = #False
                EndIf
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHREAL, *this\parseTree[*this\expressionSymbolIndex]\constant\value\q)
              Case #minScript_STRING
                ;First add the string to the string table. 
                  *this\parseTree[*this\expressionSymbolIndex]\symbol$ = *this\parseTree[*this\expressionSymbolIndex]\constant\value$
                  result = minScriptCodeCompilerClass_EXPRESSION_AddSymbolToStringTable(*this, *this\parseTree[*this\expressionSymbolIndex]\symbol$, @index)
                  If result = #minScript_OKAY
                    If blnIsAddressOf
                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHADDRSTR, index)
                      blnIsAddressOf = #False
                    Else
                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHSTR, index)
                    EndIf
                  EndIf
            EndSelect
            *this\expressionSymbolIndex + 1
          Else
            result = #minScript_CONSTANTNOTFOUND          
          EndIf
      Case #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION, #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION
        If *this\parseTree[*this\expressionSymbolIndex]\symbolType = #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION
          *runtimeFN = *this\parseTree[*this\expressionSymbolIndex]\runtimeFunction
        EndIf
        fnName$ = *this\parseTree[*this\expressionSymbolIndex]\symbol$
        ;The next symbol will already have been verified as being an open parentheses by the parser.
        *this\expressionSymbolIndex + 2
        If *this\expressionSymbolIndex < *this\numParseSymbols
          ;If the current symbol is a close paranthesis then we have no parameters.
          If *this\parseTree[*this\expressionSymbolIndex]\symbol$ <> ")"
            Repeat
              result = minScriptCodeCompilerClass_EXPRESSION_Recurse(*this, 0)
              If result = #minScript_OKAY
                numParamsPassed + 1
                If *this\expressionSymbolIndex >= *this\numParseSymbols
                  result = #minScript_SYNTAXERROR
                  Break
                EndIf
                If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ")"
                  Break
                ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ","
                  *this\expressionSymbolIndex + 1
                Else
                  result = #minScript_SYNTAXERROR
                  Break
                EndIf
              Else
                Break
              EndIf
            ForEver
          EndIf
          If result = #minScript_OKAY
            *this\expressionSymbolIndex + 1
            If *runtimeFN
              ;Emit object code to call the runtime function.
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_CALLRTFUNCTION, numParamsPassed<<32 + *runtimeFN\fnId)
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHEAX)
            ElseIf numParamsPassed <= #minScript_MAXNUMFUNCTIONPARAMETERS
              ;We pass additional null parameters to give the prerequisite fixed number of parameters.
                For i = numParamsPassed + 1 To #minScript_MAXNUMFUNCTIONPARAMETERS
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHNULL)
                Next
              ;Emit object code to call the function. We do this by referring to the function name in the object code itself
              ;to allow for functions to be added in an adhoc fashion.
                result = minScriptCodeCompilerClass_EXPRESSION_AddSymbolToStringTable(*this, fnName$, @index)
                  If result = #minScript_OKAY
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_CALLFUNCTION, index)
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHEAX)
                  EndIf
            Else
              result = #minScript_TOOMANYPARAMETERS
            EndIf
          EndIf
        Else
          result = #minScript_SYNTAXERROR
        EndIf        
      Case #minScript_COMPILER_SYMBOL_SEPARATOR
        ;We can only accept either an open bracket here as the start of a nested sub-expression.
        If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "(" And *this\expressionSymbolIndex < *this\numParseSymbols - 1
          *this\expressionSymbolIndex + 1 ;Now points at the start of the sub-expression.
          result = minScriptCodeCompilerClass_EXPRESSION_Recurse(*this, expressionVT)
          If result = #minScript_OKAY
            If *this\expressionSymbolIndex < *this\numParseSymbols And *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ")"
              *this\expressionSymbolIndex + 1 ;Now points at the symbol following the close parentheses.
            Else
              result = #minScript_SYNTAXERROR
            EndIf
          EndIf
        Else
          result = #minScript_SYNTAXERROR
        EndIf
      Case #minScript_COMPILER_SYMBOL_KEYWORD
        result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
      Default
        result = #minScript_SYNTAXERROR
    EndSelect
    ;In the case that there are no errors, we may have a series of method invokations.
    ;We delay each object code output in case there is an @ operator in force. In which case only the final invokation is involved, and then it has to be a 'get_' invokation.
    If result = #minScript_OKAY And *this\expressionSymbolIndex < *this\numParseSymbols And symbolType <> #minScript_COMPILER_SYMBOL_LITERALNUMBER And symbolType <> #minScript_COMPILER_SYMBOL_LITERALSTRING And symbolType <> #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
      If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "."
        ;In the case of a variable, we can check right now that it is a reference type.
        If (symbolType <> #minScript_COMPILER_SYMBOL_LOCALVARIABLE And symbolType <> #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE And symbolType <> #minScript_COMPILER_SYMBOL_MODULEVARIABLE) Or blnIsVarReference
          While *this\expressionSymbolIndex < *this\numParseSymbols - 1 And *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "."
            If methodName$
              ;Good to go and emit appropriate object code for the previous method. We do this by referring to the method name in the object code itself
              ;since methods are invoked through late-binding.
              result = minScriptCodeCompilerClass_EXPRESSION_AddSymbolToStringTable(*this, methodName$, @index)
              If result = #minScript_OKAY
                numParamsPassed + 1 ;To include the object pointer itself.
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_METHOD_GETOBJECT, numParamsPassed<<32 + index + blnBracketsEncountered)
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHEAX)
              Else
                Break
              EndIf
            EndIf
            methodName$ = ""
            numParamsPassed = 0
            blnBracketsEncountered = 0
            invokeType = 0
            If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "."
              *this\expressionSymbolIndex + 1
              ;The current symbol should be a method call. The following protects against a literal number etc.
              If *this\parseTree[*this\expressionSymbolIndex]\symbolType <> #minScript_COMPILER_SYMBOL_OBJECTMETHOD
                result = #minScript_SYNTAXERROR
                Break
              EndIf
              methodName$ = *this\parseTree[*this\expressionSymbolIndex]\symbol$
              invokeType = *this\parseTree[*this\expressionSymbolIndex]\invokeType
              *this\expressionSymbolIndex + 1
              ;If the next symbol is an open parentheses then we have to employ the recursive expression compiler on the method parameters.
              If *this\expressionSymbolIndex < *this\numParseSymbols - 1
                If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = "("
                  blnBracketsEncountered = 1<<62
                  numParamsPassed = 0
                  *this\expressionSymbolIndex + 1
                  If *this\expressionSymbolIndex < *this\numParseSymbols
                    ;If the current symbol is a close paranthesis then we have no parameters.
                      If *this\parseTree[*this\expressionSymbolIndex]\symbol$ <> ")"
                        Repeat
                          result = minScriptCodeCompilerClass_EXPRESSION_Recurse(*this, 0)
                          If result = #minScript_OKAY
                            numParamsPassed + 1
                            If *this\expressionSymbolIndex >= *this\numParseSymbols
                              result = #minScript_SYNTAXERROR
                              Break 2
                            EndIf
                            If *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ")"
                              *this\expressionSymbolIndex + 1
                              Break
                            ElseIf *this\parseTree[*this\expressionSymbolIndex]\symbol$ = ","
                              *this\expressionSymbolIndex + 1
                            Else
                              result = #minScript_SYNTAXERROR
                              Break 2
                            EndIf
                          Else
                            Break 2
                          EndIf
                        ForEver
                      Else
                        *this\expressionSymbolIndex + 1
                      EndIf
                  Else
                    result = #minScript_SYNTAXERROR
                    Break
                  EndIf
                EndIf
              EndIf
            Else
              Break
            EndIf
          Wend
          ;There may be a method invokation outstanding.
          If result = #minScript_OKAY And methodName$
            ;In the case of an @ operator being in force we must check for a get property invokation.
            If blnIsAddressOf
              If invokeType <> #minScript_METHOD_SET 
                methodName$ = "@" + methodName$ ;In case there is a convoluted (object property).(index) construct which the execution engine will sort out.
              Else
                result = #minscript_SYNTAXERROR
              EndIf
              blnIsAddressOf = #False             
            EndIf
            If result = #minScript_OKAY
              result = minScriptCodeCompilerClass_EXPRESSION_AddSymbolToStringTable(*this, methodName$, @index)
              If result = #minScript_OKAY
                numParamsPassed + 1 ;To include the object pointer itself.
                Select invokeType
                  Case #minScript_METHOD_INVOKE
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_METHOD_INVOKE, numParamsPassed<<32 + index + blnBracketsEncountered)
                  Case #minScript_METHOD_GET
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_METHOD_GET, numParamsPassed<<32 + index + blnBracketsEncountered)
                  Case #minScript_METHOD_SET
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_METHOD_SET, numParamsPassed<<32 + index + blnBracketsEncountered)
                  Case #minScript_METHOD_GETOBJECT
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_METHOD_GETOBJECT, numParamsPassed<<32 + index + blnBracketsEncountered)
                EndSelect
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHEAX)
              EndIf
            EndIf
          EndIf
        Else
          result = #minScript_INVALIDORNULLOBJECTVARIABLE
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ;Check for a validated type modifier.
    If result = #minScript_OKAY 
      If blnIsAddressOf
        result = #minScript_SYNTAXERROR
      ElseIf *this\parseTree[*this\expressionSymbolIndex-1]\typeModifierVT
        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_CAST, *this\parseTree[*this\expressionSymbolIndex-1]\typeModifierVT)
      EndIf
      If blnIsNot
        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_BITWISENOT)
      EndIf
    EndIf
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following function is the recursive element of the expression engine.
Procedure.i minScriptCodeCompilerClass_EXPRESSION_Recurse(*this.st_minScriptCodeCompilerClass, expressionVT)
  Protected result = #minScript_OKAY
  ;Process terms of lowest precedence.
    result = minScriptCodeCompilerClass_EXPRESSION_ORXORTerms(*this, expressionVT)
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following standalone function evaluates a constant expression for internal use.
;This is needed because this compiler does not feature expression folding.
;Mutex protection will already be provided.
Procedure.i minScriptCodeCompilerClass_EXPRESSION_EvaluateConstantExpression(*this.st_minScriptCodeCompilerClass, expression$, *module.st_minScriptModule)
  Protected result = #minScript_OKAY, tempSnippetModule
  expression$ = "RETURN " + expression$
  ;Reset the constant expression code object.
    minScript_Globals\constantExpressionObject\Reset(#minScript_RESET_REMOVEALL)
  ;Temporarily disable the output of ASM as we don't want the constant expression evaluation to clutter the output.
    minScript_Globals\blnNoDebugASM = #True
  ;Temporarily switch the constant expression code object's snippet module to point to the current one. This gives access to the current module's constants.
    tempSnippetModule = minScript_Globals\ptrConstantExpressionObject\snippetModule
    minScript_Globals\ptrConstantExpressionObject\snippetModule = *module
  ;Compile and execute the expression.
    result = minScript_Globals\constantExpressionObject\Execute(expression$)
  ;Reinstate the constant expression code object's snippet module to point to the current one.
    minScript_Globals\ptrConstantExpressionObject\snippetModule = tempSnippetModule
  ;Reinstate the output of ASM.
    minScript_Globals\blnNoDebugASM = #False
  ;Adjust any error code to make it more relevant to the underlying object as opposed to the global constant expression code object.
    Select result
      Case #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD, #minScript_VARIABLENOTFOUND
        result = #minScript_INVALIDCONSTANTEXPRESSION
    EndSelect
  ProcedureReturn result
EndProcedure
;////////////////////////////////////////////////////////////////////////


;-CLASS METHODS (PUBLIC).

;/////////////////////////////////////////////////////////////////////////////////
Procedure minScriptCodeCompilerClass_Destroy(*this.st_minScriptCodeCompilerClass)
  *this\thisEX\Reset()
  FreeStructure(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure minScriptCodeCompilerClass_Reset(*this.st_minScriptCodeCompilerClass, flags=0)
  With *this\code
    \lastCompilationErrorLine$ = ""
    \lastErrorCode = #minScript_OKAY
    \lastErrorLineNumber = 0
  EndWith
  *this\memoryFile\Reset()
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method is responsible for compiling code.
;It will either compile a bunch of functions or standalone code (outside of a function block) depending on the value of the flags parameter.
Procedure.i minScriptCodeCompilerClass_CompileCode(*this.st_minScriptCodeCompilerClass, *ptrSource.CHARACTER, *module.st_minScriptModule=0)
  Protected result = #minScript_OKAY, blnInsideFunction, blnInsideClassDefinition, i, numOpenBrackets, blnShortAssignment, numParams, index
  Protected *ptrpoKeyword.st_minScriptpoKeyword
  ;Reset fields.
    *this\thisEX\Reset()
    With *this
      \currentLine$ = ""
      \currentLineNumber = 0
      \ptrSource = *ptrSource
      ClearMap(\codeLabels())
      \numTotalConstructs = 0
      ClearList(\constructs())
      \blnLastStatementASelect = #False
      \expressionSymbolIndex = 0
      ;Reset memory file.
        \memoryFile\Reset()
    EndWith
  ;In the case that we are compiling a snippet, we proceed as if a function block had been entered.
    If *module = *this\code\snippetModule
      blnInsideFunction = #True
      ;Since a snippet is allocated to a 'mangled' function, we must carefully set this up including some object code. 
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHEBP)
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_ESPTOEBP)
    EndIf
  ;Fetch successive lines from the source stream and repeat until there are no more lines or an error ensues.
  ;Empty lines are ignored.
    While result = #minScript_OKAY And minScriptCodeCompilerClass_PARSER_GetNextSourceLine(*this)
      ;Output the source line to the debug 'file' if the relevant constant is set.
        CompilerIf Defined(minScript_OUTPUTASMTODEBUGGER, #PB_Constant)
          minScriptCodeCompilerClass_FN_OutputDebugASM(*this, "Line " + Str(*this\currentLineNumber) + " : " + *this\currentLine$)
        CompilerEndIf
      ;Initialise the parse tree and check for errors. This will only extract the first symbol as for some script commands we do not
      ;want to extract all symbols at this stage.
        result = minScriptCodeCompilerClass_PARSER_BeginParseTree(*this, *module)
        If result <> #minScript_OKAY
          Break
        EndIf
      If *this\numParseSymbols ;Just in case.
        If *this\parseTree[0]\symbol$ = "[" ;A compiler directive.
          ;We do not issue a #MS_OP_DEBUG_NOP instruction.
          result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
          If result = #minScript_OKAY
            result = *this\thisEX\DIRECTIVE()
          EndIf
        ElseIf *this\parseTree[0]\symbol$ = "debug"
          ;Only proceed if we are inside a function.
          If blnInsideFunction
            ;Emit a debugging #MS_OP_DEBUG_NOP instruction.
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_DEBUG_NOP, *this\currentLineNumber)
            result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
            If result = #minScript_OKAY
              result = *this\thisEX\DEBUG()
            EndIf
          EndIf
        ElseIf *this\currentClass
          ;Only ENDCLASS and PROPERTY are allowed.
          Select *this\parseTree[0]\symbol$
            Case "class"
              result = #minScript_NESTEDCLASSES
            Case "endclass"
              result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
              If result = #minScript_OKAY
                result = *this\thisEX\ENDCLASS()
              EndIf
            Default ;Only a property declaration is permitted.
              result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
              If result = #minScript_OKAY
                result = *this\thisEX\PROPERTY()
              EndIf
          EndSelect
        Else ;We are outside of a class definition.
          If *this\parseTree[0]\symbol$ = "endclass" Or *this\parseTree[0]\symbol$ = "property"
            result = #minScript_NOTALLOWEDOUTSIDEOFCLASS
          ElseIf *this\parseTree[0]\symbol$ = "class"
            If *Module = *this\code\snippetModule
              result = #minScript_NOTALLOWEDOUTSIDEOFMODULE
            ElseIf blnInsideFunction
              result = #minScript_MISSINGENDFUNCTIONORMETHOD
            Else ;Good to go.
              result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
              If result = #minScript_OKAY
                result = *this\thisEX\CLASS(*module)
              EndIf
            EndIf          
          Else  
            ;Emit a debugging #MS_OP_DEBUG_NOP instruction.
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_DEBUG_NOP, *this\currentLineNumber)
            ;Check whether an appropriate statement follows a SELECT.
              If *this\blnLastStatementASelect
                If *this\parseTree[0]\symbol$ <> "case" And *this\parseTree[0]\symbol$ <> "default" And *this\parseTree[0]\symbol$ <> "endselect"
                  result = #minScript_CASEORDEFAULTEXPECTED
                  Break
                EndIf
                *this\blnLastStatementASelect = #False
              EndIf
            ;Examine the first token.
            ;We separate those commands which can reside outside of a function block or cannot have their entire parse trees created at this stage.
            Select *this\parseTree[0]\symbol$
              Case "break"    ;Cannot process parse tree.
                If blnInsideFunction = 0
                  result = #minScript_CODEOUTSIDEFUNCTIONBLOCK
                Else
                  result = *this\thisEX\BREAK(*module.st_minScriptModule)
                EndIf
              Case "dim"      ;Can be placed outside of a function block. If inside the snippet module, then we are guaranteed to be inside a function.
                result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
                If result = #minScript_OKAY
                  result = *this\thisEX\DIM(*module, blnInsideFunction)
                EndIf
              Case "for"      ;Cannot necessarily process the entire parse tree.
                If blnInsideFunction = 0
                  result = #minScript_CODEOUTSIDEFUNCTIONBLOCK
                Else
                  result = *this\thisEX\FOR(*module)
                EndIf
              Case "private", "function", "method" ;Must be placed outside of a function block.
                If *Module = *this\code\snippetModule
                  result = #minScript_NOTALLOWEDOUTSIDEOFMODULE
                ElseIf blnInsideFunction
                  result = #minScript_NESTEDFUNCTIONSORMETHODS
                Else
                  result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
                  If result = #minScript_OKAY
                    ;Hand over to the \FUNCTION() method for further processing.
                      result = *this\thisEX\FUNCTION(*module)
                      If result = #minScript_OKAY
                        blnInsideFunction = #True        
                      EndIf
                  EndIf
                EndIf
              Case "let"  ;Cannot process parse tree until we have decided between a constant assignment and a regular one, and can be placed outside of a function block.
                ;We need to determine if we are dealing with a constant assignment.
                result = minScriptCodeCompilerClass_PARSER_ContinueParseTree(*this, *module, "=")
                If result = #minScript_OKAY 
                  If *this\parseTree[*this\numParseSymbols-1]\symbol$ = "="
                    If *this\parseTree[1]\symbolType = #minScript_COMPILER_SYMBOL_GLOBALCONSTANT
                      result = *this\thisEX\DEFINE(*module)
                    ElseIf blnInsideFunction
                      result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
                      If result = #minScript_OKAY
                        result = *this\thisEX\LET()
                      EndIf
                    Else
                      result = #minScript_CODEOUTSIDEFUNCTIONBLOCK
                    EndIf
                  Else
                    result = #minScript_SYNTAXERROR
                  EndIf
                EndIf
              Default ;All remaining possible commands (bar a certain shortcut!) must reside within a function block and we can complete the parse tree up front.
                If *this\parseTree[0]\symbolType = #minScript_COMPILER_SYMBOL_GLOBALCONSTANT 
                  ;Only option is for a shortcut DEFINE statement. We'll check for an undefined constant later.
                  ;Process the parse tree up to the first '='.
                  result = minScriptCodeCompilerClass_PARSER_ContinueParseTree(*this, *module, "=")
                  If result = #minScript_OKAY
                    If *this\parseTree[*this\numParseSymbols-1]\symbol$ = "="
                      result = *this\thisEX\DEFINE(*module, #True)            
                    Else
                      result = #minScript_SYNTAXERROR
                    EndIf
                  EndIf
                Else          
                  result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module)
                  If result = #minScript_OKAY
                    If blnInsideFunction
                      Select *this\parseTree[0]\symbol$
                        Case "case"
                          result = *this\thisEX\CASE()
                        Case "default"
                          result = *this\thisEX\DEFAULT()
                        Case "else"
                          result = *this\thisEX\ELSE()
                        Case "elseif"
                          result = *this\thisEX\ELSEIF()
                        Case "endfunction", "endmethod"
                          If *module = *this\code\snippetModule
                            result = #minScript_CODEOUTSIDEFUNCTIONBLOCK
                          Else
                          ;Hand over to the \ENDFUNCTION() method for further processing.
                            result = *this\thisEX\ENDFUNCTION()
                            blnInsideFunction = #False
                          EndIf
                        Case "endif"
                          result = *this\thisEX\ENDIF()
                        Case "endselect"
                          result = *this\thisEX\ENDSELECT()
                        Case "if"
                          result = *this\thisEX\IF()
                        Case "next"
                          result = *this\thisEX\NEXT()
                        Case "repeat"
                          result = *this\thisEX\REPEAT()
                        Case "return"
                          result = *this\thisEX\RETURN()
                        Case "select"
                          result = *this\thisEX\SELECT()
                        Case "static"
                          If *Module = *this\code\snippetModule
                            result = #minScript_NOTALLOWEDOUTSIDEOFMODULE
                          Else
                            result = *this\thisEX\STATIC(*module)
                          EndIf
                        Case "until"
                          result = *this\thisEX\UNTIL()
                        Case "wend"
                          result = *this\thisEX\WEND()
                        Case "while"
                          result = *this\thisEX\WHILE()
                      Default
                        ;First check for a poKeyword.
                        CompilerIf #PB_Compiler_Thread
                          LockMutex(minScript_Globals\threadMutex)
                        CompilerEndIf
                        *ptrpoKeyword = FindMapElement(minScript_Globals\minScriptpoKeywords(), *this\parseTree[0]\symbol$)
                        CompilerIf #PB_Compiler_Thread
                          UnlockMutex(minScript_Globals\threadMutex)
                        CompilerEndIf
                        If *ptrpoKeyword
                          result = *this\thisEX\POKEYWORD(*ptrpoKeyword\address)
                        Else
                          ;We first try a shortcut assignment. The possibility of a shortcut constant assignment has already been covered above.
                            result = *this\thisEX\LET(#True)
                          If result <> #minScript_OKAY
                            ;We allow certain shorthand function calls etc. and simply discard the result.
                            Select *this\parseTree[0]\symbolType
                              Case #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION, #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
                                *this\expressionSymbolIndex = 0
                                result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
                                If result = #minScript_OKAY
                                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POP, 1)
                                EndIf
                              Case #minScript_COMPILER_SYMBOL_LOCALVARIABLE, #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE, #minScript_COMPILER_SYMBOL_MODULEVARIABLE
                                ;Do we have a method invokation? If so, we proceed as in the above case.
                                If *this\numParseSymbols > 1 And *this\parseTree[1]\symbol$ = "."
                                  *this\expressionSymbolIndex = 0
                                  result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
                                  If result = #minScript_OKAY
                                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POP, 1)
                                  EndIf
                                ElseIf *this\numParseSymbols > 1
                                  ;Finally, we allow an 'intelligent' assignment such as 'a*2 + 1' which will be equivalent to 'a = a*2 + 1'.
                                  *this\expressionSymbolIndex = 0
                                  result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
                                  If result = #minScript_OKAY
                                    If *this\parseTree[0]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE
                                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPLOCALVAR, *this\parseTree[0]\localVariable\EBPoffset)
                                    Else
                                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPSCRIPTVAR, *this\parseTree[0]\scriptVariable)
                                    EndIf
                                  EndIf
                                Else
                                  result = #minScript_SYNTAXERROR
                                EndIf
                              Default
                                result = #minScript_SYNTAXERROR
                            EndSelect
                          EndIf
                        EndIf
                      EndSelect
                    Else
                      result = #minScript_CODEOUTSIDEFUNCTIONBLOCK
                    EndIf
                  EndIf
                EndIf
            EndSelect
          EndIf
        EndIf
      EndIf
    Wend
  ;Undertake additional error checking.
    If result = #minScript_OKAY
      If blnInsideFunction And *module <> *this\code\snippetModule 
        result = #minScript_MISSINGENDFUNCTIONORMETHOD
        *this\code\lastErrorCode = result
      ElseIf *this\currentClass
        result = #minScript_MISSINGENDCLASS
        *this\code\lastErrorCode = result
      ;Was there an error with the memory file.
      ElseIf *this\memoryFile\CheckForErrors()
        result = #minScript_OUTOFMEMORY
        *this\code\lastErrorCode = result
      EndIf
      If result = #minScript_OKAY And *module = *this\code\snippetModule
        ;In the case of a code snippet, we need to copy the object file and run through the object code fixing up jmp labels etc.
        ;We can do this by invoking the ENDFUNCTION method - after a bit of juggling. This method will check for outstanding constructs etc.     
          *this\numParseSymbols = 1 ;Required to pass the \ENDFUNCTION() vlaidation.
          result = *this\thisEX\ENDFUNCTION()
          *this\code\lastErrorCode = result
      EndIf
  ;Set error info as appropriate.
    Else
      *this\code\lastErrorCode = result
      *this\code\lastCompilationErrorLine$ = *this\currentLine$
      *this\code\lastErrorLineNumber = *this\currentLineNumber    
    EndIf
  ;If an error was recorded then we remove any part function or class. Keeps things tidy and avoids any memory leaks.
    If result <> #minScript_OKAY 
      If *this\code\currentScriptFunction
        DeleteMapElement(*module\functions(), *this\code\currentScriptFunction\name$)
      ElseIf *this\currentClass
        DeleteMapElement(*this\code\classes(), *this\currentClass\name$)
      EndIf
    EndIf
  *this\code\currentScriptFunction = 0
  *this\code\blnEnableExplicit = #True
  *this\currentClass = 0
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS METHODS (INTERNAL).

;/////////////////////////////////////////////////////////////////////////////////
;The following is the central hub through which all object code is emitted. We do not perform any 'peep-hole' optimisation at this time (unlike with our eScript engine).
;Error checking is left until the entire object file has been written as we can safely attempt to write to the memory file regardless.
Procedure minScriptCodeCompilerEXClass_EmitObjectCodeInstruction(*this.st_minScriptCodeCompilerClass, opCode, operand.q=0)
  CompilerIf Defined(minScript_OUTPUTASMTODEBUGGER, #PB_Constant)
    Protected loc, asm$
    loc = *this\memoryFile\Loc()
  CompilerEndIf
  ;Write op-code.
    *this\memoryFile\WriteWord(OpCode)
  ;Write operand.
    *this\memoryFile\WriteQuad(Operand)
  ;Output a suitable mnemonic to the debug 'file' if the relevant constant is set.
  ;We essentially 'disassemble' the object code instruction.
    CompilerIf Defined(minScript_OUTPUTASMTODEBUGGER, #PB_Constant)
      If minScript_Globals\blnNoDebugASM = #False
        ;No point proceeding if there is a memory file error.
        If *this\memoryFile\CheckForErrors() = #False
          asm$ = minScriptCodeClass_DISASSEMBLY_DisassembleObjectCodeInstruction(*this\code, opCode, operand, *this)
          minScriptCodeCompilerClass_FN_OutputDebugASM(*this, "  " + RSet(Str(loc), 4, "0") + " : " + asm$)
        EndIf
      EndIf
    CompilerEndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_BREAK(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule)
  Protected result = #minScript_OKAY, breakLevels=1, codeAddress, loc, t1, constantExp$
  ;First job is to evaluate what should be a constant expression. Because we do not employ expression folding, we have to cheat!
  ;Identify the 'command string'.  
    constantExp$ = Right(*this\currentLine$, Len(*this\currentLine$)-6)
    If constantExp$
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
      result = minScriptCodeCompilerClass_EXPRESSION_EvaluateConstantExpression(*this, constantExp$, *module)
      If result = #minScript_OKAY
        If minScript_Globals\constantExpressionReturnVar\vt = #minScript_INTEGER
          breakLevels = minScript_Globals\constantExpressionReturnVar\i
        Else
          result = #minScript_INVALIDBREAKLEVEL
        EndIf
      EndIf
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
    EndIf
  If result = #minScript_OKAY
    ;We have to back-track through the constructs() stack looking for a construct which can be broken.
    If *this\numTotalConstructs
      result = #minScript_INVALIDUSEOFBREAK
      t1 = breakLevels
      Repeat
        Select *this\constructs()\type
          Case #minScript_COMPILER_CONSTRUCT_FOR
            breakLevels - 1
            If breakLevels = 0
              ;No need to check the following line for errors.
              codeAddress = *this\thisEX\AddVirtualCodeLabel("next_" + Str(*this\constructs()\constructIndex1))
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeAddress)
              result = #minScript_OKAY
              Break
            EndIf
          Case #minScript_COMPILER_CONSTRUCT_REPEAT
            breakLevels - 1
            If breakLevels = 0
              codeAddress = *this\thisEX\AddVirtualCodeLabel("until_" + Str(*this\constructs()\constructIndex1))
              If codeAddress
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeAddress)
                result = #minScript_OKAY
              Else
                result = #minScript_OUTOFMEMORY
              EndIf
              Break
            EndIf
          Case #minScript_COMPILER_CONSTRUCT_WHILE
            breakLevels - 1
            If breakLevels = 0
              ;No need to check the following line for errors.
              codeAddress = *this\thisEX\AddVirtualCodeLabel("wend_" + Str(*this\constructs()\constructIndex1))
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeAddress)
              result = #minScript_OKAY
              Break
            EndIf
        EndSelect
      Until PreviousElement(*this\constructs()) = 0
      LastElement(*this\constructs())
      If result = #minScript_INVALIDUSEOFBREAK And breakLevels <> t1
        result = #minScript_INVALIDBREAKLEVEL
      EndIf
    Else
      result = #minScript_INVALIDUSEOFBREAK
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_CASE(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeLabelAddress
  If *this\numParseSymbols > 1
    ;Check we are within a SELECT/ENDSELECT.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_SELECT
        ;Make sure no DEFAULT has been processed.
          If *this\constructs()\blnDEFAULTEncountered = #False
            ;Do we need to issue a JMP because of a preceeding CASE?
              If *this\constructs()\numCASEs
                ;No need to check the return value as this is guaranteed to be valid by the \SELECT() method.
                  codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("endselect_" + Str(*this\constructs()\constructIndex1))
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeLabelAddress)
                ;Add a case bypass label for the previous CASE statement.
                  If *this\thisEX\AddVirtualCodeLabel("select_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numCASEs), *this\memoryFile\Loc()) = 0
                    result = #minScript_OUTOFMEMORY
                  EndIf
              EndIf
            If result = #minScript_OKAY
              *this\constructs()\numCASEs + 1
              ;Now arrange for a comparison with the SELECT expression.            
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_DUP)
                ;Hand over to the expression engine.
                  *this\expressionSymbolIndex = 1
                  result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
                If result = #minScript_OKAY
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_EQU)
                  codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("select_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numCASEs))
                  If codeLabelAddress
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMPZ, codeLabelAddress)
                  Else
                    result = #minScript_OUTOFMEMORY
                  EndIf
                EndIf
            EndIf
          Else
            result = #minScript_DEFAULTORELSEALREADYENCOUNTERED
          EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Class name will already have been validated.
Procedure.i minScriptCodeCompilerEXClass_CLASS(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule)
  Protected result = #minScript_OKAY, *ptrClass.st_minScriptClass
  If *this\numParseSymbols = 2
    *this\currentClass = 0 ;Just in case.
    Select *this\parseTree[1]\symbolType
      Case #minScript_COMPILER_SYMBOL_LOCALVARIABLE, #minScript_COMPILER_SYMBOL_MODULEVARIABLE
        ;First, try and locate an existing class of the given name.
          *ptrClass = FindMapElement(*this\code\classes(), *this\parseTree[1]\symbol$)
        If *ptrClass = 0
          ;Allocate a new map entry.
            *ptrClass = AddMapElement(*this\code\classes(), *this\parseTree[1]\symbol$)
          If *ptrClass
            *ptrClass\name$ = *this\parseTree[1]\symbol$
            *ptrClass\code = *this\code
            *ptrClass\module = *module
            ;If the threadsafe switch is set then we attempt to create threaded fields. For this we need to employ the global mutex.
              CompilerIf #PB_Compiler_Thread
                LockMutex(minScript_Globals\threadMutex)
                  *ptrClass\tsiConstructorActive = CreateThreadStorage()
                  If *ptrClass\tsiConstructorActive
                    *ptrClass\tsiDestructorActive = CreateThreadStorage()
                    If *ptrClass\tsiDestructorActive = 0
                      DeleteThreadStorage(*ptrClass\tsiConstructorActive)
                      DeleteMapElement(*this\code\classes())
                      result = #minScript_OUTOFMEMORY
                    EndIf
                  Else
                    DeleteMapElement(*this\code\classes())
                    result = #minScript_OUTOFMEMORY
                  EndIf
              CompilerEndIf
            If result = #minScript_OKAY            
              ;Determine an appropriate class VT.
              ;Does a class VT already exist.
                If FindMapElement(*this\code\classVTs(), *this\parseTree[1]\symbol$)
                  *ptrClass\classVT = *this\code\classVTs()
                Else
                  ;Create a new class VT.
                  If AddMapElement(*this\code\classVTs(), *this\parseTree[1]\symbol$)
                    minScript_Globals\currentClassPartVT + 1
                    *this\code\classVTs() = #minScript_CLASS + minScript_Globals\currentClassPartVT
                    *ptrClass\classVT = *this\code\classVTs()
                  Else
                    CompilerIf #PB_Compiler_Thread
                      DeleteThreadStorage(*ptrClass\tsiConstructorActive)
                      DeleteThreadStorage(*ptrClass\tsiDestructorActive)
                    CompilerEndIf
                    DeleteMapElement(*this\code\classes())
                    result = #minScript_OUTOFMEMORY
                  EndIf
                EndIf
            EndIf          
            CompilerIf #PB_Compiler_Thread
              UnlockMutex(minScript_Globals\threadMutex)
            CompilerEndIf
          Else
            result = #minScript_OUTOFMEMORY
          EndIf
        Else
          result = #minScript_CLASSALREADYDECLARED
        EndIf
        If result = #minScript_OKAY
          *this\currentClass = *ptrClass
        EndIf
      Case #minScript_COMPILER_SYMBOL_KEYWORD
        result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
      Default
        result = #minScript_SYNTAXERROR    
    EndSelect
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

  
;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_DEBUG(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols > 1 ;The minimum number of symbols.
    ;Hand over to the expression engine.
      *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
      result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
    If result = #minScript_OKAY
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_DEBUG)
    EndIf
  Else
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_DEFAULT(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeLabelAddress
  If *this\numParseSymbols = 1
    ;Check we are within a SELECT/ENDSELECT.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_SELECT
        ;Make sure no DEFAULT has been processed.
          If *this\constructs()\blnDEFAULTEncountered = #False
            *this\constructs()\blnDEFAULTEncountered = #True
            ;Do we need to issue a JMP because of a preceeding CASE?
              If *this\constructs()\numCASEs
                ;No need to check the return value as this is guaranteed to be valid by the \SELECT() method.
                  codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("endselect_" + Str(*this\constructs()\constructIndex1))
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeLabelAddress)
                ;Add a case bypass label for the previous CASE statement.
                  If *this\thisEX\AddVirtualCodeLabel("select_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numCASEs), *this\memoryFile\Loc()) = 0
                    result = #minScript_OUTOFMEMORY
                  EndIf
              EndIf
          Else
            result = #minScript_DEFAULTORELSEALREADYENCOUNTERED
          EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The parse tree will have been processed up to the first '=' symbol.
;Options are a statement of the form LET #constant = ...
;or #constant = ...
Procedure.i minScriptCodeCompilerEXClass_DEFINE(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule, blnShortCutAssignment=0)
  Protected result = #minScript_OKAY, constantExp$
  ;Make sure we are not compiling a snippet etc.
  If *module <> *this\code\snippetModule
    ;Do we have the beginnings of a correctly formatted assignment?
    If *this\numParseSymbols = 3 - blnShortCutAssignment
    ;Has the constant already been declared?
      If *this\parseTree[1-blnShortCutAssignment]\constant = 0
        ;Evaluate what should be a constant expression. Because we do not employ expression folding, we have to cheat!
          ;Identify the 'command string'.  
            constantExp$ = PeekS(*this\parseCharPtr)
          If constantExp$
            CompilerIf #PB_Compiler_Thread
              LockMutex(minScript_Globals\threadMutex)
            CompilerEndIf
            result = minScriptCodeCompilerClass_EXPRESSION_EvaluateConstantExpression(*this, constantExp$, *module)
            If result = #minScript_OKAY
              ;Add a new module constant.
                If AddMapElement(*module\moduleConstants(), *this\parseTree[1-blnShortCutAssignment]\symbol$)
                  *module\moduleConstants()\friendlyName$ = *this\parseTree[1-blnShortCutAssignment]\symbol$
                  *module\moduleConstants()\value\vt = minScript_Globals\constantExpressionReturnVar\vt                 
                  If minScript_Globals\constantExpressionReturnVar\vt = #minScript_STRING
                    If minScript_Globals\constantExpressionReturnVar\sPtr
                      *module\moduleConstants()\value$ = PeekS(minScript_Globals\constantExpressionReturnVar\sPtr)
                    EndIf
                  Else ; A numeric constant.
                    *module\moduleConstants()\value\q = minScript_Globals\constantExpressionReturnVar\q
                  EndIf
                Else
                  result = #minScript_OUTOFMEMORY
                EndIf
            EndIf
            CompilerIf #PB_Compiler_Thread
              UnlockMutex(minScript_Globals\threadMutex)
            CompilerEndIf
          Else
            result = #minScript_SYNTAXERROR
          EndIf
      Else
        result = #minScript_CONSTANTALREADYDECLARED      
      EndIf
    Else
      result = #minScript_SYNTAXERROR
    EndIf
  Else
    result = #minScript_NOTALLOWEDOUTSIDEOFMODULE  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_DIM(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule, blnInsideFunction)
  Protected result = #minScript_OKAY, varType.u, index
  If *this\numParseSymbols >= 4 ;The minimum number of symbols.
    If *this\parseTree[*this\numParseSymbols-2]\symbol$ = "as"
      ;Sort out the variable type.
        varType = minScriptCodeClass_HELPER_String2VT(*this\code, *this\parseTree[*this\numParseSymbols-1]\symbol$)
        If varType = 0  
          result = #minScript_UNRECOGNISEDVARIABLETYPE
        EndIf
      If result = #minScript_OKAY
        ;We need to examine each variable declaration.
        index = 1
        While index < *this\numParseSymbols-2
          If blnInsideFunction
            ;Make sure we have a local variable. Undeclared module variables will be recorded, at this stage, as local variables.
            If *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE Or *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
              ;Check variable is not already defined. Rather than use any existing value of the parseTree *localVariable field,
              ;which may not help in this instance if this DIM statement contains repeated declarations, we turn to the local symbols map instead.
                *this\parseTree[index]\localVariable = FindMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *this\parseTree[index]\symbol$)           
                If *this\parseTree[index]\localVariable = 0
                  result = minScriptCodeCompilerClass_HELPER_AddLocalVariable(*this, @*this\parseTree[index], varType)
                ElseIf *this\parseTree[index]\localVariable\vt <> varType Or *this\parseTree[index]\localVariable\staticModuleVar
                  result = #minScript_VARIABLEORPROPERTYALREADYDECLARED
                EndIf
            ElseIf *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
              Goto minScriptCodeCompilerEXClass_DIM_L1
            Else ;Set an appropriate error code.
              Select *this\parseTree[index]\symbolType
                Case #minScript_COMPILER_SYMBOL_KEYWORD
                  result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
                Default
                  result = #minScript_SYNTAXERROR
              EndSelect
            EndIf
          Else ;Outside of a function block so we are dealing with module variables (or script variables).
            ;Make sure we have a module variable. Undeclared module variables will be recorded, at this stage, as local variables.
            If *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE Or *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
              ;Check variable is not already defined. Rather than use any existing value of the parseTree *moduleVariable field,
              ;which may not help in this instance if this DIM statement contains repeated declarations, we turn to the module symbols map instead.
                *this\parseTree[index]\moduleVariable = FindMapElement(*module\moduleVariables(), *this\parseTree[index]\symbol$)           
                If *this\parseTree[index]\moduleVariable = 0
                  ;In the case of a user-defined type, we cannot allow the variable to have the same name as run-time function because such variables
                  ;can be used to invoke certain 'methods'.
                    CompilerIf #PB_Compiler_Thread
                      LockMutex(minScript_Globals\threadMutex)
                    CompilerEndIf
                    If (varType & #minScript_REFERENCE = 0 And varType <> #minScript_VARIANT) Or FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), *this\parseTree[index]\symbol$) = 0
                      ;Create a new module variable.
                      *this\parseTree[index]\moduleVariable = AddMapElement(*module\ModuleVariables(), *this\parseTree[index]\symbol$)
                      If *this\parseTree[index]\moduleVariable 
                        *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
                        *this\parseTree[index]\moduleVariable\vt = varType
                      Else
                        result = #minScript_OUTOFMEMORY                  
                      EndIf
                    Else
                      result = #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
                    EndIf
                    CompilerIf #PB_Compiler_Thread
                      UnlockMutex(minScript_Globals\threadMutex)
                    CompilerEndIf
                ElseIf *this\parseTree[index]\moduleVariable\vt <> varType
                  result = #minScript_VARIABLEORPROPERTYALREADYDECLARED
                EndIf
            ElseIf *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
minScriptCodeCompilerEXClass_DIM_L1:
              ;Cannot allow the creation of a script variable in the shared class object.
              If *this\code <> minScriptGlobalClassLib
                *this\parseTree[index]\scriptVariable = FindMapElement(*this\code\scriptVariables(), *this\parseTree[index]\symbol$)           
                If *this\parseTree[index]\scriptVariable = 0
                  result = minScriptCodeCompilerClass_HELPER_AddScriptVariable(*this, @*this\parseTree[index], varType)
                ElseIf *this\parseTree[index]\scriptVariable\vt <> varType
                  result = #minScript_VARIABLEORPROPERTYALREADYDECLARED
                EndIf
              Else
                result = #minScript_ACCESSDENIED              
              EndIf
            Else ;Set an appropriate error code.
              Select *this\parseTree[index]\symbolType
                Case #minScript_COMPILER_SYMBOL_KEYWORD
                  result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
                Default
                  result = #minScript_SYNTAXERROR
              EndSelect
            EndIf
          EndIf
          If result = #minScript_OKAY
            If index = *this\numParseSymbols - 3
              Break
            Else ;Next symbol needs to be a comma.
              If *this\parseTree[index+1]\symbol$ = "," And index < *this\numParseSymbols - 4
                index + 2
              Else
                result = #minScript_SYNTAXERROR
                Break
              EndIf
            EndIf
          Else
            Break
          EndIf
        Wend
      EndIf
    Else
      result = #minScript_SYNTAXERROR    
    EndIf
  Else
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_DIRECTIVE(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols >= 3 And *this\parseTree[*this\numParseSymbols-1]\symbol$ = "]"
    Select *this\parseTree[1]\symbol$
      Case "disableexplicit"
        If *this\numParseSymbols = 3 
          *this\code\blnEnableExplicit = #False
        Else
          result = #minScript_SYNTAXERROR  
        EndIf
      Case "enableexplicit"
        If *this\numParseSymbols = 3 
          *this\code\blnEnableExplicit = #True
        Else
          result = #minScript_SYNTAXERROR  
        EndIf
      Default    
        result = #minScript_SYNTAXERROR  
    EndSelect
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_ELSE(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeLabelAddress
  If *this\numParseSymbols = 1
    ;Check we are within an IF/ENDIF block.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_IF
        If *this\constructs()\blnELSEEncountered = #False
          ;Emit object code to bypass subsequent instructions within the 'Else' clause.
            codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("endif_" + Str(*this\constructs()\constructIndex1))
            If codeLabelAddress  
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeLabelAddress)
              ;Add a bypass label for the previous IF/ELSEIF statement.
              If *this\thisEX\AddVirtualCodeLabel("if_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numELSEIFs), *this\memoryFile\Loc())
                *this\constructs()\blnELSEEncountered = #True
              Else
                result = #minScript_OUTOFMEMORY
              EndIf
            Else
              result = #minScript_OUTOFMEMORY
            EndIf
        Else
          result = #minScript_DEFAULTORELSEALREADYENCOUNTERED
        EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_ELSEIF(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeLabelAddress
  If *this\numParseSymbols > 1
    ;Check we are within a IF/ENDIF.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_IF
        ;Make sure no ELSE has been processed.
          If *this\constructs()\blnELSEEncountered = #False
            ;We need to issue a JMP 
              ;No need to check the return value as this is guaranteed to be valid by the \IF() method.
                codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("endif_" + Str(*this\constructs()\constructIndex1))
                *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeLabelAddress)
            ;Add a bypass label for the previous IF/ELSEIF statement.
              If *this\thisEX\AddVirtualCodeLabel("if_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numELSEIFs), *this\memoryFile\Loc()) = 0
                result = #minScript_OUTOFMEMORY
              EndIf
            If result = #minScript_OKAY
              *this\constructs()\numELSEIFs + 1
              ;Hand over to the expression engine.
                *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
                result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
              If result = #minScript_OKAY
                codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("if_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numELSEIFs))
                If codeLabelAddress
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMPZ, codeLabelAddress)
                Else
                  result = #minScript_OUTOFMEMORY
                EndIf
              EndIf
            EndIf
          Else
            result = #minScript_DEFAULTORELSEALREADYENCOUNTERED
          EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_ENDCLASS(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols <> 1
    result = #minScript_SYNTAXERROR
  EndIf
  *this\currentClass = 0
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_ENDFUNCTION(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, memoryFileSize
  If *this\numParseSymbols = 1
    ;Are there any outstanding constructs?
    If ListSize(*this\constructs()) = 0
      ;Emit code to tidy up the stack frame.
        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_EBPTOESP)
        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPEBP)
        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_RET)
      ;Set the object code buffer.
      memoryFileSize = *this\memoryFile\Lof()
      If memoryFileSize 
        *this\code\currentScriptFunction\objectCodeBufferPtr = AllocateMemory(memoryFileSize)
        If *this\code\currentScriptFunction\objectCodeBufferPtr 
          CopyMemory(*this\memoryFile\GetFileMemoryBase(), *this\code\currentScriptFunction\objectCodeBufferPtr, memoryFileSize)
          ;Run through the object code and fix up any references to virtual code labels.
            *this\thisEX\FixUpCodeReferences(*this\code\currentScriptFunction\objectCodeBufferPtr)
          ;Reset certain compilation fields in case the source string contains additional functions.
            With *this
              \code\currentScriptFunction = 0
              ClearMap(\codeLabels())
              \numTotalConstructs = 0
              ClearList(\constructs())  ;Just in case!
              \expressionSymbolIndex = 0
              ;Reset memory file.
              \memoryFile\Reset()
            EndWith
        Else
          result = #minScript_OUTOFMEMORY
        EndIf
      EndIf
    Else
      result = #minScript_MISSINGENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_ENDIF(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols = 1
    ;Check we are within an IF/ENDIF block.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_IF
        ;We need to set an appropriate virtual code label.
        If *this\constructs()\blnELSEEncountered= #False
          ;We need to set up a virtual jump label.
          If *this\thisEX\AddVirtualCodeLabel("if_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numCASEs), *this\memoryFile\Loc()) = 0
            result = #minScript_OUTOFMEMORY
          EndIf
        EndIf
        If result = #minScript_OKAY
          If *this\thisEX\AddVirtualCodeLabel("endif_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc())
            ;Pop the 'If/EndIf' construct.
            *this\thisEX\PopConstruct()
          Else
            result = #minScript_OUTOFMEMORY
          EndIf
        EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_ENDSELECT(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeLabelAddress
  If *this\numParseSymbols = 1
    ;Check we are within a SELECT/ENDSELECT block.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_SELECT
        If *this\constructs()\numCASEs And *this\constructs()\blnDEFAULTEncountered = #False
          ;We need to set up a virtual jump label.
          If *this\thisEX\AddVirtualCodeLabel("select_" + Str(*this\constructs()\constructIndex1) + "_" + Str(*this\constructs()\numCASEs), *this\memoryFile\Loc()) = 0
            result = #minScript_OUTOFMEMORY
          EndIf
        EndIf
        If result = #minScript_OKAY
          If *this\thisEX\AddVirtualCodeLabel("endselect_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc())
            ;Emit object code to remove the top of the stack.
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POP, 1)
            *this\thisEX\PopConstruct()
          Else
            result = #minScript_OUTOFMEMORY
          EndIf
        EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_FOR(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule)
  Protected result = #minScript_OKAY, stepValue=1, constantExp$, opPopVar, codeAddress
  result = minScriptCodeCompilerClass_PARSER_FinishParseTree(*this, *module, "step")
  If result = #minScript_OKAY
    If *this\numParseSymbols >= 6
      result = *this\thisEX\PushConstruct(#minScript_COMPILER_CONSTRUCT_FOR)
      If result = #minScript_OKAY
        ;Is there a STEP clause?
          If *this\parseTree[*this\numParseSymbols-1]\symbol$ = "step"
            ;Extract the relevant substring.
              constantExp$ = PeekS(*this\parseSymbol)
            If constantExp$
              CompilerIf #PB_Compiler_Thread
                LockMutex(minScript_Globals\threadMutex)
              CompilerEndIf
              result = minScriptCodeCompilerClass_EXPRESSION_EvaluateConstantExpression(*this, constantExp$, *module)
              If result = #minScript_OKAY
                If minScript_Globals\constantExpressionReturnVar\vt = #minScript_INTEGER
                  stepValue = minScript_Globals\constantExpressionReturnVar\i
                  ;Effectively remove the 'STEP' clause from the parse tree.
                    *this\numParseSymbols-1
                Else
                  result = #minScript_INVALIDBREAKLEVEL
                EndIf
              EndIf
              CompilerIf #PB_Compiler_Thread
                UnlockMutex(minScript_Globals\threadMutex)
              CompilerEndIf
            Else
              result = #minScript_SYNTAXERROR            
            EndIf
          EndIf
          *this\constructs()\stepValue = stepValue
        If result = #minScript_OKAY
          ;Check that the index variable is an integer type.
            If *this\parseTree[1]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE 
              If *this\parseTree[1]\localVariable
minScriptCodeCompilerEXClass_FOR_L1:
                opPopVar = #MS_OP_POPLOCALVAR
                *this\constructs()\opPushVar = #MS_OP_PUSHLOCALVAR
                *this\constructs()\varOperand = *this\parseTree[1]\localVariable\EBPoffset
                If *this\parseTree[1]\localVariable\vt <> #minScript_INTEGER And *this\parseTree[1]\localVariable\vt & #minScript_VARIANT = 0
                  result = #minScript_TYPEMISMATCH
                EndIf
              ElseIf *this\code\blnEnableExplicit = #False
                result = minScriptCodeCompilerClass_HELPER_AddLocalVariable(*this, @*this\parseTree[1], #minScript_VARIANT)
                If result = #minScript_OKAY
                  Goto minScriptCodeCompilerEXClass_FOR_L1
                EndIf
              Else
                result = #minScript_VARIABLENOTFOUND            
              EndIf
            ElseIf *this\parseTree[1]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE Or *this\parseTree[1]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
              If *this\parseTree[1]\scriptVariable
minScriptCodeCompilerEXClass_LET_L2:
                opPopVar = #MS_OP_POPSCRIPTVAR
                *this\constructs()\opPushVar = #MS_OP_PUSHSCRIPTVAR
                *this\constructs()\varOperand = *this\parseTree[1]\scriptVariable
                If *this\parseTree[1]\scriptVariable\vt <> #minScript_INTEGER And *this\parseTree[1]\scriptVariable\vt & #minScript_VARIANT = 0
                  result = #minScript_TYPEMISMATCH
                EndIf
              ElseIf *this\code\blnEnableExplicit = #False
                If *this\code <> minScriptGlobalClassLib
                  result = minScriptCodeCompilerClass_HELPER_AddScriptVariable(*this, @*this\parseTree[1], #minScript_VARIANT)
                  If result = #minScript_OKAY
                    Goto minScriptCodeCompilerEXClass_LET_L2
                  EndIf
                Else
                    result = #minScript_ACCESSDENIED              
                EndIf
              Else
                result = #minScript_VARIABLENOTFOUND            
              EndIf
            Else
              result = #minScript_SYNTAXERROR
            EndIf
          If result = #minScript_OKAY
            If *this\parseTree[2]\symbol$ = "="
              ;Process the fromExpression.
              *this\expressionSymbolIndex = 3
              result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this, 0, "to")
              If result = #minScript_OKAY
                If *this\expressionSymbolIndex < *this\numParseSymbols - 1
                  *this\constructs()\forVar$ = *this\parseTree[1]\symbol$
                  If *this\thisEX\AddVirtualCodeLabel("for_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc())
                    *this\thisEX\EmitObjectCodeInstruction(opPopVar, *this\constructs()\varOperand)
                    ;Now we process the to Expression.                        
                    *this\expressionSymbolIndex + 1
                    result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this, 0)
                    If result = #minScript_OKAY
                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_CAST, #minScript_INTEGER)
                      *this\thisEX\EmitObjectCodeInstruction(*this\constructs()\opPushVar, *this\constructs()\varOperand)
                      If *this\constructs()\stepValue >= 0
                        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_LT)
                      Else
                        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_GT)
                      EndIf 
                      codeAddress = *this\thisEX\AddVirtualCodeLabel("next_" + Str(*this\constructs()\constructIndex1))
                      If codeAddress
                        *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMPNZ, codeAddress)
                      Else
                        result = #minScript_OUTOFMEMORY
                      EndIf                
                    EndIf
                  Else
                    result = #minScript_OUTOFMEMORY
                  EndIf
                Else
                  result = #minScript_SYNTAXERROR
                EndIf
              EndIf
            Else
              result = #minScript_SYNTAXERROR
            EndIf
          EndIf
        EndIf
      EndIf
    Else
      result = #minScript_SYNTAXERROR  
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_FUNCTION(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule=0)
  Protected result = #minScript_OKAY, blnIsPrivate, parameters.st_minScript_ParseParameters, numParameters, index, paramType.u, i
  Protected t1, *ptrChar.CHARACTER, skipIndexes
  ;Do we have a valid 'PRIVATE' declaration.
  If *this\parseTree[0]\symbol$ = "private"
    blnIsPrivate = #True
    If *this\numParseSymbols > 1 
      If *this\parseTree[1]\symbol$ <> "function"
        result = #minScript_INVALIDUSEOFPRIVATE
      EndIf
    Else
      result = #minScript_SYNTAXERROR
    EndIf  
  ElseIf *this\parseTree[0]\symbol$ = "method"
    ;Check the name is of the form class:method.
    If *this\numParseSymbols > 1 
      t1 = FindString(*this\parseTree[1]\symbol$, ":")
      If t1
        ;Switch the ':' for a '@'.
          *ptrChar = @*this\parseTree[1]\symbol$ + (t1-1)<<(SizeOf(CHARACTER)-1)
          *ptrChar\c = '@'
        PushMapPosition(*this\code\Modules())
        ;Need to check all modules (other than the current one) for an existing class method of the same name.
          ForEach *this\code\Modules()
            ForEach *this\code\Modules()\functions()
              If *this\code\Modules() <> *module And *this\code\Modules()\functions()\name$ = *this\parseTree[1]\symbol$
                result = #minScript_FUNCTIONORMETHODALREADYDEFINED
                Break 2
              EndIf
            Next          
          Next
        PopMapPosition(*this\code\Modules())
        ;We need to set up a hidden first parameter ('this') and we have to give it a minScript_VARIANT type because the class may not be defined at this point.
        ;Also, the execution engine does not promote parameters of a minScript_VARIANT type.
          parameters\varName$[0] = "this"
          parameters\vt[0] = #minScript_VARIANT|#minScript_REFERENCE
          numParameters = 1
      ElseIf *this\numParseSymbols > 4
        ;May have a class destructor which would be recorded as 4 different symbols.
        If *this\parseTree[2]\symbol$ = ":" And *this\parseTree[3]\symbol$ = "~" And *this\parseTree[4]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
          *this\parseTree[1]\symbol$ + "@~" + *this\parseTree[4]\symbol$
          *this\parseTree[1]\symbolType = #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
          skipIndexes = 3
        Else        
          result = #minScript_INVALIDFUNCTIONORMETHODDECLARATION
        EndIf
      Else
        result = #minScript_SYNTAXERROR
      EndIf
    Else
      result = #minScript_SYNTAXERROR
    EndIf  
  EndIf
  If result = #minScript_OKAY  
    If *this\numParseSymbols >= 4 + blnIsPrivate + skipIndexes ;The minimum number of symbols.
      ;Does the function already exist within this module?
      If *this\parseTree[1+blnIsPrivate]\scriptFunction = 0
        ;Check we do not have the name of a run-time function or a keyword etc.
        If *this\parseTree[1+blnIsPrivate]\symbolType = #minScript_COMPILER_SYMBOL_KEYWORD
          result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
        ElseIf *this\parseTree[1+blnIsPrivate]\symbolType <> #minScript_COMPILER_SYMBOL_SCRIPTEDFUNCTION
          result = #minScript_SYNTAXERROR
        ElseIf FindString(*this\parseTree[1+blnIsPrivate]\symbol$, ":") 
          ;A function name of the form 'moduleName:functionName' was given and whilst the names would have been checked for valid characters, the ':' could get through.
          result = #minScript_SYNTAXERROR
        Else ;Good thus far!
          ;Open parantheses will already have been checked. We check for the close parantheses.
          If *this\parseTree[*this\numParseSymbols - 1]\symbol$ <> ")"
            result = #minScript_INVALIDFUNCTIONORMETHODDECLARATION
          Else
            ;We store the info in a temporary static array in case of an error.
              index = 3 + blnIsPrivate + skipIndexes
              While (index < *this\numParseSymbols - 1) And result = #minScript_OKAY
                ;Do we have an entry of the form "var AS TYPE" ?
                If index < *this\numParseSymbols - 3
                  If *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE Or *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
                    If *this\parseTree[index+1]\symbol$ = "as"
                      paramType = minScriptCodeClass_HELPER_String2VT(*this\code, *this\parseTree[index+2]\symbol$)
                      If paramType
                        ;We cannot allow a user-defined variable to have the same name as run-time function because such variables
                        ;can be used to invoke certain 'methods'.
                        CompilerIf #PB_Compiler_Thread
                          LockMutex(minScript_Globals\threadMutex)
                        CompilerEndIf
                        If paramType&#minScript_REFERENCE And FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), *this\parseTree[index]\symbol$)                        
                          result = #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
                          Break
                        EndIf
                        CompilerIf #PB_Compiler_Thread
                          UnlockMutex(minScript_Globals\threadMutex)
                        CompilerEndIf
                      Else
                        result = #minScript_UNRECOGNISEDVARIABLETYPE
                        Break
                      EndIf
                      If index = *this\numParseSymbols - 4 Or *this\parseTree[index+3]\symbol$ = ","
                        ;Have we too many parameters.
                          If numParameters < #minScript_MAXNUMFUNCTIONPARAMETERS
                            ;Check the parameter has not already been added.
                              For i = 0 To numParameters-1
                                If parameters\varName$[i] = *this\parseTree[index]\symbol$
                                  result = #minScript_VARIABLEORPROPERTYALREADYDECLARED
                                  Break 2
                                EndIf
                              Next
                            ;Add the parameter to the temporary parameters array.
                              parameters\varName$[numParameters] = *this\parseTree[index]\symbol$
                              parameters\vt[numParameters] = paramType
                              numParameters + 1
                            index+4
                          Else
                            result = #minScript_TOOMANYPARAMETERS
                          EndIf
                      Else
                        result = #minScript_SYNTAXERROR
                      EndIf
                    Else
                      result = #minScript_SYNTAXERROR
                    EndIf
                  Else
                    ;Emit the most appropriate error.
                      Select *this\parseTree[index]\symbolType
                        Case #minScript_COMPILER_SYMBOL_KEYWORD
                          result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
                        Case #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
                          result = #minScript_INVALIDPARAMETER                
                        Default
                          result = #minScript_SYNTAXERROR
                      EndSelect
                  EndIf 
                Else
                  result = #minScript_SYNTAXERROR
                EndIf
              Wend
            ;If all is well then we can add the function to the underlying code object's module function map.
            If result = #minScript_OKAY
              *this\code\currentScriptFunction = AddMapElement(*module\functions(), *this\parseTree[1+blnIsPrivate]\symbol$)
              If *this\code\currentScriptFunction
                *this\code\currentScriptFunction\name$ = *this\parseTree[1+blnIsPrivate]\symbol$
                *this\code\currentScriptFunction\module = *module
                *this\code\currentScriptFunction\blnIsPrivate = blnIsPrivate
                *this\code\currentScriptFunction\numLocalVariables = #minScript_MAXNUMFUNCTIONPARAMETERS
                ;We write some initial object code to set up the stack frame.            
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHEBP)
                  *this\thisEX\EmitObjectCodeInstruction(#MS_OP_ESPTOEBP)
                ;Emit some object code for the undeclared parameters.
                  For i = #minScript_MAXNUMFUNCTIONPARAMETERS - numParameters To 1 Step -1
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_LOCALPARAM, #minScript_EMPTY)
                  Next 
                ;Now the declared parameters.
                  For i = numParameters-1 To 0 Step -1
                    If AddMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), parameters\varName$[i])              
                      *this\code\currentScriptFunction\localVariablesSymbols()\EBPoffset = i * SizeOf(minScriptVariant)
                      *this\code\currentScriptFunction\localVariablesSymbols()\vt = parameters\vt[i]
                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_LOCALPARAM, parameters\vt[i])
                    Else
                      result = #minScript_OUTOFMEMORY
                      Break
                    EndIf        
                  Next
              Else
                result = #minScript_OUTOFMEMORY
              EndIf
            EndIf  
          EndIf
        EndIf
      Else
        If *this\parseTree[1+blnIsPrivate]\symbolType = #minScript_COMPILER_SYMBOL_RUNTIMEFUNCTION
          result = #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
        Else
          result = #minScript_FUNCTIONORMETHODALREADYDEFINED
        EndIf
      EndIf  
    Else
      result = #minScript_INVALIDFUNCTIONORMETHODDECLARATION
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_IF(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeLabelAddress
  If *this\numParseSymbols > 1 ;The minimum number of symbols.
    ;Hand over to the expression engine.
      *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
      result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
    If result = #minScript_OKAY
      result = *this\thisEX\PushConstruct(#minScript_COMPILER_CONSTRUCT_IF)
      If result = #minScript_OKAY
        ;We need to set up a virtual jump label of the form 'endif_' + Str(*this\constructs()\constructIndex1).
        If *this\thisEX\AddVirtualCodeLabel("endif_" + Str(*this\constructs()\constructIndex1))
          ;We need to set up a virtual jump label of the form 'if_' + Str(*this\constructs()\constructIndex1) + "_0".
          codeLabelAddress = *this\thisEX\AddVirtualCodeLabel("if_" + Str(*this\constructs()\constructIndex1) + "_0")
          If codeLabelAddress
            ;Emit object code to bypass subsequent instructions if the 'If' clause is false.
              *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMPZ, codeLabelAddress)
          Else
            result = #minScript_OUTOFMEMORY
          EndIf
        Else
          result = #minScript_OUTOFMEMORY
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Constant assignments will already have been filtered out.
Procedure.i minScriptCodeCompilerEXClass_LET(*this.st_minScriptCodeCompilerClass, blnShortCutAssignment=0)
  Protected result = #minScript_OKAY, vt, i, j, numOpenBrackets, idxOpenBracketSymbol, idxCloseBracketSymbol, idxEqualsSymbol, *ptrChar.CHARACTER, t1
  If *this\numParseSymbols >= 4 - blnShortCutAssignment ;The minimum number of symbols.
    ;We first search for the first un-nested close bracket and unbracketed equals symbol.
      For i = 2-blnShortCutAssignment To *this\numParseSymbols - 1
        If *this\parseTree[i]\symbol$ = "("
          If numOpenBrackets = 0
            idxOpenBracketSymbol = i
          EndIf
          numOpenBrackets + 1
        ElseIf *this\parseTree[i]\symbol$ = ")"
          numOpenBrackets - 1
          If numOpenBrackets < 0 ;Why not!
            result = #minScript_SYNTAXERROR
            Break
          ElseIf numOpenBrackets = 0
            idxCloseBracketSymbol = i
          EndIf
        ElseIf *this\parseTree[i]\symbol$ = "="
          If idxEqualsSymbol = 0 And numOpenBrackets = 0
            idxEqualsSymbol = i
            Break
          EndIf
        EndIf
      Next
    ;Did we find a valid equals sign?
    If result = #minScript_OKAY
      If idxEqualsSymbol And idxEqualsSymbol < *this\numParseSymbols - 1
        ;Do we have a straight-forward 'LET variable = expression' construct?
        If idxEqualsSymbol = 2-blnShortCutAssignment
          ;Check for an initialised variable.
          Select *this\parseTree[1-blnShortCutAssignment]\symbolType
            Case #minScript_COMPILER_SYMBOL_LOCALVARIABLE, #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE, #minScript_COMPILER_SYMBOL_MODULEVARIABLE
              If *this\parseTree[1-blnShortCutAssignment]\localVariable ;Same field as \scriptVariable and \moduleVariable
minScriptCodeCompilerEXClass_LET_L1:
                ;Hand over to the expression engine.
                  *this\expressionSymbolIndex = 3-blnShortCutAssignment ;Start node of the expression in the parse tree.
                  If *this\parseTree[1-blnShortCutAssignment]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE
                    vt = *this\parseTree[1-blnShortCutAssignment]\localVariable\vt
                  Else
                    vt = *this\parseTree[1-blnShortCutAssignment]\scriptVariable\vt
                  EndIf
                  result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this, vt)
                If result = #minScript_OKAY
                  ;Emit code to pop the top of the variant stack to the relevant variable.            
                  If *this\parseTree[1-blnShortCutAssignment]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPLOCALVAR, *this\parseTree[1-blnShortCutAssignment]\localVariable\EBPoffset)
                  Else
                    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPSCRIPTVAR, *this\parseTree[1-blnShortCutAssignment]\scriptVariable)
                  EndIf
                EndIf
              ElseIf *this\code\blnEnableExplicit = #False 
              ;We attempt to create a new variant type variable on the fly.
                If *this\parseTree[1-blnShortCutAssignment]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE
                  result = minScriptCodeCompilerClass_HELPER_AddLocalVariable(*this, *this\parseTree[1-blnShortCutAssignment], #minScript_VARIANT)
                  If result = #minScript_OKAY
                    Goto minScriptCodeCompilerEXClass_LET_L1
                  EndIf
                Else ;Script variable.
                  If *this\code <> minScriptGlobalClassLib
                    result = minScriptCodeCompilerClass_HELPER_AddScriptVariable(*this, @*this\parseTree[1-blnShortCutAssignment], #minScript_VARIANT)
                    If result = #minScript_OKAY
                      Goto minScriptCodeCompilerEXClass_LET_L1
                    EndIf
                  Else
                    result = #minScript_ACCESSDENIED              
                  EndIf
                EndIf
              Else
                result = #minScript_VARIABLENOTFOUND            
              EndIf
            Case #minScript_COMPILER_SYMBOL_KEYWORD
              result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
            Default
              result = #minScript_SYNTAXERROR
          EndSelect
        Else
          ;A couple of permitted options here all of which involve the setting of a property.           
          ;First, nested object methods/properties finishing with a simple property set.
          If *this\parseTree[idxEqualsSymbol-1]\symbolType = #minScript_COMPILER_SYMBOL_OBJECTMETHOD And *this\parseTree[idxEqualsSymbol-1]\invokeType = #minScript_METHOD_GET
            ;Change to a property set.
              *this\parseTree[idxEqualsSymbol-1]\invokeType = #minScript_METHOD_SET
            ;The '=' symbol needs to be switched to an open bracket.
              *this\parseTree[idxEqualsSymbol]\symbol$ = "("
              *this\parseTree[idxEqualsSymbol]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
              *this\parseTree[idxEqualsSymbol]\typeModifierVT = 0
              *this\parseTree[idxEqualsSymbol]\invokeType = 0
            ;Add an additional ')' symbol.
              If *this\numParseSymbols < #minScript_COMPILER_SIZEOFPARSETREE
                *this\parseTree[*this\numParseSymbols]\symbol$ = ")"
                *this\parseTree[*this\numParseSymbols]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                *this\parseTree[*this\numParseSymbols]\typeModifierVT = 0
                *this\parseTree[*this\numParseSymbols]\invokeType = 0
                *this\numParseSymbols + 1
                ;Hand over to the expression engine to compile just a single operand. 
                ;This will allow us to determine whether what is otherwise a valid expression is in fact a valid property assignment
                ;as a valid assignment in this case will constitute a single operand (including a trailing series of invokations).
                  *this\expressionSymbolIndex = 1-blnShortCutAssignment ;Start node of the expression in the parse tree.
                  result = minScriptCodeCompilerClass_EXPRESSION_PushNextOperand(*this, 0, @t1)
                If result = #minScript_OKAY 
                  If *this\expressionSymbolIndex <> *this\numParseSymbols                  
                    result = #minscript_SYNTAXERROR
                  Else
                    ;Emit code to discard the top of the variant stack.            
                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POP, 1)
                  EndIf
                EndIf
              Else
                result = #minScript_STATEMENTTOOCOMPLEX
              EndIf
          Else
            ;Only allowable option now is for a close bracket immediately preceeding the = symbol.
            If idxOpenBracketSymbol And idxCloseBracketSymbol And (idxEqualsSymbol = idxCloseBracketSymbol + 1)
              If *this\parseTree[idxOpenBracketSymbol-1]\symbol$ = #minScript_DEFAULTPROPERTY$ And *this\parseTree[idxOpenBracketSymbol-1]\invokeType = #minScript_METHOD_GET
                ;Assignment must in the form : LET ...{#minScript_DEFAULTPROPERTY$}(...) = ...
                ;We need to carefully edit the parse tree to switch the assignment for a suitable property set.
                *this\parseTree[idxOpenBracketSymbol-1]\invokeType = #minScript_METHOD_SET
                ;Two possible cases.
                If idxCloseBracketSymbol = idxOpenBracketSymbol + 1 ;Empty parantheses.
                  ;Shunt all the remaining symbols down.
                  For j = idxCloseBracketSymbol To *this\numParseSymbols-3
                    With *this\parseTree[j]
                      \symbol$ = *this\parseTree[j+2]\symbol$
                      \symbolType = *this\parseTree[j+2]\symbolType
                      \localVariable = *this\parseTree[j+2]\localVariable
                      \typeModifierVT = *this\parseTree[j+2]\typeModifierVT
                      \invokeType = *this\parseTree[j+2]\invokeType
                    EndWith
                  Next
                  ;Reduce the number of symbols.
                    *this\numParseSymbols - 1
                Else ;Index parameters.
                  *this\parseTree[idxCloseBracketSymbol]\symbol$ = ","
                  *this\parseTree[idxCloseBracketSymbol]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                  ;Shunt all the remaining symbols down.
                    For j = idxEqualsSymbol To *this\numParseSymbols-2
                      With *this\parseTree[j]
                        \symbol$ = *this\parseTree[j+1]\symbol$
                        \symbolType = *this\parseTree[j+1]\symbolType
                        \localVariable = *this\parseTree[j+1]\localVariable
                        \typeModifierVT = *this\parseTree[j+1]\typeModifierVT
                        \invokeType = *this\parseTree[j+1]\invokeType
                      EndWith
                    Next
                EndIf
                ;The last symbol needs to be a close bracket.
                  *this\parseTree[*this\numParseSymbols-1]\symbol$ = ")"
                  *this\parseTree[*this\numParseSymbols-1]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                  *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = 0
                  *this\parseTree[*this\numParseSymbols-1]\invokeType = 0
                ;Hand over to the expression engine to compile just a single operand. 
                ;This will allow us to determine whether what is otherwise a valid expression is in fact a valid property assignment
                ;as a valid assignment in this case will constitute a single operand (including a trailing series of invokations).
                  *this\expressionSymbolIndex = 1-blnShortCutAssignment ;Start node of the expression in the parse tree.
                  result = minScriptCodeCompilerClass_EXPRESSION_PushNextOperand(*this, 0, @t1)
                If result = #minScript_OKAY 
                  If *this\expressionSymbolIndex <> *this\numParseSymbols                  
                    result = #minscript_SYNTAXERROR
                  Else
                    ;Emit code to discard the top of the variant stack.            
                      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POP, 1)
                  EndIf
                EndIf
              Else
                ;Assignment must in the form : LET ...method(...) = ...
                ;We replace the method with get_method and then introduce a set_IndexedValue.
                If *this\parseTree[idxOpenBracketSymbol-1]\symbolType = #minScript_COMPILER_SYMBOL_OBJECTMETHOD
                  *this\parseTree[idxOpenBracketSymbol-1]\invokeType = #minScript_METHOD_GET
                  ;Need to insert a '.' symbol and a 'set_IndexedValue'. How many items we need to add depends on whether there are any parameters.
                  If idxCloseBracketSymbol = idxOpenBracketSymbol + 1 ;Empty parantheses.
                    If *this\numParseSymbols < #minScript_COMPILER_SIZEOFPARSETREE
                      ;Modify a couple of the given parameters.
                        *this\parseTree[idxOpenBracketSymbol]\symbol$ = "."
                        *this\parseTree[idxOpenBracketSymbol]\typeModifierVT = 0
                        *this\parseTree[idxOpenBracketSymbol]\invokeType = 0
                        *this\parseTree[idxCloseBracketSymbol]\symbol$ = #minScript_DEFAULTPROPERTY$
                        *this\parseTree[idxCloseBracketSymbol]\symbolType = #minScript_COMPILER_SYMBOL_OBJECTMETHOD
                        *this\parseTree[idxCloseBracketSymbol]\typeModifierVT = 0
                        *this\parseTree[idxCloseBracketSymbol]\invokeType = #minScript_METHOD_SET
                        *this\parseTree[idxEqualsSymbol]\symbol$ = "("
                        *this\parseTree[idxEqualsSymbol]\typeModifierVT = 0
                        *this\parseTree[idxEqualsSymbol]\invokeType = 0
                      ;Need to add one additional item (a close bracket) to the parse tree.
                        *this\numParseSymbols + 1
                        *this\parseTree[*this\numParseSymbols-1]\symbol$ = ")"
                        *this\parseTree[*this\numParseSymbols-1]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                        *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = 0
                        *this\parseTree[*this\numParseSymbols-1]\invokeType = 0
                    Else
                      result = #minScript_STATEMENTTOOCOMPLEX
                    EndIf
                  Else ;Index parameters.
                    If *this\numParseSymbols + 1 < #minScript_COMPILER_SIZEOFPARSETREE
                      ;Need to add two additional items to the parse tree.
                        *this\numParseSymbols + 2
                        *this\parseTree[*this\numParseSymbols-1]\symbol$ = ")"
                        *this\parseTree[*this\numParseSymbols-1]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                        *this\parseTree[*this\numParseSymbols-1]\typeModifierVT = 0
                        *this\parseTree[*this\numParseSymbols-1]\invokeType = 0
                      ;Shunt some symbols up.
                        For j = *this\numParseSymbols-2 To idxEqualsSymbol + 2 Step -1 
                          With *this\parseTree[j]
                            \symbol$ = *this\parseTree[j-1]\symbol$
                            \symbolType = *this\parseTree[j-1]\symbolType
                            \localVariable = *this\parseTree[j-1]\localVariable
                            \typeModifierVT = *this\parseTree[j-1]\typeModifierVT
                            \invokeType = *this\parseTree[j-1]\invokeType
                          EndWith
                        Next
                        *this\parseTree[idxEqualsSymbol+1]\symbol$ = ","
                        *this\parseTree[idxEqualsSymbol+1]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                        *this\parseTree[idxEqualsSymbol+1]\typeModifierVT = 0
                        *this\parseTree[idxEqualsSymbol+1]\invokeType = 0
                        For j = idxEqualsSymbol To idxOpenBracketSymbol + 2 Step -1 
                          With *this\parseTree[j]
                            \symbol$ = *this\parseTree[j-2]\symbol$
                            \symbolType = *this\parseTree[j-2]\symbolType
                            \localVariable = *this\parseTree[j-2]\localVariable
                            \typeModifierVT = *this\parseTree[j-2]\typeModifierVT
                            \invokeType = *this\parseTree[j-2]\invokeType
                          EndWith
                        Next
                        *this\parseTree[idxOpenBracketSymbol]\symbol$ = "."
                        *this\parseTree[idxOpenBracketSymbol]\symbolType = #minScript_COMPILER_SYMBOL_SEPARATOR
                        *this\parseTree[idxOpenBracketSymbol]\typeModifierVT = 0
                        *this\parseTree[idxOpenBracketSymbol]\invokeType = 0
                        *this\parseTree[idxOpenBracketSymbol+1]\symbol$ = #minScript_DEFAULTPROPERTY$
                        *this\parseTree[idxOpenBracketSymbol+1]\symbolType = #minScript_COMPILER_SYMBOL_OBJECTMETHOD
                        *this\parseTree[idxOpenBracketSymbol+1]\typeModifierVT = 0
                        *this\parseTree[idxOpenBracketSymbol+1]\invokeType = #minScript_METHOD_SET
                    Else
                      result = #minScript_STATEMENTTOOCOMPLEX
                    EndIf
                  EndIf
                  If result = #minScript_OKAY
                    ;Hand over to the expression engine to compile just a single operand. 
                    ;This will allow us to determine whether what is otherwise a valid expression is in fact a valid property assignment
                    ;as a valid assignment in this case will constitute a single operand (including a trailing series of invokations).
                      *this\expressionSymbolIndex = 1-blnShortCutAssignment ;Start node of the expression in the parse tree.
                      result = minScriptCodeCompilerClass_EXPRESSION_PushNextOperand(*this, 0, @t1)
                    If result = #minScript_OKAY 
                      If *this\expressionSymbolIndex <> *this\numParseSymbols                  
                        result = #minscript_SYNTAXERROR
                      Else
                        ;Emit code to discard the top of the variant stack.            
                          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POP, 1)
                      EndIf
                    EndIf
                  EndIf
                Else
                  result = #minScript_SYNTAXERROR
                EndIf
              EndIf
            Else
              result = #minScript_SYNTAXERROR
            EndIf
          EndIf
        EndIf
      Else
        result = #minScript_SYNTAXERROR  
      EndIf
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_NEXT(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeAddress
  If *this\numParseSymbols = 1
    ;Check we are within a FOR/NEXT block.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_FOR
          *this\thisEX\EmitObjectCodeInstruction(*this\constructs()\opPushVar, *this\constructs()\varOperand)
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_PUSHINT, *this\constructs()\stepValue)
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_ADD)
          ;No need to check the following line for errors.
            codeAddress = *this\thisEX\AddVirtualCodeLabel("for_" + Str(*this\constructs()\constructIndex1))
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeAddress)
          ;No need to check the following line for errors.
            *this\thisEX\AddVirtualCodeLabel("next_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc())
          *this\thisEX\PopConstruct()
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_POKEYWORD(*this.st_minScriptCodeCompilerClass, address)
  Protected result = #minScript_OKAY
  ;Two options; either no paramaters passed or a single parameter.
  If *this\numParseSymbols = 1 ;No parameters.
    *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POKEYWORD0, address)
  Else
    ;We hand over to the expression engine.
      *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
      result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
    If result = #minScript_OKAY
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POKEYWORD1, address)
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_PROPERTY(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, propType.u, blnIsPrivate
  If *this\numParseSymbols = 3 
    blnIsPrivate = #True
  ElseIf *this\numParseSymbols <> 4 Or *this\parseTree[0]\symbol$ <> "property"
    result = #minScript_SYNTAXERROR
  EndIf
  If result = #minScript_OKAY
    If *this\parseTree[*this\numParseSymbols-2]\symbol$ = "as"
      ;Sort out the variable type.
        propType = minScriptCodeClass_HELPER_String2VT(*this\code, *this\parseTree[*this\numParseSymbols-1]\symbol$)
      If propType
        ;We need to examine the property declaration.
        ;Make sure the parser has recorded either a local or module variable.
        If *this\parseTree[1-blnIsPrivate]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE Or *this\parseTree[1-blnIsPrivate]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
          ;Check we haven't already a property of the given name.
          If FindMapElement(*this\currentClass\propertyInfo(), *this\parseTree[1-blnIsPrivate]\symbol$) = 0
            ;Add a new property.
            If AddMapElement(*this\currentClass\propertyInfo(), *this\parseTree[1-blnIsPrivate]\symbol$)
              *this\currentClass\propertyInfo()\isPrivateProperty = blnIsPrivate
              *this\currentClass\propertyInfo()\vt = propType
              *this\currentClass\numProperties+1
            Else
              result = #minScript_OUTOFMEMORY            
            EndIf
          ElseIf *this\currentClass\propertyInfo()\vt <> propType
            result = #minScript_VARIABLEORPROPERTYALREADYDECLARED
          EndIf
        Else ;Set an appropriate error code.
          Select *this\parseTree[1-blnIsPrivate]\symbolType
            Case #minScript_COMPILER_SYMBOL_KEYWORD
              result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
            Case #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
              result = #minScript_INVALIDPARAMETER                
            Default
              result = #minScript_SYNTAXERROR
          EndSelect
        EndIf
      Else
        result = #minScript_UNRECOGNISEDVARIABLETYPE
      EndIf
    Else
      result = #minScript_SYNTAXERROR    
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_REPEAT(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols = 1
    result = *this\thisEX\PushConstruct(#minScript_COMPILER_CONSTRUCT_REPEAT)
    If result = #minScript_OKAY
      If *this\thisEX\AddVirtualCodeLabel("repeat_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc()) = 0
        result = #minScript_OUTOFMEMORY
      EndIf
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_RETURN(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols > 1
    ;Hand over to the expression compiler.
    *this\expressionSymbolIndex = 1
    result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
    If result = #minScript_OKAY
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPEAX)
    EndIf
  EndIf
  If result = #minScript_OKAY
    ;Emit code to tidy up the stack frame.
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_EBPTOESP)
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_POPEBP)
      *this\thisEX\EmitObjectCodeInstruction(#MS_OP_RET)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_SELECT(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY
  If *this\numParseSymbols > 1 ;The minimum number of symbols.
    ;Hand over to the expression engine.
      *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
      result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
    If result = #minScript_OKAY
      result = *this\thisEX\PushConstruct(#minScript_COMPILER_CONSTRUCT_SELECT)
      If result = #minScript_OKAY
        ;We need to set up a virtual jump label of the form 'endSelect_' + Str(*this\constructs()\constructIndex1).
        If *this\thisEX\AddVirtualCodeLabel("endselect_" + Str(*this\constructs()\constructIndex1)) = 0
          result = #minScript_OUTOFMEMORY
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  *this\blnLastStatementASelect = #True
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_STATIC(*this.st_minScriptCodeCompilerClass, *module.st_minScriptModule)
  Protected result = #minScript_OKAY, varType.u, index
  If *this\numParseSymbols >= 4 ;The minimum number of symbols.
    If *this\parseTree[*this\numParseSymbols-2]\symbol$ = "as"
      ;Sort out the variable type.
        varType = minScriptCodeClass_HELPER_String2VT(*this\code, *this\parseTree[*this\numParseSymbols-1]\symbol$)
        If varType = 0  
          result = #minScript_UNRECOGNISEDVARIABLETYPE
        EndIf
      If result = #minScript_OKAY
        ;We need to examine each variable declaration.
        index = 1
        While index < *this\numParseSymbols-2
          ;Make sure we have a local variable. Undeclared module variables will be recorded, at this stage, as local variables.
          If *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_LOCALVARIABLE Or *this\parseTree[index]\symbolType = #minScript_COMPILER_SYMBOL_MODULEVARIABLE
            ;Check variable is not already defined. Rather than use any existing value of the parseTree *localVariable field,
            ;which may not help in this instance if this DIM statement contains repeated declarations, we turn to the local symbols map instead.
              *this\parseTree[index]\localVariable = FindMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *this\parseTree[index]\symbol$)
              If *this\parseTree[index]\localVariable = 0
                ;In the case of a user-defined type, we cannot allow the variable to have the same name as run-time function because such variables
                ;can be used to invoke certain 'methods'.
                  CompilerIf #PB_Compiler_Thread
                    LockMutex(minScript_Globals\threadMutex)
                  CompilerEndIf
                    If varType & #minScript_REFERENCE = 0 Or FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), *this\parseTree[index]\symbol$) = 0
                      ;Create a new local variable symbol.
                        *this\parseTree[index]\localVariable = AddMapElement(*this\code\currentScriptFunction\localVariablesSymbols(), *this\parseTree[index]\symbol$)
                        If *this\parseTree[index]\localVariable 
                          *this\parseTree[index]\localVariable\vt = varType
                          ;Create a new module variable.
                            *this\parseTree[index]\localVariable\staticModuleVar = AddMapElement(*module\ModuleVariables(), "s_" + *this\code\currentScriptFunction\name$ + ":" + *this\parseTree[index]\symbol$)
                            If *this\parseTree[index]\localVariable\staticModuleVar
                              *this\parseTree[index]\localVariable\staticModuleVar\vt = varType
                            Else
                              result = #minScript_OUTOFMEMORY                  
                            EndIf
                        Else
                          result = #minScript_OUTOFMEMORY                  
                        EndIf
                    Else
                      result = #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
                    EndIf
                  CompilerIf #PB_Compiler_Thread
                    UnlockMutex(minScript_Globals\threadMutex)
                  CompilerEndIf
              ElseIf *this\parseTree[index]\localVariable\vt <> varType Or *this\parseTree[index]\localVariable\staticModuleVar = 0
                result = #minScript_VARIABLEORPROPERTYALREADYDECLARED
              EndIf
          Else ;Set an appropriate error code.
            Select *this\parseTree[index]\symbolType
              Case #minScript_COMPILER_SYMBOL_SCRIPTVARIABLE
                result = #minScript_INVALIDPARAMETER
              Case #minScript_COMPILER_SYMBOL_KEYWORD
                result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
              Default
                result = #minScript_SYNTAXERROR
            EndSelect
          EndIf
          If result = #minScript_OKAY
            If index = *this\numParseSymbols - 3
              Break
            Else ;Next symbol needs to be a comma.
              If *this\parseTree[index+1]\symbol$ = "," And index < *this\numParseSymbols - 4
                index + 2
              Else
                result = #minScript_SYNTAXERROR
                Break
              EndIf
            EndIf
          Else
            Break
          EndIf
        Wend
      EndIf
    Else
      result = #minScript_SYNTAXERROR    
    EndIf
  Else
    result = #minScript_SYNTAXERROR
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_UNTIL(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeAddress
  If *this\numParseSymbols > 1
    ;Check we are within a FOR/NEXT block.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_REPEAT
        ;Hand over to the expression engine.
          *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
          result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
        If result = #minScript_OKAY
          ;No need to check the following line for errors.
            codeAddress = *this\thisEX\AddVirtualCodeLabel("repeat_" + Str(*this\constructs()\constructIndex1))
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMPZ, codeAddress)
          If *this\thisEX\AddVirtualCodeLabel("until_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc()) = 0
            result = #minScript_OUTOFMEMORY
          EndIf
          *this\thisEX\PopConstruct()
        EndIf
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_WEND(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeAddress
  If *this\numParseSymbols = 1
    ;Check we are within an WHILE/WEND block.
    If ListSize(*this\constructs())
      If *this\constructs()\type = #minScript_COMPILER_CONSTRUCT_WHILE
          ;No need to check the following line for errors.
            codeAddress = *this\thisEX\AddVirtualCodeLabel("while_" + Str(*this\constructs()\constructIndex1))
          *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMP, codeAddress)
          If *this\thisEX\AddVirtualCodeLabel("wend_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc()) = 0
            result = #minScript_OUTOFMEMORY
          EndIf
          *this\thisEX\PopConstruct()
      Else
        result = #minScript_UNEXPECTEDENDCONSTRUCT
      EndIf
    Else
      result = #minScript_UNEXPECTEDENDCONSTRUCT
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_WHILE(*this.st_minScriptCodeCompilerClass)
  Protected result = #minScript_OKAY, codeAddress
  If *this\numParseSymbols > 1
    result = *this\thisEX\PushConstruct(#minScript_COMPILER_CONSTRUCT_WHILE)
    If result = #minScript_OKAY
      If *this\thisEX\AddVirtualCodeLabel("while_" + Str(*this\constructs()\constructIndex1), *this\memoryFile\Loc())
        ;Hand over to the expression engine.
          *this\expressionSymbolIndex = 1 ;Start node of the expression in the parse tree.
          result = minScriptCodeCompilerClass_EXPRESSION_Compile(*this)
        If result = #minScript_OKAY
          codeAddress = *this\thisEX\AddVirtualCodeLabel("wend_" + Str(*this\constructs()\constructIndex1))
          If codeAddress
            *this\thisEX\EmitObjectCodeInstruction(#MS_OP_JMPZ, codeAddress)
          Else
            result = #minScript_OUTOFMEMORY
          EndIf                
        EndIf
      Else
        result = #minScript_OUTOFMEMORY
      EndIf
    EndIf
  Else
    result = #minScript_SYNTAXERROR  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-Code label METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;No error checking is required because all virtual code labels are internally generated.
Procedure.i minScriptCodeCompilerEXClass_AddVirtualCodeLabel(*this.st_minScriptCodeCompilerClass, label$, objectCodeOffset=-1)
  Protected result
  ;If the code label has already been added then we alter the value only if the given offset <> -1.
  result = FindMapElement(*this\codeLabels(), label$)
  If result
    If objectCodeOffset <> -1
      *this\codeLabels() = objectCodeOffset
    EndIf
  Else
    result = AddMapElement(*this\codeLabels(), label$)
    If result
      *this\codeLabels() = objectCodeOffset
    EndIf
  EndIf
  CompilerIf Defined(minScript_OUTPUTASMTODEBUGGER, #PB_Constant)
    ;In the case that a virtual label is being properly defined, we can output a suitable debug command.
    If objectCodeOffset <> -1
      minScriptCodeCompilerClass_FN_OutputDebugASM(*this, "." + label$)
    EndIf
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;No error checking is required because all virtual code labels are internally generated.
Procedure minScriptCodeCompilerEXClass_FixUpCodeReferences(*this.st_minScriptCodeCompilerClass, *objectCodeBufferPtr.st_minScript_memoryPointer)
  Protected memSize, i, *ptrOperand.INTEGER, *ptrCodeLabel.INTEGER
  memSize = MemorySize(*objectCodeBufferPtr)
  For i = 1 To memSize/#minScript_SIZEOFOBJECTCODEINSTRUCTION
    ;Do we have one of the jump mnemonics?
    If *objectCodeBufferPtr\w >= #MS_OP_JMP And *objectCodeBufferPtr\w <= #MS_OP_JMPZ
      ;The temporary operand in this instruction points to the relevant entry in the code labels map.
      *ptrOperand = *objectCodeBufferPtr + SizeOf(WORD)
      If *ptrOperand ;Just in case.
        *ptrCodeLabel = *ptrOperand\i
        ;Now we replace the temporary operand with the actual code label offset.
          *ptrOperand\i = *ptrCodeLabel\i      
      EndIf
    EndIf
    *objectCodeBufferPtr + #minScript_SIZEOFOBJECTCODEINSTRUCTION
  Next
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-Construct METHODS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_PopConstruct(*this.st_minScriptCodeCompilerClass)
  Protected result
  If ListSize(*this\constructs())
    DeleteElement(*this\constructs()) 
    result = #True
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeCompilerEXClass_PushConstruct(*this.st_minScriptCodeCompilerClass, constructType)
  Protected result = #minScript_OKAY
  If AddElement(*this\constructs())
    *this\numTotalConstructs + 1
    *this\constructs()\type = constructType
    *this\constructs()\constructIndex1 = *this\numTotalConstructs
  Else
    result = #minScript_OUTOFMEMORY
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



DataSection 
  VTable_minScriptCodeCompilerClass: 
    Data.i @minScriptCodeCompilerClass_Destroy()
    Data.i @minScriptCodeCompilerClass_Reset()
    Data.i @minScriptCodeCompilerClass_CompileCode()
    Data.i @minScriptCodeCompilerEXClass_EmitObjectCodeInstruction()
    Data.i @minScriptCodeCompilerEXClass_BREAK()
    Data.i @minScriptCodeCompilerEXClass_CASE()
    Data.i @minScriptCodeCompilerEXClass_CLASS()
    Data.i @minScriptCodeCompilerEXClass_DEBUG()
    Data.i @minScriptCodeCompilerEXClass_DEFAULT()
    Data.i @minScriptCodeCompilerEXClass_DEFINE()
    Data.i @minScriptCodeCompilerEXClass_DIM()
    Data.i @minScriptCodeCompilerEXClass_DIRECTIVE()
    Data.i @minScriptCodeCompilerEXClass_ELSE()
    Data.i @minScriptCodeCompilerEXClass_ELSEIF()
    Data.i @minScriptCodeCompilerEXClass_ENDCLASS()
    Data.i @minScriptCodeCompilerEXClass_ENDFUNCTION()
    Data.i @minScriptCodeCompilerEXClass_ENDIF()
    Data.i @minScriptCodeCompilerEXClass_ENDSELECT()
    Data.i @minScriptCodeCompilerEXClass_FOR()
    Data.i @minScriptCodeCompilerEXClass_FUNCTION()
    Data.i @minScriptCodeCompilerEXClass_IF()
    Data.i @minScriptCodeCompilerEXClass_LET()
    Data.i @minScriptCodeCompilerEXClass_NEXT()
    Data.i @minScriptCodeCompilerEXClass_POKEYWORD()
    Data.i @minScriptCodeCompilerEXClass_PROPERTY()
    Data.i @minScriptCodeCompilerEXClass_REPEAT()
    Data.i @minScriptCodeCompilerEXClass_RETURN()
    Data.i @minScriptCodeCompilerEXClass_SELECT()
    Data.i @minScriptCodeCompilerEXClass_STATIC()
    Data.i @minScriptCodeCompilerEXClass_UNTIL()
    Data.i @minScriptCodeCompilerEXClass_WEND()
    Data.i @minScriptCodeCompilerEXClass_WHILE()
    Data.i @minScriptCodeCompilerEXClass_AddVirtualCodeLabel()
    Data.i @minScriptCodeCompilerEXClass_FixUpCodeReferences()
    Data.i @minScriptCodeCompilerEXClass_PopConstruct()
    Data.i @minScriptCodeCompilerEXClass_PushConstruct()
EndDataSection

; IDE Options = PureBasic 5.62 (Windows - x86)
; ExecutableFormat = Shared dll
; CursorPosition = 2910
; FirstLine = 2895
; Folding = ------------
; Markers = 894
; Executable = d.exe.dll.exe.dll