﻿CompilerIf Defined(INCLUDE_threadStorage, #PB_Constant)=0
#INCLUDE_threadStorage=1

;/////////////////////////////////////////////////////////////////////////////////
;threadStorage.
;==============
;
;©Big Eared Software 2017.
;=========================
;   Created with Purebasic 5.60 for Windows.
;
;   Platforms:  ALL.
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;   i)    Dynamic creation of thread local storage.
;   ii)   The creation and deletion functions CreateThreadStorage()/DeleteThreadStorage() should really only be invoked from the main process.
;         If they might be invoked from multiple threads then Mutex protection should be employed.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-MACROS.
;========
  ;The following allows us to define dynamic variables of type 'ThreadStorageIndex' for convenience.
    Macro ThreadStorageIndex
      i
    EndMacro
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS and STRUCTURES.
;==========================
  ;The following constant determines how many possible indexes can be in use at any given time (this does not limit the number of threads).
  ;Change to a suitable value as appropriate.
    #THREADSTORAGE_MAXNUMUSEDINDEXES = 1000
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBAL and THREADED VARIABLES.
;===============================
  Global Dim threadStorage_gUsedIndexes.i(#THREADSTORAGE_MAXNUMUSEDINDEXES)
  Threaded Dim threadStorage_tIndexes.i(#THREADSTORAGE_MAXNUMUSEDINDEXES)
;/////////////////////////////////////////////////////////////////////////////////


;-FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following creates a new thread storage index.
;Returns 0 if an error such as no more indexes available.
Procedure.i CreateThreadStorage()
  Protected index, i
  ;Need to find the first free index.
  For i = 1 To #THREADSTORAGE_MAXNUMUSEDINDEXES
    If threadStorage_gUsedIndexes(i) = 0
      ;Mark it as used.    
        threadStorage_gUsedIndexes(i) = 1
      index = i
      Break
    EndIf
  Next
  ProcedureReturn index
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following frees a thread storage index for reuse.
Procedure DeleteThreadStorage(index)
  If index >0 And index <= #THREADSTORAGE_MAXNUMUSEDINDEXES
    threadStorage_tIndexes(index) = 0
    threadStorage_gUsedIndexes(index) = 0
  EndIf
  ProcedureReturn index
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following retrieves the value associated with the given 1-based thread storage index for the current thread.
Procedure.i GetThreadStorageValue(index)
  If index >0 And index <= #THREADSTORAGE_MAXNUMUSEDINDEXES
    ;Only proceed if the index is in use.
    If threadStorage_gUsedIndexes(index)
      ProcedureReturn threadStorage_tIndexes(index)
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following sets a value against the given 1-based thread storage index for the current thread.
Procedure SetThreadStorageValue(index, value)
  If index >0 And index <= #THREADSTORAGE_MAXNUMUSEDINDEXES
    ;Only proceed if the index is in use.
    If threadStorage_gUsedIndexes(index)
      threadStorage_tIndexes(index) = value
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

CompilerEndIf

; IDE Options = PureBasic 5.60 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 109
; FirstLine = 76
; Folding = -
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode