﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Global Constants Module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;   i)  Constant names must begin with a '#' character.
;/////////////////////////////////////////////////////////////////////////////////

;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_InitStandardGlobalConstants()
  Protected result = #minScript_OKAY
  result + minScript_AddIntegerGlobalConstant("#ARRAY", #minScript_ARRAY)
  result + minScript_AddIntegerGlobalConstant("#ASCII", #PB_Ascii)
  result + minScript_AddIntegerGlobalConstant("#BYTE", #minScript_BYTE)
  result + minScript_AddIntegerGlobalConstant("#CHARACTER", #minScript_CHARACTER)
  result + minScript_AddIntegerGlobalConstant("#CURRENCY", #minScript_CURRENCY)
  result + minScript_AddIntegerGlobalConstant("#DOUBLE", #minScript_REAL)
  result + minScript_AddIntegerGlobalConstant("#EMPTY", #minScript_EMPTY)
  result + minScript_AddIntegerGlobalConstant("#FALSE", 0)
  result + minScript_AddIntegerGlobalConstant("#FLOAT", #minScript_FLOAT)
  result + minScript_AddIntegerGlobalConstant("#INTEGER", #minScript_INTEGER)
  result + minScript_AddIntegerGlobalConstant("#LONG", #minScript_LONG)
  result + minScript_AddIntegerGlobalConstant("#NOCASE", #PB_String_NoCase)
  result + minScript_AddIntegerGlobalConstant("#REAL", #minScript_REAL)
  result + minScript_AddIntegerGlobalConstant("#STRING", #minScript_STRING)
  result + minScript_AddIntegerGlobalConstant("#TRUE", 1)
  result + minScript_AddIntegerGlobalConstant("#UNICODE", #PB_Unicode)
  result + minScript_AddIntegerGlobalConstant("#UTF8", #PB_UTF8)
  result + minScript_AddIntegerGlobalConstant("#VARIANT", #minScript_VARIANT)
  result + minScript_AddIntegerGlobalConstant("#WORD", #minScript_WORD)
  ;Error constants - intended for use with the THROWERROR() runtime function.
    result + minScript_AddIntegerGlobalConstant("#ERROR_UNEXPECTEDERROR", #minScript_UNEXPECTEDERROR)
    result + minScript_AddIntegerGlobalConstant("#ERROR_OUTOFMEMORY", #minScript_OUTOFMEMORY)
    result + minScript_AddIntegerGlobalConstant("#ERROR_INVALIDPARAMETER", #minScript_INVALIDPARAMETER)
    result + minScript_AddIntegerGlobalConstant("#ERROR_INCORRECTNUMBEROFPARAMETERS", #minScript_INCORRECTNUMBEROFPARAMETERS)
    result + minScript_AddIntegerGlobalConstant("#ERROR_INCOMPATIBLEDATATYPES", #minScript_INCOMPATIBLEDATATYPES)
    result + minScript_AddIntegerGlobalConstant("#ERROR_INVALIDDATATYPE", #minScript_INVALIDDATATYPE)
    result + minScript_AddIntegerGlobalConstant("#ERROR_ACCESSDENIED", #minScript_ACCESSDENIED)
    result + minScript_AddIntegerGlobalConstant("#ERROR_OBJECTNOTINITIALIZED", #minScript_OBJECTNOTINITIALIZED)
    result + minScript_AddIntegerGlobalConstant("#ERROR_TYPEMISMATCH", #minScript_TYPEMISMATCH)
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following helper function adds a new entry to the global constants map and is called whenever a new constant is being added.
;It over-rides any existing constant of the same name.
Procedure.i minScript_GlobalConstants_HELPER_AddGlobalConstant(name$)
  Protected result = #minScript_OKAY, *ptr.CHARACTER
  If name$
    ;Check the constant name and prepend it with a # character if required.
      *ptr = @name$
      If *ptr\c <> '#'
        name$ = "#" + name$
        *ptr = @name$
      EndIf
      *ptr + SizeOf(CHARACTER)
      result = minScript_SymbolsIsSymbolNameValid(*ptr)
      If result = #minScript_OKAY
        ;Add the constant to the minScriptGlobalConstants map.
          If AddMapElement(minScript_Globals\minScriptGlobalConstants(), LCase(name$))
            minScript_Globals\minScriptGlobalConstants()\friendlyName$ = name$
          Else
            result = #minScript_OUTOFMEMORY           
          EndIf
      EndIf
  Else
    result = #minScript_INVALIDSYMBOLNAME
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-PUBLIC FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_AddCurrencyGlobalConstant(name$, value.q)
  Protected result = #minScript_OKAY
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ;Add a new entry to the global constants map.
    result = minScript_GlobalConstants_HELPER_AddGlobalConstant(name$)
  If result = #minScript_OKAY
    minScript_Globals\minScriptGlobalConstants()\value\vt = #minScript_CURRENCY
    minScript_Globals\minScriptGlobalConstants()\value\c = value
  EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_AddIntegerGlobalConstant(name$, value)
  Protected result = #minScript_OKAY
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ;Add a new entry to the global constants map.
    result = minScript_GlobalConstants_HELPER_AddGlobalConstant(name$)
  If result = #minScript_OKAY
    minScript_Globals\minScriptGlobalConstants()\value\vt = #minScript_INTEGER
    minScript_Globals\minScriptGlobalConstants()\value\i = value
  EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_AddRealGlobalConstant(name$, value.d)
  Protected result = #minScript_OKAY
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ;Add a new entry to the global constants map.
    result = minScript_GlobalConstants_HELPER_AddGlobalConstant(name$)
  If result = #minScript_OKAY
    minScript_Globals\minScriptGlobalConstants()\value\vt = #minScript_REAL
    minScript_Globals\minScriptGlobalConstants()\value\r = value
  EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_AddStringGlobalConstant(name$, value$)
  Protected result = #minScript_OKAY
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ;Add a new entry to the global constants map.
    result = minScript_GlobalConstants_HELPER_AddGlobalConstant(name$)
  If result = #minScript_OKAY
    minScript_Globals\minScriptGlobalConstants()\value\vt = #minScript_STRING
    minScript_Globals\minScriptGlobalConstants()\value$ = value$
  EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Returns a tab separated list of constant IDs.
Procedure.s minScript_GetListOfGlobalConstants()
  Protected result$
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ForEach minScript_Globals\minScriptGlobalConstants()
    result$ + minScript_Globals\minScriptGlobalConstants()\friendlyName$ + #TAB$  
  Next
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 42
; FirstLine = 21
; Folding = --
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode