﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Symbols Module.
;
;© Big Eared Software 2016.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;/////////////////////////////////////////////////////////////////////////////////


Declare.i minScript_SymbolsIsNegativeInteger(*ptrChar.CHARACTER, *ptrVal.QUAD = 0)
Declare.i minScript_SymbolsIsPositiveInteger(*ptrChar.CHARACTER, *ptrVal.QUAD = 0)
Declare.i minScript_SymbolsIsNegativeReal(*ptrChar.CHARACTER, *ptrVal.DOUBLE = 0, *ptrDecDivider.QUAD=0)
Declare.i minScript_SymbolsIsPositiveReal(*ptrChar.CHARACTER, *ptrVal.DOUBLE = 0, *ptrDecDivider.QUAD=0)


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents a hexadecimal number.
;If it does then the value of the integer is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
Procedure.i minScript_SymbolsIsHexadecimal(*ptrChar.CHARACTER, *ptrVal.INTEGER = 0)
  Protected result = #minScript_OKAY, value, numDigits
  If *ptrVal
    *ptrVal\i = 0
  EndIf
  While *ptrChar\c
    Select *ptrChar\c
      Case '0' To '9'
        numDigits + 1
        value * 16
        value + *ptrChar\c - '0'
      Case 'a' To 'f'
        numDigits + 1
        value * 16
        value + *ptrChar\c - 'a' + 10
      Case 'A' To 'F'
        numDigits + 1
        value * 16
        value + *ptrChar\c - 'A' + 10
      Default
        result = #minScript_INVALIDPARAMETER
        Break
    EndSelect
    *ptrChar + SizeOf(CHARACTER)
  Wend
  If result = #minScript_OKAY 
    If numDigits = 0
      result = #minScript_INVALIDPARAMETER
    ElseIf *ptrVal
      *ptrVal\i = value
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents either a negative or a positive integer.
;If it does then the value of the integer is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
;Note that only one '-' character is allowed and no + symbols etc.
Procedure.i minScript_SymbolsIsInteger(*ptrChar.CHARACTER, *ptrVal.INTEGER = 0)
  Protected result = #minScript_OKAY, value.i, blnNegative
  If *ptrVal
    *ptrVal\i = 0
  EndIf
  If *ptrChar\c = '-'
    blnNegative = #True
    *ptrChar + SizeOf(CHARACTER)
  EndIf
  result = minScript_SymbolsIsPositiveInteger(*ptrChar, @value)
  If result = #minScript_OKAY And *ptrVal
    If blnNegative
      *ptrVal\i = -value
    Else
      *ptrVal\i = value
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents a negative integer.
;If it does then the value of the integer is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
;Note that only one '-' character is allowed and no mixture of -+ etc.
Procedure.i minScript_SymbolsIsNegativeInteger(*ptrChar.CHARACTER, *ptrVal.INTEGER = 0)
  Protected result = #minScript_OKAY, value.i
  If *ptrVal
    *ptrVal\i = 0
  EndIf
  If *ptrChar\c = '-'
    *ptrChar + SizeOf(CHARACTER)
    result = minScript_SymbolsIsPositiveInteger(*ptrChar, @value)
    If result = #minScript_OKAY And *ptrVal
      *ptrVal\i = -value
    EndIf
  Else
    result = #minScript_INVALIDPARAMETER
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents a positive integer.
;If it does then the value of the integer is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
;Note that this function does not allow the symbol to be prepended with a '+' character.
Procedure.i minScript_SymbolsIsPositiveInteger(*ptrChar.CHARACTER, *ptrVal.INTEGER = 0)
  Protected result = #minScript_OKAY, numDigits, value.i, *tempPtr.CHARACTER
  *tempPtr = *ptrChar
  If *ptrVal
    *ptrVal\i = 0
  EndIf
  While *ptrChar\c
    Select *ptrChar\c
      Case '0' To '9'
        numDigits + 1
        value * 10
        value + *ptrChar\c - '0'
      Default
        result = #minScript_INVALIDPARAMETER
        Break
    EndSelect
    *ptrChar + SizeOf(CHARACTER)
  Wend
  If result = #minScript_OKAY 
    If numDigits = 0
      result = #minScript_INVALIDPARAMETER
    ElseIf *ptrVal
      *ptrVal\i = value
    EndIf
  Else ;Try a hexadecimal number.
    If *tempPtr\c = '0'
      *tempPtr + SizeOf(CHARACTER)
      If *tempPtr\c = 'x' Or *tempPtr\c = 'X'
        result = minScript_SymbolsIsHexadecimal(*tempPtr + SizeOf(CHARACTER), *ptrVal)
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents a negative real.
;If it does then the value of the real is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
;Note that only one '-' character is allowed and no mixture of -+ etc.
Procedure.i minScript_SymbolsIsNegativeReal(*ptrChar.CHARACTER, *ptrVal.DOUBLE = 0, *ptrblnIsCurrency.INTEGER=0)
  Protected result = #minScript_OKAY, value.d
  If *ptrVal
    *ptrVal\d = 0
  EndIf
  If *ptrChar\c = '-'
    *ptrChar + SizeOf(CHARACTER)
    result = minScript_SymbolsIsPositiveReal(*ptrChar, @value, *ptrblnIsCurrency)
    If result = #minScript_OKAY And *ptrVal
      *ptrVal\d = -value
    EndIf
  Else
    result = #minScript_INVALIDPARAMETER
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents a positive real.
;If it does then the value of the real is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
;Note that this function does not allow the symbol to be prepended with a '+' character.
Procedure.i minScript_SymbolsIsPositiveReal(*ptrChar.CHARACTER, *ptrVal.DOUBLE = 0, *ptrblnIsCurrency.INTEGER=0)
  Protected result = #minScript_OKAY, numDigits, intPart.q, dblPart.q, decimalDivider.q, blnDecimalPointFound
  If *ptrVal
    *ptrVal\d = 0
  EndIf
  decimalDivider = 1
  While *ptrChar\c
    Select *ptrChar\c
      Case '.'
        If blnDecimalPointFound
          result = #minScript_INVALIDPARAMETER
          Break
        Else
          blnDecimalPointFound = #True
        EndIf
      Case '0' To '9'
        numDigits + 1
        If blnDecimalPointFound
          dblPart * 10        
          dblPart + *ptrChar\c - '0'
          decimalDivider * 10
        Else
          intPart * 10        
          intPart + *ptrChar\c - '0'
        EndIf
      Default
        result = #minScript_INVALIDPARAMETER
        Break
    EndSelect
    *ptrChar + SizeOf(CHARACTER)
  Wend
  If result = #minScript_OKAY
    If numDigits = 0
      result = #minScript_INVALIDPARAMETER
    Else
      ;We might be able to reduce the value of decimalDivider by cancelling down.
        While decimalDivider > 1
          If 10*(dblPart / 10) = dblPart
            decimalDivider / 10
            dblPart / 10
          Else
            Break
          EndIf
        Wend
      If *ptrVal
        *ptrVal\d = intPart + dblPart/decimalDivider
      EndIf
      If *ptrblnIsCurrency And decimalDivider <= 10000
        *ptrblnIsCurrency\i = #True
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol to see if it represents either a negative or a positive real.
;If it does then the value of the double is placed within the given buffer.
;Returns either #minScript_OKAY or #minScript_INVALIDPARAMETER.
;Note that only one '-' character is allowed and no + symbols etc.
Procedure.i minScript_SymbolsIsReal(*ptrChar.CHARACTER, *ptrVal.DOUBLE = 0, *ptrblnIsCurrency.INTEGER=0)
  Protected result = #minScript_OKAY, value.d, blnNegative
  If *ptrVal
    *ptrVal\d = 0
  EndIf
  If *ptrChar\c = '-'
    blnNegative = #True
    *ptrChar + SizeOf(CHARACTER)
  EndIf
  result = minScript_SymbolsIsPositiveReal(*ptrChar, @value, *ptrblnIsCurrency)
  If result = #minScript_OKAY And *ptrVal
    If blnNegative
      *ptrVal\d = -value
    Else
      *ptrVal\d = value
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function checks the given symbol name to see if it is valid.
;Returns either #minScript_OKAY or #minScript_INVALIDSYMBOLNAME.
Procedure.i minScript_SymbolsIsSymbolNameValid(*ptrChar.CHARACTER)
  Protected result = #minScript_OKAY, numDigits, numAlpha, numSymbols
  While *ptrChar\c
    Select *ptrChar\c
      Case '0' To '9'
        If numAlpha = 0
          result = #minScript_INVALIDSYMBOLNAME
          Break
        EndIf
        numDigits + 1
      Case 'a' To 'z', 'A' To 'Z'
        numAlpha + 1
      Case '_'
        numSymbols + 1
      Default
        result = #minScript_INVALIDSYMBOLNAME
        Break
    EndSelect
    *ptrChar + SizeOf(CHARACTER)
  Wend
  If result = #minScript_OKAY And numAlpha = 0   
    result = #minScript_INVALIDSYMBOLNAME
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////






; IDE Options = PureBasic 5.62 (Windows - x86)
; ExecutableFormat = Shared dll
; CursorPosition = 287
; FirstLine = 273
; Folding = --
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode