﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;User defined types module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.


;-PUBLIC USER DEFINED TYPES FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;Add a 'global' user defined type to the compiler.
;No error checking is performed and so we allow duplicate VT's which would be silly of a host app to allow!
Procedure.i minScript_AddUserDefinedType(name$, vt)
  Protected result = #minScript_OKAY, t1$
  If name$ And vt > #minScript_REFERENCE
    t1$ = LCase(name$)
    result = minScript_SymbolsIsSymbolNameValid(@t1$)
    If result = #minScript_OKAY
      ;Only proceed if the name does not match an existing compiler keyword or run-time function.
      If minScript_IsCompilerKeyword(t1$) = #False
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScript_Globals\threadMutex)
        CompilerEndIf
        If FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), t1$) = 0
          ;We add this symbol to the list of compiler keywords to help the parser.
          If AddMapElement(minScript_Globals\minScriptKeywords(), t1$)
            minScript_Globals\minScriptKeywords() = name$
            If AddMapElement(minScript_Globals\userDefinedTypes(), t1$)
              minScript_Globals\userDefinedTypes() = vt
            Else
              DeleteMapElement(minScript_Globals\minScriptKeywords())
              result = #minScript_OUTOFMEMORY
            EndIf
          Else
            result = #minScript_OUTOFMEMORY
          EndIf
        Else
          result = #minScript_SYMBOLNAMEMATCHESRUNTIMEFUNCTION
        EndIf
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScript_Globals\threadMutex)
        CompilerEndIf
      Else
        result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
      EndIf
    EndIf
  Else
    result = #minScript_INVALIDARGUMENT
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Returns a tab separated list of user-defined types.
Procedure.s minScript_GetListOfUserDefinedTypes()
  Protected name$, fnID
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ForEach minScript_Globals\userDefinedTypes()
    name$ + MapKey(minScript_Globals\userDefinedTypes())
  Next
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn name$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////




; IDE Options = PureBasic 5.46 LTS Beta 2 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 51
; FirstLine = 22
; Folding = -
; EnableUnicode
; EnableThread
; Executable = d.exe.dll.exe.dll